/* 
 * Basis.cc -- ePiX2::Basis class and operations
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 15, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <ahwang@mathcs.holycross.edu>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include "Vector.h"
#include "Errors.h"
#include "Constants.h"
#include "Functions.h"
#include "Basis.h"

namespace ePiX2 {

  Basis::Basis() : m_O(), e1(E_1()), e2(E_2()), e3(E_3()) {}

  Basis::Basis(const Point& loc, 
	       const Vector& v1, const Vector& v2, const Vector& v3) 
    : m_O(loc), e1(v1),e2(v2), e3(v3)
  {
    e1.move_to(loc);
    e2.move_to(loc);
    e3.move_to(loc);

    if (lin_dependent(v1,v2,v3)) // N.B. no graceful attempt to comply
      {
	epix_warning("Linearly independent vectors in basis");
	e1 = E_1(loc);
	e2 = E_2(loc);
	e3 = E_3(loc);
      }

    else
      {
	// Gram-Schmidt constructor, suitable for (sea,sky,eye) bases: 
	// Usually we know sky and eye, and want to preserve eye's direction.
	bool errors = false;

	e3.normalize();
	e2 %= e3;       // orthogonalize

	e2.normalize();
	(e1 %= e3) %= e2;

	e1.normalize();
      }
  } // end of Basis()

  // convert arg to specified coords/scale
  Point Basis::coords(const Point& arg, const double scale) const
  {
    return m_O + scale*(arg.x1()*e1 + arg.x2()*e2 + arg.x3()*e3);
  }

  // convert arg from specified coords/scale
  Point Basis::anti_coords(const Point& arg, const double scale) const
  {
    Vector dir = (1.0/scale)*(arg - m_O);
    return Point((e1|dir), (e2|dir), (e3|dir));
  }


  Basis& Basis::move_to(const Point& arg)
  {
    m_O = arg;
    e1.move_to(arg);
    e2.move_to(arg);
    e3.move_to(arg);
    return *this;
  }

  // rotations about basis elements; separated from rotate() for efficiency
  Basis& Basis::rotate_sea(double angle)
  {
    Vector temp2 = e2;
    Vector temp3 = e3;
    double c=Cos(angle), s=Sin(angle);

    e2 = c*temp2 - s*temp3;
    e3 = s*temp2 + c*temp3;
    return *this;
  }

  Basis& Basis::rotate_sky(double angle)
  {
    Vector temp3 = e3;
    Vector temp1 = e1;
    double c=Cos(angle), s=Sin(angle);

    e3 = c*temp3 - s*temp1;
    e1 = s*temp3 + c*temp1;
    return *this;
  }

  Basis& Basis::rotate_eye(double angle)
  {
    Vector temp1 = e1;
    Vector temp2 = e2;
    double c=Cos(angle), s=Sin(angle);

    e1 = c*temp1 - s*temp2;
    e2 = s*temp1 + c*temp2;
    return *this;
  }

  Basis& Basis::rotate(const double angle, const Vector& axis)
  {
    e1.rotate_by(angle, axis);
    e2.rotate_by(angle, axis);
    e3.rotate_by(angle, axis);

    return *this;
  } // end of Basis::rotate(double, Vector)

  Basis& Basis::reflect(const Vector& axis)
  {
    e1.reflect_by(axis);
    e2.reflect_by(axis);
    e3.reflect_by(axis);

    return *this;
  } // end of Basis::reflect(Vector)

  bool operator== (const Basis& b1, const Basis& b2)
  {
    return ((b1.here() == b2.here()) &&
	    (b1.sea()  == b2.sea())  &&
	    (b1.sky()  == b2.sky())  &&
	    (b1.eye()  == b2.eye()));
  }

} // end of namespace
