/* 
 * Camera.cc -- Camera functions
 *
 * This file is part of ePiX, a preprocessor for creating high-quality 
 * line figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 25, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// n.b. ePiX::trig functions are sensitive to current angle units

#ifdef DEBUG
#include "Debug.h"
#endif

#include "Pair.h"
#include "Point.h"
#include "Vector.h"
#include "Basis.h"
#include "Constants.h"
#include "Functions.h"
#include "Errors.h"
#include "Camera.h"

namespace ePiX2 {

  // viewpt, target, orient
  Camera::Camera(const Point vpt)
    : the_viewpt(vpt), the_target(Origin), the_filter(Neutral())
  {
#ifdef DEBUG
    std::cerr << "Camera()" << std::endl;
#endif
    if (the_viewpt == Origin)
      {
	epix_warning("Cannot initialize Camera at Origin, using (0, 0, oo)");
	the_viewpt = Point(0, 0, EPIX2_INFTY);
      }
    Vector viewer(the_viewpt - the_target);
    the_distance = norm(viewer);

    viewer.normalize();

    // viewer is "eye", if possible E_3 is "sky"
    Vector tmp_sea(E_3()*viewer);

    if (norm(tmp_sea) < EPIX2_EPSILON) // we're on the "z-axis through vpt"
      tmp_sea = E_1();

    // eye*sea = sky
    the_orient = Basis(Origin, tmp_sea, viewer*tmp_sea, viewer);
    adjust();

    the_lens = new Perspective();
  }

  Camera::~Camera(void) { delete the_lens; }

  Camera& Camera::perspective(void)
  {
    delete the_lens;
    the_lens = new Perspective;
    return *this;
  }

  Camera& Camera::orthog(void)
  {
    delete the_lens;
    the_lens = new Orthog;
    return *this;
  }

  Camera& Camera::fisheye(void)
  {
    delete the_lens;
    the_lens = new Fisheye;
    return *this;
  }

  Camera& Camera::bubble(void)
  {
    delete the_lens;
    the_lens = new Bubble;
    return *this;
  }

  // rotate up/down
  Camera& Camera::tilt(const double angle)
  {
    the_orient.rotate_sea(-angle);
    the_target = the_viewpt + (-the_distance)*the_orient.eye();
    return *this;
  }

  // rotate left/right
  Camera& Camera::pan(const double angle)
  {
    the_orient.rotate_sky(-angle);
    the_target = the_viewpt + (-the_distance)*the_orient.eye();
    return *this;
  }

  // rotate about viewing axis
  Camera& Camera::roll(const double angle)
  { 
    the_orient.rotate_eye(-angle); // target unchanged
    return *this;
  }

  bool Camera::is_linear(void)
  {
    return (*the_lens).is_linear();
  }


  Color Camera::operator() (const Color& in_color) const
  {
    Color temp = the_filter;
    return temp.filter(in_color);
  }

  // lens interface
  Pair Camera::operator() (const Point& arg) const
  {
    return (*the_lens)(arg, the_orient, the_viewpt, the_distance);
  }

  Vector Camera::eye(void) const
  { 
    return the_orient.eye();
  }

  Point Camera::viewpt(void) const
  {
    return the_viewpt;
  }


  /* * * lens implementations * * */
  Pair Camera::Perspective::operator()
    (const Point& arg, const Basis& orient, const Point& viewpt,
     const double dist) const
  {
    Vector arg_vector = arg - viewpt;

    // get arg's location in our coordinates
    double u1(orient.sea()|arg_vector);
    double u2(orient.sky()|arg_vector);
    double u3(orient.eye()|arg_vector);

    if (u3 > -EPIX2_EPSILON)
      epix_warning("Perspective looking at point on or behind film plane");

    return (-dist/u3)*Pair(u1, u2);
  } // end of Camera::Perspective::operator()

  Pair Camera::Orthog::operator()
    (const Point& arg, const Basis& orient, const Point& viewpt,
     const double dist) const
  {
    Vector arg_vector(arg - viewpt);
    double u1(orient.sea()|arg_vector);
    double u2(orient.sky()|arg_vector);

    return Pair(u1, u2);
  } // end of Camera::Orthog::operator()

  Pair Camera::Fisheye::operator()
    (const Point& arg, const Basis& orient, const Point& viewpt,
     const double dist) const
  {
    Vector arg_vector(arg - viewpt);
    arg_vector.normalize();

    double u1(orient.sea()|arg_vector);
    double u2(orient.sky()|arg_vector);

    // radial projection to camera sphere, then orthogonal projection
    return dist*Pair(u1, u2);
  } // end of Camera::Fisheye::operator()

  Pair Camera::Bubble::operator()
    (const Point& arg, const Basis& orient, const Point& viewpt,
     const double dist) const
  {
    Vector arg_vector(arg - viewpt);
    arg_vector.normalize();

    double u1(orient.sea()|arg_vector);
    double u2(orient.sky()|arg_vector);
    double u3(orient.eye()|arg_vector);

    // radial projection to camera sphere, then stereographic projection
    return (2.0/(1-u3))*Pair(u1, u2);
  } // end of Camera::Bubble::operator()


  // fix target, move viewpt radially along eye()
  Camera& Camera::range(double d)
  {
    if (d == 0)
      d = EPIX2_INFTY;

    the_distance = d;
    the_viewpt = the_target + d*the_orient.eye();
    the_orient.move_to(the_viewpt);

    return *this;
  }

  // fix viewpt, move target radially along eye()
  Camera& Camera::focus(double d)
  {
    if (d == 0)
      d = EPIX2_INFTY;

    the_distance = d;
    the_target = the_viewpt - d*the_orient.eye();

    return *this;
  }

  // private function
  void Camera::adjust()
  {
#ifdef DEBUG
    std::cerr << "Camera::adjust()" << std::endl;
#endif
    Vector temp(the_target - the_viewpt);
    the_distance = norm(temp);

    Point& vpt(the_viewpt);

    if (the_distance < EPIX2_EPSILON)
      {
	epix_warning("Cannot make viewpoint and target the same, no action");
	return;
      }

    // else
    temp *= -1.0/the_distance; // reverse direction (N.B. not vpt - tgt)

    double z_ht(temp|E_3(vpt));

    // too close to "z-axis" through target?
    if (sqrt((1-z_ht)*(1+z_ht)) < EPIX2_EPSILON)
      {
        if (z_ht >= 0)
          the_orient = Basis(vpt,  E_1(vpt), E_2(vpt),  E_3(vpt));

        else // z_ht < 0
          the_orient = Basis(vpt, -E_1(vpt), E_2(vpt), -E_3(vpt));
      }

    else // far enough from axis
      the_orient = Basis(vpt, Vector(vpt, -temp.dx2(), temp.dx1(), 0), 
			 E_3(vpt), temp);
  }

  // fix target, set viewpt
  Camera& Camera::at(const Point& vpt)
  {
    the_viewpt = vpt;
    adjust();

    return *this;
  }

  // fix viewpt, set target
  Camera& Camera::look_at(const Point& tgt)
  {
    the_target = tgt;
    adjust();

    return *this;
  }

  Camera& Camera::filter(const Color& filter)
  {
    the_filter=filter;
    return *this;
  }

} /* end of namespace */
