/* 
 * Circle.cc -- epix2::Circle class
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 15, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include <cmath>

#include "Functions.h"
#include "Point.h"
#include "Vector.h"

#include "Styles.h"

#include "Shard_1d.h"

#include "Circle.h"

namespace ePiX2 {

  const int PTS(48); // points per half-circle

  Circle::Circle(const Point& ctr, const double rad, const Vector& N)
    : m_ctr(ctr), m_rad(rad), m_perp(N)
  {
#ifdef DEBUG
    std::cerr << "Circle::ctor()" << std::endl;
#endif
    Outline3d data(N);

    m_perp.move_to(ctr);
    m_perp.normalize();

    // construct orthonormal basis
    Vector e1(E_1(ctr));

    if (fabs(m_perp|e1) > 0.7) // Magic Number (0.7 ~ sqrt(0.5))
      e1 = E_2(ctr);

    e1 %= m_perp;
    e1.normalize();
    Vector e2(m_perp*e1);

    for (int i=0; i<2*PTS; ++i) // Magic Number
      data.add(Edge3d(ctr+rad*(cos(i*M_PI/PTS)*e1+sin(i*M_PI/PTS)*e2),
		      ctr+rad*(cos((i+1)*M_PI/PTS)*e1+sin((i+1)*M_PI/PTS)*e2),
		      true));

    data.update();
    add_shard(data);
  }


  Circle::Circle(const Circle& circ)
    : Object(circ),  m_ctr(circ.m_ctr), m_rad(circ.m_rad), m_perp(circ.m_perp)
  {
#ifdef DEBUG
    std::cerr << "Circle::copy_ctor()" << std::endl;
#endif
  }

  Circle& Circle::operator=(const Circle& circ)
  {
#ifdef DEBUG
    std::cerr << "Circle::operator=" << std::endl;
#endif
    if (this != &circ)
      {
	Object::operator=(circ); // copy base class

	m_ctr = circ.m_ctr;
	m_rad = circ.m_rad;
	m_perp = circ.m_perp;
      }

    return *this;
  }


  // class bookkeeping
  Circle* Circle::clone(void) const
  {
#ifdef DEBUG
    std::cerr << "Circle::clone()" << std::endl;
#endif
    return new Circle(*this);
  }


  bool Circle::operator== (const Circle& circ) const
  {
    return ((m_ctr == circ.m_ctr) &&
	    (fabs(m_rad-circ.m_rad) < EPIX2_EPSILON) &&
	    (norm(m_perp*circ.m_perp) < EPIX2_EPSILON));
				      
  }
} /* end of namespace */


