/*
 * Color_RGB.cc -- ePiX2::RGB_Color
 *
 * This file is part of ePiX, a program for creating high-quality
 * figures in LaTeX
 *
 * Version 2.0pre
 *
 * Last Change: August 27, 2006
 */

/*
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 */
 
/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include <sstream>
#include <string>

#include <cmath>

#include "Functions.h"
#include "Output.h"

#include "Color.h"
#include "Color_Utils.h"
#include "Color_RGB.h"

namespace ePiX2 {

  RGB_Color::RGB_Color(const double r, const double g, const double b)
    : red(make_r(r, g, b)), green(make_g(r, g, b)), blue(make_b(r, g, b)) { }


  RGB_Color::RGB_Color(const RGB_Densities& s)
    : red(s.m_dens_red), green(s.m_dens_green), blue(s.m_dens_blue) { }

  Color_Base::RGB_Densities RGB_Color::to_rgb() const
  {
    return Color_Base::RGB_Densities(red, green, blue);
  }

  RGB_Color* RGB_Color::clone() const
  {
    return new RGB_Color(*this);
  }


  Color RGB_Color::filter(const Color_Base& color) const
  {
    Color_Base::RGB_Densities s(color.to_rgb());

    return RGB(min(red,   s.m_dens_red),
	       min(green, s.m_dens_green),
	       min(blue,  s.m_dens_blue));
  }

  RGB_Color& RGB_Color::operator*= (const double c)
  {
    double r(red*c);
    double g(green*c);
    double b(blue*c);

    red   = make_r(r, g, b);
    green = make_g(r, g, b);
    blue  = make_b(r, g, b);

    return *this;
  }

  RGB_Color& RGB_Color::blend(const Color_Base& color, const double d)
  {
    Color_Base::RGB_Densities s(color.to_rgb());
    double wt(clip_0_1(d));

    red   += wt*(s.m_dens_red   - red);
    green += wt*(s.m_dens_green - green);
    blue  += wt*(s.m_dens_blue  - blue);

    return *this;
  }

  RGB_Color& RGB_Color::superpose(const Color_Base& color)
  {
    Color_Base::RGB_Densities s(color.to_rgb());

    red   = clip_0_1(red   + s.m_dens_red);
    green = clip_0_1(green + s.m_dens_green);
    blue  = clip_0_1(blue  + s.m_dens_blue);

    return *this;
  }

  RGB_Color& RGB_Color::invert()
  {
    red   = 1 - red;
    green = 1 - green;
    blue  = 1 - blue;

    return *this;
  }

  // string identifier -- not output-specific
  std::string RGB_Color::name() const
  {
    std::ostringstream nm;

    nm << "rgb_" << dtohex(red) << dtohex(green) << dtohex(blue);

    return nm.str();
  }

  std::string RGB_Color::print_to(const Format& fmt) const
  {
    // rd = "round off"
    return fmt.print_color("rgb", name(), rd(red), rd(green), rd(blue));
  }
} /* end of namespace */
