/*
 * Color_Sep.cc -- ePiX2 CMYK Color separation functions
 *
 * This file is part of ePiX, a program for creating high-quality
 * figures in LaTeX
 *
 * Version 2.0pre
 *
 * Last Change: August 14, 2006
 *
 *
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include <sstream>
#include <string>

#include <cmath>

#include "Functions.h"
#include "Output.h"

#include "Color_Utils.h"
#include "Color.h"
#include "Color_Sep.h"

namespace ePiX2 {

  Cyan_Layer::Cyan_Layer(const double arg)
    : m_dens(clip_0_1(arg)) { }

  // Return Cyan channel of color
  Color Cyan_Layer::filter(const Color_Base& color) const
  {
    Color_Base::RGB_Densities s(color.to_rgb());
    if (Color(color).is_unset())
      s = Color_Base::RGB_Densities(1,1,1);

    return CyanK(min(m_dens, 1-s.m_dens_red));
  }

  Cyan_Layer* Cyan_Layer::clone() const
  {
    return new Cyan_Layer(*this);
  }

  Color_Base::RGB_Densities Cyan_Layer::to_rgb() const
  {
    return Color_Base::RGB_Densities(1-m_dens, 1, 1);
  }

  // string identifier -- not output-specific
  std::string Cyan_Layer::name() const
  {
    std::ostringstream nm;
    nm << "c_proc_" << dtohex(m_dens);
    return nm.str();
  }

  std::string Cyan_Layer::print_to(const Format& fmt) const
  {
    return fmt.print_color("cmy", name(), rd(m_dens), rd(0), rd(0), rd(0));
  }


  // Magenta_Layer
  Magenta_Layer::Magenta_Layer(const double arg)
    : m_dens(clip_0_1(arg)) { }

  // Return Magenta channel of color
  Color Magenta_Layer::filter(const Color_Base& color) const
  {
    Color_Base::RGB_Densities s(color.to_rgb());
    if (Color(color).is_unset())
      s = Color_Base::RGB_Densities(1,1,1);

    return MagentaK(min(m_dens, 1-s.m_dens_green));
  }

  Magenta_Layer* Magenta_Layer::clone() const
  {
    return new Magenta_Layer(*this);
  }

  Color_Base::RGB_Densities Magenta_Layer::to_rgb() const
  {
    return Color_Base::RGB_Densities(1, 1-m_dens, 1);
  }

  // string identifier -- not output-specific
  std::string Magenta_Layer::name() const
  {
    std::ostringstream nm;
    nm << "m_proc_"  << dtohex(m_dens);
    return nm.str();
  }

  std::string Magenta_Layer::print_to(const Format& fmt) const
  {
    return fmt.print_color("cmy", name(), rd(0), rd(m_dens), rd(0), rd(0));
  }


  // Yellow_Layer
  Yellow_Layer::Yellow_Layer(const double arg)
    : m_dens(clip_0_1(arg)) { }

  // Return Yellow channel of color
  Color Yellow_Layer::filter(const Color_Base& color) const
  {
    Color_Base::RGB_Densities s(color.to_rgb());
    if (Color(color).is_unset())
      s = Color_Base::RGB_Densities(1,1,1);

    return YellowK(min(m_dens, 1-s.m_dens_blue));
  }

  Yellow_Layer* Yellow_Layer::clone() const
  {
    return new Yellow_Layer(*this);
  }

  Color_Base::RGB_Densities Yellow_Layer::to_rgb() const
  {
    return Color_Base::RGB_Densities(1, 1, 1-m_dens);
  }

  // string identifier -- not output-specific
  std::string Yellow_Layer::name() const
  {
    std::ostringstream nm;
    nm << "y_proc_" << dtohex(m_dens);
    return nm.str();
  }

  std::string Yellow_Layer::print_to(const Format& fmt) const
  {
    return fmt.print_color("cmy", name(), rd(0), rd(0), rd(m_dens), rd(0));
  }


  // Black_Layer
  Black_Layer::Black_Layer(const double arg)
    : m_dens(clip_0_1(arg)) { }

  // Return Black channel of color
  Color Black_Layer::filter(const Color_Base& color) const
  {
    Color_Base::RGB_Densities s(color.to_rgb());
    if (Color(color).is_unset())
      s = Color_Base::RGB_Densities(1,1,1);

    double c(1-s.m_dens_red), m(1-s.m_dens_green), y(1-s.m_dens_blue);
    double k(min(min(c, m), y));

    return CMYK_Black(min(m_dens, k));
  }

  Black_Layer* Black_Layer::clone() const
  {
    return new Black_Layer(*this);
  }

  Color_Base::RGB_Densities Black_Layer::to_rgb() const
  {
    return Color_Base::RGB_Densities(1-m_dens, 1-m_dens, 1-m_dens);
  }

  // string identifier -- not output-specific
  std::string Black_Layer::name() const
  {
    std::ostringstream nm;
    nm << "k_proc_" << dtohex(m_dens);
    return nm.str();
  }

  std::string Black_Layer::print_to(const Format& fmt) const
  {
    return fmt.print_color("cmy", name(), rd(0), rd(0), rd(0), rd(m_dens));
  }


  // null operations
  Cyan_Layer& Cyan_Layer::operator*= (const double x) { return *this; }
  Cyan_Layer& Cyan_Layer::blend(const Color_Base& col, const double d)
  { return *this; }

  Cyan_Layer& Cyan_Layer::superpose(const Color_Base& col) { return *this; }
  Cyan_Layer& Cyan_Layer::invert() { return *this; }


  Magenta_Layer& Magenta_Layer::operator*= (const double x) { return *this; }
  Magenta_Layer& Magenta_Layer::blend(const Color_Base& col, const double d)
  { return *this; }

  Magenta_Layer& Magenta_Layer::superpose(const Color_Base& col) { return *this; }
  Magenta_Layer& Magenta_Layer::invert() { return *this; }


  Yellow_Layer& Yellow_Layer::operator*= (const double x) { return *this; }
  Yellow_Layer& Yellow_Layer::blend(const Color_Base& col, const double d)
  { return *this; }

  Yellow_Layer& Yellow_Layer::superpose(const Color_Base& col) { return *this; }
  Yellow_Layer& Yellow_Layer::invert() { return *this; }


  Black_Layer& Black_Layer::operator*= (const double x) { return *this; }
  Black_Layer& Black_Layer::blend(const Color_Base& col, const double d)
  { return *this; }

  Black_Layer& Black_Layer::superpose(const Color_Base& col) { return *this; }
  Black_Layer& Black_Layer::invert() { return *this; }

} /* end of namespace */
