/* 
 *  Compound.cc -- epix2::Compound Object class
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef DEBUG
#include "Debug.h"
#endif

#include <vector>

#include "Functions.h"
#include "Point.h"
#include "Vector.h"
#include "Basis.h"
#include "Gyro.h"

#include "Knife.h"

#include "Polygon.h"

#include "Compound.h"

namespace ePiX2 {

  typedef std::vector<Object*>::iterator Op;
  typedef std::vector<Object*>::const_iterator Opc;

  Compound::Compound(void) { }

  Compound::Compound(const Compound& comp)
    : Object(comp)
  {
    for (Opc obj=comp.m_atoms.begin(); obj!=comp.m_atoms.end(); ++obj)
      *this << *((*obj)->clone());
  }

  Compound& Compound::operator= (const Compound& comp)
  {
    if (this != &comp) // Compounds distinguished by address
      {
	std::vector<Object*> new_atoms;

	for (Opc obj=comp.m_atoms.begin(); obj!=comp.m_atoms.end(); ++obj)
	  new_atoms.push_back((*obj)->clone()); // copy atoms

	for (Opc obj=comp.m_atoms.begin(); obj!=comp.m_atoms.end(); ++obj)
	  delete *obj; // free old memory

	Object::operator=(comp);
	swap(m_atoms, new_atoms);
      }
    return *this;
  }

  Compound::~Compound()
  {
    for (Opc obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      delete *obj;
  }

  Compound* Compound::clone(void) const
  {
    return new Compound(*this);
  }

  // add an Object
  Compound& Compound::operator<< (const Object& obj)
  {
    m_atoms.push_back(obj.clone());
    return *this;
  }

  // User interface
  Compound& Compound::chop(const Knife& knife)
  {
    // chop in internal coordinates; convert to external coordinates at shatter
    Knife our_knife(knife.anti_coords(orient()));

    for (Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->chop(our_knife);

    return *this;
  }

  Compound& Compound::slice(const Knife& knife)
  {
    Knife our_knife(knife.anti_coords(orient()));

    for (Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->slice(our_knife);

    return *this;
  }


  // set attributes
  Compound& Compound::front_pen(const Pen& pen)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->front_pen(pen);

    return *this;
  }

  Compound& Compound::front_ink(const Ink& ink)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->front_ink(ink);

    return *this;
  }

  Compound& Compound::back_pen(const Pen& pen)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->back_pen(pen);

    return *this;
  }

  Compound& Compound::back_ink(const Ink& ink)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->back_ink(ink);

    return *this;
  }

  Compound& Compound::line_pen(const Pen& pen)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->line_pen(pen);

    return *this;
  }

  Compound& Compound::base_pen(const Pen& pen)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->base_pen(pen);

    return *this;
  }

  Compound& Compound::stencil(const Stencil& stencil)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->stencil(stencil);

    return *this;
  }

  Compound& Compound::text(const Color& color)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->text(color);

    return *this;
  }

  Compound& Compound::back(const Color& color)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->back(color);

    return *this;
  }

  Compound& Compound::pad(const Length& len)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->pad(len);

    return *this;
  }

  Compound& Compound::edge(const Pen& pen)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->edge(pen);

    return *this;
  }


  Compound& Compound::align(const align_type A)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->align(A);

    return *this;
  }

  Compound& Compound::font(const font_type F)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->font(F);

    return *this;
  }

  Compound& Compound::angle(const double theta)
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->angle(theta);

    return *this;
  }

  Compound& Compound::offset(std::string arg) // need to pass by value
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->offset(arg);

    return *this;
  }

  Compound& Compound::fix()
  {
    for(Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      (*obj)->fix();

    return *this;
  }


  bool Compound::hides(const Point& vpt, const Point& X) const
  {
    Point new_vpt(Object::anti_coords(vpt));
    Point new_X(Object::anti_coords(X));

    for (Opc obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      if ((*obj)->hides(new_vpt, new_X))
	return true;

    // N.B. Needs updating if our shard list becomes non-empty
    return false;
  }

  // add our Shards to container
  void Compound::shatter(std::vector<Shard*>& S)
  {
    for (Op obj=m_atoms.begin(); obj!=m_atoms.end(); ++obj)
      {
	(*obj)->in_coords(orient()); // update subobjects' positions
	(*obj)->shatter(S);
      }
  }

  /*
  Compound Tetrahedron(const Point& v1, const Point& v2,
		       const Point& v3, const Point& v4)
  {
    Compound value;
    value << Polygon(v1, v2, v4)
	  << Polygon(v2, v3, v4)
	  << Polygon(v3, v1, v4)
	  << Polygon(v1, v3, v2);

    return value;
  }

  // opposite corners, sides parallel to coord planes
  // Compound Block(const Point&, const Point&);

  */
} // end of namespace

