/* 
 * Functions.cc -- non-standard mathematical functions
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 12, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <string>
#include <cmath>

#include "Constants.h"
#include "Errors.h"

#include "Point.h"

#include "Functions.h"

namespace ePiX2 {

  const double radians_per_degree(0.01745329251994329577);   // M_PI/180
  const double degrees_per_radian(57.29577951308232087685);  // 180/M_PI
  const double radians_per_rotation(6.28318530717958647692); // 2*M_PI

  // extended reciprocal, with 1/0=0
  double recip (double x) 
  { 
    return (x== 0 || fabs(x) > EPIX2_INFTY) ? 0 : 1/x;
  }

  // map small args to zero
  double trunc(double x)
  {
    // semi-Magic Number 0.0001; smaller x print as xe-05
    return (fabs(x) < 0.0001) ? 0 : x;
  }

  double snip_to(double var, const double arg1, const double arg2)
  {
    if      (var < min(arg1,arg2)) var = min(arg1,arg2);
    else if (var > max(arg1,arg2)) var = max(arg1,arg2);
    return var;
  }

  // faster special case
  double clip_0_1(const double arg)
  {
    return (arg <= 0) ? 0 : ( arg >= 1 ? 1 : arg );
  }

  // Angle_Units public functions
  double Angle_Units::value() const
  {
#ifdef DEBUG
    std::cerr << "Angle_Units::value() = "
	      << m_units_per_radian << std::endl;
#endif
    return m_units_per_radian;
  }


  // private constructor
  Angle_Units::Angle_Units(const double mult)
    : m_units_per_radian(mult) { }


  Angle_Units& angle_units()
  {
    static Angle_Units* current_angle_units(new Angle_Units(1.0));
    return *current_angle_units;
  }

  // global units-setting functions
  void radians()
  {
    angle_units().m_units_per_radian = 1.0;
  }

  void degrees()
  {
    angle_units().m_units_per_radian = radians_per_degree;
  }

  void rotations()
  {
    angle_units().m_units_per_radian = radians_per_rotation;
  }


  // trig functions with angle units
  double Cos(double theta)
  {
    return std::cos(angle_units().value()*theta);
  }
  double Sin(double theta)
  {
    return std::sin(angle_units().value()*theta);
  }
  double Tan(double theta)
  {
    return std::tan(angle_units().value()*theta);
  }

  // non-mathematical behavior (i.e. vanishing:) at poles
  double Sec(double theta)
  {
    return recip(std::cos(angle_units().value()*theta));
  }
  double Csc(double theta)
  {
    return recip(std::sin(angle_units().value()*theta));
  }
  double Cot(double theta)
  {
    return recip(std::tan(angle_units().value()*theta));
  }

  // and inverses
  double Acos(double arg)
  {
    return std::acos(arg)/angle_units().value();
  }
  double Asin(double arg)
  {
    return std::asin(arg)/angle_units().value();
  }
  double Atan(double arg)
  {
    return std::atan(arg)/angle_units().value();
  }

  // identity function for angle units
  double angle(double t)
  {
    return angle_units().value()*t;
  }

  // standard coordinate systems
  Point cartesian(double x, double y, double z)
  {
    return Point(x, y, z);
  }

  Point cylindrical(double x, double y, double z)
  {
    return Point(x*Cos(y), x*Sin(y), z);
  }

  Point spherical(double x, double y, double z)
  {
    return Point(x*Cos(y)*Cos(z), x*Sin(y)*Cos(z), x*Sin(z));
  }

  // sin(x)/x
  double sinx (double x) { return (x != 0) ? Sin(x)/angle(x) : 1; }

  // signum, x/|x|, defined to be 0 at 0
  double sgn (double x)
  {
    if (fabs(x) < EPIX2_EPSILON)
      return 0;
    else
      return x > 0 ? 1 : -1;
  }

  // Charlie Brown: Period-2 extension of |x| on [-1,1] /\/\/\/\/\/
  double cb (double x)
  {
    return fabs(x - 2*floor((x+1)/2));
  }

  // N.B.: gcd(0,i) = |i|
  int gcd (int i, int j)
  {
    i=abs(i);
    j=abs(j);

    if (i==0 || j==0) // (1,0) and (0,1) coprime, others not
      return i+j;

    else {

      int temp;

      if (j < i) // swap them
	{
	  temp = j;
	  j=i;
	  i=temp;
	}
      // Euclidean algorithm
      while ((temp = j%i)) // i does not evenly divide j
	{
	  j=i;
	  i=temp;
	}
    
      return i;
    }
  }

  /*
  // inf and sup of f on [a,b]
  double inf (double f(double), double a, double b)
  {
    const int N = (int) ceil(fabs(b-a)); // N >= 1 unless a=b
    double y = f(a);
    double dx = (b-a)/(N*EPIX2_ITERATIONS);

    for (int i=1; i <= N*EPIX2_ITERATIONS; ++i)
      if (f(a + i*dx) < y)
	y = f(a + i*dx);

    return y;
  }

  double sup (double f(double), double a, double b)
  {
    const int N = (int) ceil(fabs(b-a)); // N >= 1 unless a=b
    double y = f(a);
    double dx = (b-a)/(N*EPIX2_ITERATIONS);

    for (int i=1; i <= N*EPIX2_ITERATIONS; ++i)
      if (f(a + i*dx) > y)
	y = f(a + i*dx);

    return y;
  }
  */

  static double diff(double f(double), double t)
  {
    /* Hardwired constant 0.01 */
    double dt(0.01*EPIX2_EPSILON); // step size
    return (f(t+0.5*dt)-f(t-0.5*dt))/dt;
  }

  double newton (double f(double), double start)
  {
    double guess = start;
    int count=0; // number of iterations

    /* Hardwired constant 20 */
    while ( (fabs(f(guess)) > EPIX2_EPSILON) && (count < 20) )
      {
	if (fabs(diff(f, guess)) < EPIX2_EPSILON)
	  {
	    epix_warning("Critical point in Newton's method");
	    return guess;
	  }

	guess -= f(guess)/diff(f, guess);
	++count;
      }

    if (count == 20)
      epix_warning("20 iterations in Newton's method");

    return(guess);
  }
} /* end of namespace */
