/*
 *  Functions.h -- epix2 utility functions
 *
 * This file is part of ePiX, a preprocessor for creating high-quality 
 * line figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 12, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef EPIX2_FUNCTIONS
#define EPIX2_FUNCTIONS

namespace ePiX2 {

  class Point;

  // ePiX:: trig functions are sensitive to angle units
  double Cos(double theta);
  double Sin(double theta);
  double Tan(double theta); 

  // non-mathematical behavior (i.e. vanishing:) at poles
  double Sec(double theta);
  double Csc(double theta);
  double Cot(double theta);

  // and inverses
  double Acos(double arg);
  double Asin(double arg);
  double Atan(double arg);

  // identity function for angle units
  double angle(double);

  template<typename T> T identity(T arg) { return arg; }
  template<typename T1,typename T2> T1 proj1(T1 arg1, T2 arg2) { return arg1; }
  template<typename T1,typename T2> T2 proj2(T1 arg1, T2 arg2) { return arg2; }

  // standard coordinate systems
  Point cartesian(double x, double y, double z);
  Point cylindrical(double r, double theta, double z);
  Point spherical(double rho, double longitude, double latitude);

  // utility functions
  double sinx (double);  // sin(x)/x
  double sgn (double);   // x/|x|
  double recip (double); // 1/x if |x| isn't small
  double trunc(double);  // x if |x| isn't small

  // push var to [arg1, arg2] or [0,1]
  double snip_to(double var, const double arg1, const double arg2);
  double clip_0_1(const double);

  // period-2 extension of absolute value on [-1,1]: \/\/\/
  double cb (double);

  int gcd (int, int);

  template<typename T> inline T min(T a, T b) { return a<b ? a : b; }
  template<typename T> inline T max(T a, T b) { return a<b ? b : a; }

  double inf (double f(double), double, double);
  double sup (double f(double), double, double);

  double newton (double f(double), double seed);

  class Angle_Units;
  Angle_Units& angle_units();
  void radians();
  void degrees();
  void rotations();

  class Angle_Units {
  public:
    double value() const;

  private:
    double m_units_per_radian;

    Angle_Units(const double th);

    friend Angle_Units& angle_units();
    friend void radians();
    friend void degrees();
    friend void rotations();

  }; // end of class Angle_Units


  /* Coordinate systems
  void cartesian();
  void cylindrical();
  void spherical();

  // coordinate-dependent constructors
  Point  Pt(const double, const double, const double) const;
  Vector Vec(const Point&, const double, const double, const double);

  class Coordinate_System {
  public:
    static Coordinate_System& coordinates();


  private:
    Point m_coords(const double, const double, const double);

    Coordinate_System(Point f(double, double, double));

    friend void cylindrical();
    friend void cartesian();
    friend void spherical();

  }; // end of class Coordinate_System
  */

} // end of namespace

#endif /* EPIX2_FUNCTIONS */
