/* 
 *  Layer.cc --  ePiX2 binary space partition tree structure
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: July 30, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef DEBUG
#include "Debug.h"
#endif

#include <vector>

#include "Errors.h"

#include "Functions.h"
#include "Point.h"
#include "Vector.h"

#include "Styles.h"

#include "Camera.h"

#include "Shard.h"
#include "Layer.h"

namespace ePiX2 {

  typedef std::vector<Shard*>::const_iterator SVp; // Shard vector pointer

  Layer::Layer(const Vector& N) : m_perp(N), front(0), back(0) { }

  Layer::~Layer(void)
  {
#ifdef DEBUG
    std::cerr << "~Layer()" << std::endl;
#endif
    clear();
  }

  void Layer::perp(const Vector& N)
  {
    m_perp = N;
  }

  Vector Layer::perp(void) const
  {
    return m_perp;
  }

  void Layer::clear(void)
  {
#ifdef DEBUG
    std::cerr << "Layer::clear()" << std::endl;
#endif
    if (back != 0)
      {
	back->clear();
	delete back;
      }

    m_fragments.clear();

    if (front != 0)
      {
	front->clear();
	delete front;
      }
  }

  // Indirection is necessary b/c insertion dep. on Shard_Base type.
  void Layer::add(const Shard* S)
  {
#ifdef DEBUG
    std::cerr << "Layer::add()" << std::endl;
#endif

    if (m_fragments.size() == 0)   // empty Layer
      {
	perp(S->perp());           // initialize normal
	m_fragments.push_back(S->clone()); // add data
      }

    else if (m_perp.same_perp(S->perp())) // coplanar
      {
	m_fragments.push_back(S->clone());
      }

    else
      {
	S->add_to(this);
      }
  }

  void Layer::add_f(const Shard* S)
  {
#ifdef DEBUG
    std::cerr << "Layer::add_f()" << std::endl;
#endif
    if (front == 0)
      {
	front = new Layer(S->perp());
	front->m_fragments.push_back(S->clone());
      }

    else
      front->add(S); // may need to initialize
  }

  void Layer::add_b(const Shard* S)
  {
#ifdef DEBUG
    std::cerr << "Layer::add_b()" << std::endl;
#endif
    if (back == 0)
      {
	back = new Layer(S->perp());
	back->m_fragments.push_back(S->clone());
      }

    else
      back->add(S);
  }

  void Layer::add(const std::vector<Shard*>& LS)
  {
#ifdef DEBUG
    std::cerr << "Layer::add(vector<Shard*>)" << std::endl;
#endif
    for (SVp p=LS.begin(); p!=LS.end(); ++p)
      add(*p);
  }

  // shoot back to front; pass most parameters through
  void Layer::shoot(Screen& S, const Camera& cam, const Color& ambient,
		    const std::vector<Light>& spots, const Fog& fog,
		    const photo_type T) const
  {
#ifdef DEBUG
    std::cerr << "Layer::shoot()" << std::endl;
#endif
    if (this->m_fragments.size() == 0) // nothing to see here, people
      return;

    // get direction to viewer
    Vector dir = cam.viewpt() - m_perp.tail();

    if ((dir|m_perp) >= 0)
      {
	if (back != 0)
	  back->shoot(S, cam, ambient, spots, fog, T);

	for (SVp p=m_fragments.begin(); p!=m_fragments.end(); ++p)
	  (*p)->shoot(S, cam, ambient, spots, fog, T);

	if (front != 0)
	  front->shoot(S, cam, ambient, spots, fog, T);
      }
    else
      {
	if (front != 0)
	  front->shoot(S, cam, ambient, spots, fog, T);

	for (SVp p=m_fragments.begin(); p!=m_fragments.end(); ++p)
	  (*p)->shoot(S, cam, ambient, spots, fog, T);

	if (back != 0)
	  back->shoot(S, cam, ambient, spots, fog, T);
      }
  } // end of Layer::shoot()

} // end of namespace
