/*
 *  Lighting.cc -- ePiX2::Light class and utility functions
 *
 * This file is part of ePiX, a preprocessor for creating high-quality 
 * line figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 14, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <cmath>
#include <string>

#include "Point.h"
#include "Vector.h"
#include "Color.h"

#include "Fog.h"

#include "Styles.h"
#include "Lighting.h"

namespace ePiX2 {

  double Light::distance(const Point& arg) const
  {
    return norm(arg - m_loc);
  }

  // Color we cast at arg, seen through fog
  Color Light::incident(const Fog& fog, const Point& arg) const
  {
    return fog.mutes(m_tint, norm(arg - m_loc));
  }

  // Functions that model reflection from a line or surface element
  Color face_spot(const Ink& ink,
		  const Vector& perp,
		  const Vector& in_dir,
		  const Vector& out_dir,
		  const Color& incident)
  {
    if (ink.gloss() == 0) // we reflect only ambient light
      return 0*ink.color();

    // else: reflect in_dir across perp in the plane they span,
    // then take cos(0.5*angle to viewer) to get intensity.
    Vector tmp(in_dir);
    Vector refl_dir(in_dir - 2*(tmp%=perp));
    double I_spot(pow(0.5*(1+(refl_dir|out_dir)), 10)); // Magic Number

    // light that reaches perp.tail() and is reflected to viewer
    return (I_spot*(ink.gloss()))*(ink.color().filter(incident));
  }


  Color face_ambient(const Ink& ink,
		     const Vector& perp,
		     const Vector& out_dir,
		     const Color& ambt)
  {
    if (ink.gloss() == 1) // we reflect only spot light
      return 0*ink.color();

    // else: reflect with intensity proportional to (3+cos^2(theta))/4
    double I_ambt(0.25*(3+pow((perp|out_dir),2))); // Magic Number

    return I_ambt*(1-ink.gloss())*(ink.color().filter(ambt));
  }

  /*
  Color line_spot(const Ink& ink,
		  const Vector& tgt,
		  const Vector& in_dir,
		  const Vector& out_dir,
		  const Color& incident)
  {
    if (ink.gloss() == 0) // we reflect only ambient light
      return ink.color() *= 0;

    // else: reflected intensity is the product of the sines of the
    // tangent with the incident and viewer angles
    double I_spot = ((in_dir*tgt)|(tgt*out_dir));

    // light that reaches perp.tail() and is reflected to viewer
    return I_spot*(ink.gloss())*(ink.color().filter(incident));
  }
  */

  Color line_ambient(const Ink& ink,
		     const Vector& tgt,
		     const Vector& out_dir,
		     const Color& ambt)
  {
    if (ink.gloss() == 1) // we reflect only spot light
      return ink.color() *= 0;

    // else: reflect with intensity proportional to |sin| of normal angle
    double I_ambt(norm(tgt*out_dir));
    return I_ambt*(1-ink.gloss())*(ink.color().filter(ambt));
  }
} /* end of namespace */
