/* 
 * Mask.cc -- ePiX2::Mask base class and children
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 23, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "Constants.h"
#include "Functions.h"
#include "Errors.h"

#include "Pair.h"
#include "Pair_Map.h"
#include "Edge_Data.h"
#include "Scissors.h"

#include "Length.h"
#include "Styles.h"

#include "Tile.h"
#include "Tile_1d.h"
#include "Tile_2d.h"

#include "Mask.h"

namespace ePiX2 {
  namespace Mask {

    // alignment points
    Pair Mask::tr() const { return Pair(h_max(), v_max()); }
    Pair Mask::tl() const { return Pair(h_min(), v_max()); }
    Pair Mask::bl() const { return Pair(h_min(), v_min()); }
    Pair Mask::br() const { return Pair(h_max(), v_min()); }

    Pair Mask::t()  const { return Pair(h_avg(), v_max()); }
    Pair Mask::b()  const { return Pair(h_avg(), v_min()); }
    Pair Mask::l()  const { return Pair(h_min(), v_avg()); }
    Pair Mask::r()  const { return Pair(h_max(), v_avg()); }
    Pair Mask::c()  const { return Pair(h_avg(), v_avg()); }

    Outline2d  Mask::border(const Color& color, const Length& len) const
    {
      return border(Pen(color, len));
    }

    Rect::Rect(const Pair& arg1, const Pair& arg2)
      : m_hmin(min(arg1.x1(), arg2.x1())),
	m_hmax(max(arg1.x1(), arg2.x1())),
	m_vmin(min(arg1.x2(), arg2.x2())),
	m_vmax(max(arg1.x2(), arg2.x2())) { }


    Rect* Rect::clone() const
    {
      return new Rect(*this);
    }

    // corners and midpoints
    double Rect::h_min() const { return m_hmin; }
    double Rect::v_min() const { return m_vmin; }

    double Rect::h_max() const { return m_hmax; }
    double Rect::v_max() const { return m_vmax; }

    double Rect::h_avg() const { return 0.5*(m_hmin+m_hmax); }
    double Rect::v_avg() const { return 0.5*(m_vmin+m_vmax); }

    double Rect::h_size() const { return m_hmax - m_hmin; }
    double Rect::v_size() const { return m_vmax - m_vmin; }

    /*
    // seek largest t-interval in [0,1] s.t. this->encloses(tail+t*head)
    Edge2d Rect::crop(const Edge2d& E) const
    {
    Pair t_tmp = E.tail();
    Pair h_tmp = E.head();
    double E_h = h_tmp.x1() - t_tmp.x1();
    double E_v = h_tmp.x2() - t_tmp.x2();

    Interval I(0,1);
    // t where we're between vertical lines
    if (fabs(E_h) > EPIX2_EPSILON) // not nearly vertical
      I *= Interval((m_hmin-t_tmp.x1())/E_h, (m_hmax-t_tmp.x1())/E_h);

    else if (t_tmp.x1() < m_hmin || m_hmax < t_tmp.x1()) // out of bounds
      I *= Interval::Empty;

    // otherwise assume vertical sides have no effect

    // t where we're between horizontal lines
    if (fabs(E_v) > EPIX2_EPSILON) // not nearly horizontal
      I *= Interval((m_vmin-t_tmp.x2())/E_v, (m_vmax-t_tmp.x2())/E_v);

    else if (t_tmp.x2() < m_vmin || m_vmax < t_tmp.x2()) // out of bounds
      I *= Interval::Empty;


    if (!I.empty())
      return Edge2d(t_tmp + I.min()*(h_tmp - t_tmp),
		    t_tmp + I.max()*(h_tmp - t_tmp), true);
    else
      return Edge2d(t_tmp, h_tmp, false); // undrawn
      } // end of Rect::crop()
  */

    void Rect::crop(Tile* T) const
    {
      T->crop_by(Scissors(br(), tr()));
      T->crop_by(Scissors(tr(), tl()));
      T->crop_by(Scissors(tl(), bl()));
      T->crop_by(Scissors(bl(), br()));
    }

    bool Rect::encloses(const Pair& arg) const
    {
      const double x = arg.x1();
      const double y = arg.x2();
      return ((m_hmin <= x) && (x <= m_hmax) && 
	      (m_vmin <= y) && (y <= m_vmax));
    }

    Outline2d Rect::border(const Pen& pen) const
    {
      Pen2d pen2d(pen.color(), pen.width());
      Outline2d value(pen2d, Pen2d::Xfine, false);
      value.add(Edge2d(br(), tr(), true));
      value.add(Edge2d(tr(), tl(), true));
      value.add(Edge2d(tl(), bl(), true));
      value.add(Edge2d(bl(), br(), true));

      return value;
    }

    Plate2d Rect::backing(const Color& col) const
    {
      Plate2d value(col, Pen2d::Xfine, false);
      value.add(Edge2d(br(), tr(), true));
      value.add(Edge2d(tr(), tl(), true));
      value.add(Edge2d(tl(), bl(), true));
      value.add(Edge2d(bl(), br(), true));

      return value;
    }

    /*
      //TO DO: Handle maps that aren't affine
      Rect* Rect::map_by(const Pair_Map& f) const
      {
      return new Rect(f(bl()), f(tr()));
      }
    */
  } // end of Mask namespace
} // end of ePiX2 namespace
