/* 
 *  Object.cc -- epix2::Object class
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef DEBUG
#include "Debug.h"
#endif

#include <vector>

#include "Functions.h"
#include "Point.h"
#include "Vector.h"

#include "Gyro.h"

#include "Knife.h"
#include "Shard.h"

#include "Object.h"

namespace ePiX2 {

  typedef std::vector<Shard*>::iterator Sp;
  typedef std::vector<Shard*>::const_iterator Spc;

  Object::Object() { }

  Object::Object(const Object& obj)
    : m_orient(obj.m_orient)
  {
#ifdef DEBUG
    std::cerr << "Object::copy_ctor()" << std::endl;
#endif
    for (Spc f=obj.m_fragments.begin(); f!=obj.m_fragments.end(); ++f)
      m_fragments.push_back((*f)->clone());
  }

  Object& Object::operator= (const Object& obj)
  {
#ifdef DEBUG
    std::cerr << "Object::operator=" << std::endl;
#endif
    if (this != &obj) // Objects distinguished by address
      {
	std::vector<Shard*> new_frags;

	for (Spc f=obj.m_fragments.begin(); f!=obj.m_fragments.end(); ++f)
	  new_frags.push_back((*f)->clone()); // copy data

	for (Spc f=obj.m_fragments.begin(); f!=obj.m_fragments.end(); ++f)
	  delete *f;                          // free old memory

	m_orient = obj.m_orient;

	swap(m_fragments, new_frags);
      }

    return *this;
  }


  Object::~Object()
  {
    for (Spc f=m_fragments.begin(); f!=m_fragments.end(); ++f)  delete *f;
  }

  /*
  Object* Object::clone() const
  {
    return new Object(*this);
  }
  */

  // Affine operations
  Object& Object::move_to(const Point& arg)
  {
    m_orient.move_to(arg);
    return *this;
  }

  Object& Object::move_by(const Vector& dX)
  {
    m_orient.move_by(dX);
    return *this;
  }

  Object& Object::scale (const double c)
  {
    m_orient.scale(c);
    return *this;
  }

  Object& Object::reflect(const Vector& axis)
  {
    m_orient.reflect(axis);
    return *this;
  }

  Object& Object::rotate(const double angle, const Vector& axis)
  {
    m_orient.rotate(angle, axis);
    return *this;
  }

  Object& Object::rotate(const Vector& axis, const double angle)
  {
    m_orient.rotate(angle, axis);
    return *this;
  }


  // cut
  // TO DO: Collect cuts in a separate list, then append to our fragments
  Object& Object::chop(const Knife& knife)
  {
#ifdef DEBUG
    std::cerr << "Object::chop()" << std::endl;
#endif
    if (m_fragments.size() > 0)
      {
	std::vector<Shard*> new_cuts;

	// chop in internal coordinates; convert at shatter
	Knife our_knife(knife.anti_coords(m_orient));

	for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
	  (*frag)->chop(our_knife, new_cuts);

	// append cuts
	if (new_cuts.size() > 0)
	  for (Spc f=new_cuts.begin(); f!=new_cuts.end(); ++f)
	    m_fragments.push_back(*f);  // no clone; allocated in Shard::chop()
      }

    return *this;
  }

  Object& Object::slice(const Knife& knife)
  {
    if (m_fragments.size() > 0)
      {
	std::vector<Shard*> new_cuts;

	// chop in internal coordinates; convert at shatter
	Knife our_knife(knife.anti_coords(m_orient));

	for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
	  (*frag)->slice(our_knife, new_cuts);

	// append cuts
	if (new_cuts.size() > 0)
	  for (Spc f=new_cuts.begin(); f!=new_cuts.end(); ++f)
	    m_fragments.push_back(*f);
      }

    return *this;
  }


  // front face and edge
  Object& Object::front_pen(const Pen& pen)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->front_pen(pen);

    return *this;
  }

  Object& Object::front_ink(const Ink& ink)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->front_ink(ink);

    return *this;
  }

  // back face and edge
  Object& Object::back_pen(const Pen& pen)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->back_pen(pen);

    return *this;
  }

  Object& Object::back_ink(const Ink& ink)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->back_ink(ink);

    return *this;
  }


  // line and base
  Object& Object::line_pen(const Pen& pen)
  {
#ifdef DEBUG
    std::cerr << "Object::line_pen()" << std::endl;
#endif
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->line_pen(pen);

    return *this;
  }

  Object& Object::base_pen(const Pen& pen)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->base_pen(pen);

    return *this;
  }

  // stencil
  Object& Object::stencil(const Stencil& sten)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->stencil(sten);

    return *this;
  }

  // fine-grained stencil control
  Object&  Object::text(const Color& color)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->text(color);

    return *this;
  }

  Object&  Object::back(const Color& color)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->back(color);

    return *this;
  }

  Object&  Object::pad(const Length& len)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->pad(len);

    return *this;
  }

  Object&  Object::edge(const Pen& pen)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->edge(pen);

    return *this;
  }


  Object&  Object::align(const align_type A)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->align(A);

    return *this;
  }

  Object&  Object::font(const font_type F)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->font(F);

    return *this;
  }

  Object&  Object::angle(const double th)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->angle(th);

    return *this;
  }

  Object&  Object::offset(std::string arg)
  {
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      (*frag)->offset(arg);

    return *this;
  }

  Object& Object::fix()
  {
    for (Sp f=m_fragments.begin(); f!=m_fragments.end(); ++f)
      (*f)->fix();

    return *this;
  }


  // set attributes on both sides
  Object& Object::paint(const Ink& ink)
  {
    front_ink(ink);
    back_ink(ink);
    return *this;
  }

  // face edges
  Object& Object::limn(const Pen& pen)
  {
    front_pen(pen);
    back_pen(pen);
    return *this;
  }

  // lines
  Object& Object::stroke(const Pen& pen)
  {
    line_pen(pen);
    base_pen(pen);
    return *this;
  }

  Object& Object::stroke(const Duo_Pen& pen)
  {
    line_pen(pen.line());
    base_pen(pen.base());
    return *this;
  }


  // express ourself in given coordinate system
  Object& Object::in_coords(const Gyro& gyro)
  {
    m_orient.in_coords(gyro);
    return *this;
  }

  bool Object::hides(const Point& vpt, const Point& X) const
  {
    Point new_vpt(m_orient.anti_coords(vpt));
    Point new_X(m_orient.anti_coords(X));

    for (Spc f=m_fragments.begin(); f!=m_fragments.end(); ++f)
      if ((*f)->hides(new_vpt, new_X))
	return true;

    return false;
  }

  Point Object::here(void) const
  {
    return m_orient.here();
  }

  Point  Object::in_coords(const Point& arg) const
  {
    return m_orient.in_coords(arg);
  }

  Point  Object::anti_coords(const Point& arg) const
  {
    return m_orient.anti_coords(arg);
  }

  Vector Object::in_coords(const Vector& arg) const
  {
    return m_orient.in_coords(arg);
  }

  Object& Object::add_shard(const Shard& shard)
  {
    m_fragments.push_back(shard.clone());
    return *this;
  }

  const Gyro& Object::orient() const
  {
    return m_orient;
  }

  Shard* Object::fragment(const unsigned int i)
  {
    return m_fragments.at(i); // throws if i >= m_fragments.size()
  }

  void Object::shatter(std::vector<Shard*>& S)
  {
#ifdef DEBUG
    std::cerr << "Object::shatter()" << std::endl;
#endif
    for (Sp frag=m_fragments.begin(); frag!=m_fragments.end(); ++frag)
      S.push_back(((*frag)->clone())->in_coords(m_orient));
  }
} // end of namespace

