/* 
 *  Object.h -- epix2::Object_Base and Object classes
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * This file defines the ePiX2 scene element interface:
 *
 * Object:
 * 
 *   Base class providing translation, scaling (homothety), rotatation,
 *   and reflection; can be chopped by a half-space or sliced by a plane.
 *
 *   All scenery derives from Object. Child classes:
 *   
 *   *must* implement
 *     [] Copy assignment and copy constructor (not compiler-enforced)
 *     [] clone
 *
 *   *may* reimplement (and need to if they're not rep'd by a Shard* list)
 *     [] Destructor
 *     [] chop, slice, hides
 *     [] front_ink, et. al. (fifteen functions)
 *     [] shatter (add Shards to a specified list)
 */

#ifndef EPIX2_OBJECT
#define EPIX2_OBJECT

#include <string>
#include <vector>

#include "Gyro.h"

namespace ePiX2 {

  class Point;
  class Vector;

  class Knife;

  class Pen;
  class Ink;
  class Stencil;

  class Shard;

  class Object {
  public:
    Object();
    Object(const Object&);
    Object& operator= (const Object&);

    virtual ~Object(void);

    // class bookkeeping
    virtual Object* clone(void) const = 0;
    // virtual bool operator== (const Object&) const = 0;

    // User operations
    Object& move_to(const Point&);   // set origin
    Object& move_by(const Vector&);  // translate

    Object& scale (const double);    // homothety about our origin

    Object& reflect(const Vector& axis);
    Object&  rotate(const double angle, const Vector& axis);
    Object&  rotate(const Vector& axis, const double angle);

    virtual Object&  chop(const Knife&); // half-space cut
    virtual Object& slice(const Knife&); // planar cut

    // Set attributes (low-level)
    // Surfaces
    virtual Object& front_pen(const Pen&);
    virtual Object& front_ink(const Ink&);
    virtual Object& back_pen(const Pen&);
    virtual Object& back_ink(const Ink&);

    // Paths
    virtual Object& line_pen(const Pen&);
    virtual Object& base_pen(const Pen&);

    // Labels
    virtual Object& stencil(const Stencil&) ;
    virtual Object& text(const Color&);
    virtual Object& back(const Color&);
    virtual Object&  pad(const Length&);
    virtual Object& edge(const Pen&);

    virtual Object& align(const align_type);
    virtual Object&  font(const font_type);
    virtual Object& angle(const double);
    // "h_offset x v_offset", e.g., "2x2" (pt), "4x-2pt", "1mm x -4pt"
    virtual Object& offset(std::string); // affects only Labels

    virtual Object& fix(); // make ink/pen/stencil attributes permanent


    // Set attributes (coarse but friendly)
    Object& paint(const Ink&); // color both faces
    Object& limn(const Pen&);  // color both edges

    Object& stroke(const Pen&); // set line style
    Object& stroke(const Duo_Pen&);

    // misc utilities
    // express ourself in given coordinate system
    Object& in_coords(const Gyro&);

    // do we occlude X from vpt?
    virtual bool hides(const Point& vpt, const Point& X) const;

    // express w.r.t. basis/scale
    Point  here(void) const;
    Point  in_coords(const  Point& arg) const;
    Point  anti_coords(const  Point& arg) const;

    Vector in_coords(const Vector& arg) const;

    // add our shards to specified container
    virtual void shatter(std::vector<Shard*>&);

  protected:
    Object& add_shard(const Shard&);
    const Gyro& orient() const;
    Shard* fragment(const unsigned int); // index of Shard

  private:
    Gyro m_orient;
    std::vector<Shard*> m_fragments;
  }; // end of class Object

} /* end of namespace */

#endif /* EPIX2_OBJECT */
