/*
 * PS.cc -- ePiX2::PS (PostScript) output class
 *
 * This file is part of ePiX, a program for creating high-quality
 * figures in LaTeX
 *
 * Version 2.0pre
 *
 * Last Change: December 27, 2006
 */

/*
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 */
 
/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


#ifdef DEBUG
#include "Debug.h"
#endif

#include <string>
#include <sstream>
#include <set>
#include <vector>

#include "Constants.h"
#include "Functions.h"

#include "Edge_Data.h"

#include "Color.h"

#include "Styles.h"
#include "Utils.h"

#include "PS.h"

namespace ePiX2 {

  std::string  PS::print_text(const Pair& here, const Pair& here_offset,
			      const std::string& label,
			      const mark_type mark,
			      const Color& text, const Color& back,
			      const Length& pad, const Pen2d& edge,
			      const align_type align,
			      const font_type font,
			      const double angle) const
  {
    using std::endl;
    std::stringstream obuf;

    switch(font)
      {
      case ROMAN:
	obuf << "/Times-Roman findfont" << endl;
	break;

      case ITALIC:
	obuf << "/Times-Italic findfont" << endl;
	break;

      case BFACE:
	obuf << "/Times-Bold findfont" << endl;
	break;

      case TTYPE:
	obuf << "/Courier findfont" << endl;
	break;
      }

    obuf << "0.2 scalefont" << endl
	 << "setfont" << endl
	 << print(here) << " moveto" << endl
	 << print(text) << endl
	 << "(" << label << ") show" << endl;

    return obuf.str();
  } // end of PS::print_text()


  /* * * print_plate * * */
  // fill color; border; clipped flag; edges
  std::string PS::print_plate(const Color& fc, const Pen2d& line,
			      bool cropped,
			      const std::vector<Edge2d>& edges) const
  {
    using std::endl;

    std::stringstream obuf;
    obuf << "newpath" << endl;

    std::vector<Edge2d>::const_iterator p(edges.begin());
    obuf << print(p->tail()) << " moveto" << endl;

    for (p=edges.begin(); p!=edges.end(); ++p)
      obuf << print(p->tail()) << " lineto" << endl;

    obuf << "closepath" << endl;

    if (!fc.is_unset())
      {
	obuf << "gsave" << endl
	     << print(fc) << endl 
	     << "fill"  << endl
	     << "grestore" << endl;
      }

    // TO DO: if cropped, only draw uncropped edges
    obuf << print(line.color()) << endl
	 << "stroke" << endl << endl;

    return obuf.str();
  } // end of PS::print_plate()


  /* * * print_outline * * */
  // line/base; edges
  // default line color and width, fill color, closed/filled flags, vertices
  std::string PS::print_outline(const Pen2d& line, const Pen2d& base, 
				const std::vector<Edge2d>& edges) const
  {
    const Length lw(line.width());
    const Color  lc(line.color());

    if ( (lw < Length(EPIX2_EPSILON)) || (lc.is_unset()))
      return "";

    using std::endl;

    std::stringstream obuf;
    obuf << "newpath" << endl;

    std::vector<Edge2d>::const_iterator p(edges.begin());
    obuf << print(p->tail()) << " moveto" << endl;

    for (p=edges.begin(); p!=edges.end(); ++p)
      obuf << print(p->tail()) << " lineto" << endl;


    if ((base.color() != lc) &&
	(lw < base.width())) // TO DO: set line width
      {
	obuf << "gsave" << endl
	     << print(base.color()) << endl 
	     << "stroke"  << endl
	     << "grestore" << endl;
      }

    obuf << print(lc) << endl
	 << "stroke" << endl << endl;

    return obuf.str();
  } // end of PS::print_outline()


  std::string PS::print_outline(const Pen2d& line, const Pen2d& base, 
				      const Edge2d& e) const
  {
    using std::endl;

    const Length lw(line.width());
    const Color  lc(line.color());

    if ( (lw < Length(EPIX2_EPSILON)) || (lc.is_unset()))
      return "";

    std::stringstream obuf;
    obuf << "newpath" << endl;

    obuf << print(e.tail()) << " moveto" << endl
	 << print(e.head()) << " lineto" << endl;

    if ((base.color() != lc) &&	(lw < base.width())) // TO DO: set line width
      {
	obuf << "gsave" << endl
	     << print(base.color()) << endl 
	     << "stroke"  << endl
	     << "grestore" << endl;
      }

    obuf << print(lc) << endl
	 << "stroke" << endl << endl;

    return obuf.str();
  } // end of PS::print_outline(Pen2d, Pen2d, Edge2d)


  std::string PS::file_header(void) const
  {
    std::stringstream obuf, comment_buf;
    std::string comment;
    using std::endl;

    comment_buf << "Created by " << version() << " on " << date_and_time();
    comment = comment_buf.str();
    std::string separator(comment.size()+8, '%');
    std::string blank(comment.size(), ' ');

    obuf << separator << endl
	 << "%%  " << blank   << "  %%" << endl
	 << "%%  " << comment << "  %%" << endl
	 << "%%  " << blank   << "  %%" << endl
	 << separator << endl;

    return obuf.str();
  }


  std::string PS::pic_header(const Pair& sw, const Pair& ne, 
			     const Length& len,
			     const std::set<Color>& pallet) const
  {
    using std::endl;

    std::stringstream obuf;
    Length len_in_pt = len;
    len_in_pt *= recip(len_in_pt.magnitude()); // one unit
    len_in_pt.to("pt");
    double scale = len_in_pt.magnitude();

    obuf << "%!PS-Adobe-2.0" << endl
	 << "%%Creator: ePiX-2.0pre" << endl
	 << "gsave" << endl
	 << "0 setlinecap" << endl
	 << "1 setlinejoin" << endl
	 << scale << " " << scale << " scale" << endl
      //	 << recip(scale) << " setlinewidth" << endl << endl;
	 << 0.001 << " setlinewidth" << endl << endl;

    return obuf.str();
  }

  std::string PS::pic_footer(void) const
  {
    std::stringstream obuf;
    obuf << "grestore" << std::endl
	 << "showpage" << std::endl
	 << "%%EOF" << std::endl;

    return obuf.str();
  }

  std::string PS::print_color(const std::string& model,
			      const std::string& name,
			      const double d1, const double d2,
			      const double d3) const
  {
    std::stringstream obuf;

    // write densities
    if (model == "cmy")
      {
	double bk(min(min(d1,d2),d3));
	obuf << d1-bk << " " << d2-bk << " " << d3-bk << " " << bk;
      }
    else
      {
	obuf << d1 << " " << d2 << " " << d3;
      }

    // color model
    obuf << " set" << model;
    if (model == "cmy")
      obuf << "k";

    obuf << "color";

    return obuf.str();
  }

  std::string PS::print_color(const std::string& model,
			      const std::string& name,
			      const double d1, const double d2,
			      const double d3, const double d4) const
  {
    std::stringstream obuf;
    obuf << d1 << " " << d2 << " " << d3 << " " << d4
	 << " set" << model << "color";

    return obuf.str();
  }


  std::string PS::comment(const std::string& msg) const
  {
    std::stringstream obuf;
    obuf << "%%  " << msg << std::endl;

    return obuf.str();
  }

  // Utility functions
  /*
  std::string epix2_rule(const Length& x, const Length& y,
			 const Length& dx,const Length& dy)
  {
    std::stringstream buf;
    buf << "\\kern " << dx << " \\rule[" << dy << "]"
	<< "{" << x << "}{" << y << "}";

    return buf.str();
  }
  */
  static std::string epix2_rule(const double x, const double y)
  {
    std::stringstream buf;
    buf << "\\rule{" << x << "pt}{" << y << "pt}";

    return buf.str();
  }

  std::string PS::print(const align_type arg) const
  {
    std::string val;
    switch(arg) 
      {
      case B:
	val="B";
	break;
      case Br:
	val="Br";
	break;
      case Bl:
	val="Bl";
	break;
      case c:
	val="c";
	break;
      case r:
	val="l"; // opposite of LaTeX
	break;
      case tr:
      case rt:
	val="bl";
	break;
      case t:
	val="b";
	break;
      case tl:
      case lt:
	val="br";
	break;
      case l:
	val="r";
	break;
      case bl:
      case lb:
	val="tr";
	break;
      case b:
	val="t";
	break;
      case br:
      case rb:
	val="tl";
	break;
      }

    return val;
  } // end of PS::print(align_type)

  std::string PS::print(const mark_type mark) const
  {
    std::string val="$";

    switch(mark)
      {
      case MK_NONE:
	val = ""; // N.B. reassign, not append
	break;
      case MK_CIRC:
	val += "\\circle";
	break;
      case MK_SPOT:
	val += "\\circle*";
	break;
      case MK_RING:
	val += "\\circle";
	break;
      case MK_DOT:
	val += "\\circle*";
	break;
      case MK_DDOT:
	val += "\\circle*";
	break;
      case MK_PLUS:
	val += "+";
	break;
      case MK_OPLUS:
	val += "\\oplus";
	break;
      case MK_TIMES:
	val += "\\times";
	break;
      case MK_OTIMES:
	val += "\\otimes";
	break;
      case MK_DIAMOND:
	val += "\\diamond";
	break;
      case MK_UP:
	val += "\\bigtriangleup";
	break;
      case MK_DOWN:
	val += "\\bigtriangledown";
	break;
      case MK_BOX:
	val += epix2_rule(2, 2); // Magic numbers
	break;
      case MK_BBOX:
	val += epix2_rule(1.5, 1.5);
	break;
      case MK_HTICK:
	val += epix2_rule(4, 0.5);
	break;
      case MK_VTICK:
	val += epix2_rule(0.5, 4);
	break;
      }

    if (mark != MK_NONE)
      val +="$";
    return val;
  } // end of PS::print(mark_type)


  std::string PS::print(const font_type f) const
  {
    std::string val = "\\";

    switch(f)
      {
      case ROMAN:
	val = ""; // N.B. reassign
	break;
      case ITALIC:
	val += "textit";
	break;
      case BFACE:
	val += "textbf";
	break;
      case TTYPE:
	val += "texttt";
	break;
      }

    return val;
  } // end of PS::print(font_type)


  std::string PS::print(const Pair& arg) const
  {
    std::stringstream obuf;
    obuf << trunc(arg.x1()) << " " << trunc(arg.x2());
    return obuf.str();
  }

  std::string PS::print(const Color& col) const
  {
    return col.print_to(*this);
    //return col.name();
  }

  std::string PS::print(const Length& len) const
  {
    std::stringstream o;
    o << len.magnitude() << len.units();
    return o.str();
  }
} /* end of namespace */
