/*
 * PSTricks.cc -- ePiX2 PSTricks output format
 *
 * This file is part of ePiX, a program for creating high-quality
 * figures in LaTeX
 *
 * Version 2.0pre
 *
 * Last Change: September 5, 2006
 */

/*
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 */
 
/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include <string>
#include <sstream>
#include <set>
#include <vector>

#include "Constants.h"
#include "Functions.h"

#include "Edge_Data.h"

#include "Color.h"

#include "Styles.h"
#include "Utils.h"

#include "PSTricks.h"

namespace ePiX2 {

  // formatting string to break lines in path data lists
  std::string PST_path_break_init()
  {
    std::stringstream buf;
    buf << std::endl << "  "; // newline, two spaces

    return buf.str();
  }

  const std::string PST_path_break(PST_path_break_init());

  // PSTricks maintains the current state via its data members.
  // When a plate or outline is printed, we check the new attributes
  // against the current state, print PSTricks commands to update the
  // output file state, and update our internal state. To accommodate
  // 2-D figures, the initial state has white filling and a plain black pen
  PSTricks::PSTricks()
    : m_fill(White()),
      m_edge(Black()), m_edge_width(PLAIN_WIDTH),
      m_base(White()), m_base_width(Length(0)),
      m_text_pad(LABEL_PAD) { }


  // private functions
  std::string PSTricks::update_fill_style(const Color& face,
					  const Color& edge,
					  const Length& edge_width) const
  {
    std::stringstream cmd; // output file command
    cmd << "\\psset{";

    bool empty_cmd(true);  // no options found

    if (face != m_fill)
      {
	m_fill = face;   // update current internal color
	
	if (face.is_unset())  // update file stream
	  cmd << "fillstyle=none";

	else
	  cmd << "fillcolor=" << print(face);

	empty_cmd=false;
      }

    if (edge != m_edge)
      {
	m_edge = edge;

	if (!empty_cmd) // from previous block
	  cmd << ",";

	if (edge.is_unset())  // update file stream
	  cmd << "linestyle=none";

	else
	  cmd << "linecolor=" << print(edge);

	empty_cmd=false;
      }

    if (edge_width != m_edge_width)
      {
	m_edge_width = edge_width;

	if (!empty_cmd)
	  cmd << ",";

	cmd << "linewidth=" << print(edge_width);

	empty_cmd=false;
      }

    if (!empty_cmd) // got command options
      {
	cmd << "}" << std::endl; // close \psset
#ifdef DEBUG
	std::cerr << "PSTricks::update_fill_style returning "
		  << cmd.str() << std::endl;
#endif
	return cmd.str();
      }

    else
      return "";
  } // end of update_fill_style


  // Since we don't always want to draw borders, we ask for them in the
  // output stream locally as needed.
  std::string PSTricks::update_line_style(const Color& line,
					  const Length& line_width,
					  const Color& base) const
  {
    std::stringstream cmd; // output file command
    cmd << "\\psset{";

    bool empty_cmd(true);  // no options found

    if (line != m_edge)
      {
	m_edge = line;     // update linecolor

	// TO DO: dashed/dotted lines	
	if (line.is_unset())  // update file stream
	  cmd << "linestyle=none";

	else
	  cmd << "linecolor=" << print(line);

	empty_cmd=false;
      }

    if (line_width != m_edge_width)
      {
	m_edge_width = line_width;

	if (!empty_cmd)
	  cmd << ",";

	cmd << "linewidth=" << print(line_width);

	empty_cmd=false;
      }

    if (base.is_unset()) // set m_base to sane default
      {
	m_base = White();

	if (!empty_cmd)
	  cmd << ",";

	m_base_width = Length(0);
	cmd << "border=0pt";

	empty_cmd = false;
      }

    else if (base != m_base)
      {
	m_base = base;

	if (!empty_cmd)
	  cmd << ",";

	cmd << "bordercolor=" << print(base);

	empty_cmd = false;
      }

    if (!empty_cmd) // got command options
      {
	cmd << "}" << std::endl; // close \psset
#ifdef DEBUG
	std::cerr << "PSTricks::update_line_style returning "
		  << cmd.str() << std::endl;
#endif
	return cmd.str();
      }

    else
      return "";
  } // end of update_line_style


  // Mid-level functions
  std::string  PSTricks::print_text(const Pair& here, const Pair& here_offset,
				    const std::string& label,
				    const mark_type mark,
				    const Color& text, const Color& back,
				    const Length& pad, const Pen2d& edge,
				    const align_type align,
				    const font_type font,
				    const double angle) const
  {
    std::stringstream obuf, cmd;

    if (!back.is_unset())
      {
	// collect style command, update label_pad (framesep) if nec.
	cmd << "[" << "fillcolor=" << print(back);

	if (pad != m_text_pad)
	  {
	    m_text_pad = pad;
	    // print straight to output stream
	    obuf << "\\psset{framesep=" << print(pad) << "}" << std::endl;
	  }

	if (edge.color() != m_edge) // set linecolor
	  {
	    if (edge.color().is_unset())
	      cmd << ",linestyle=none";

	    else
	      cmd << ",linecolor=" << print(edge.color());
	  }

	if (edge.width() != m_edge_width) // set linewidth
	  cmd << ",linewidth=" << print(edge.width());

	cmd << "]";
      }
    // N.B. No label decorations without background color
    else
      cmd << "[fillstyle=none,linestyle=none]";

    // cmd complete

    obuf << "\\rput";

    if (align != c) // not centered
      obuf << "[" << print(align) << "]";

    if (fabs(angle) > EPIX2_EPSILON) // not vertical
      obuf << "{" << angle << "}";

    // text label gets offset
    obuf << print(here_offset) << "{";

    if (!back.is_unset()) // we have a background color
      obuf << "\\psframebox" << cmd.str();

    obuf << "{";

    // label text proper
    obuf << "\\color{" << print(text) << "}";

    if (font != ROMAN)
      obuf << print(font) << "{";

    obuf << label;

    if (font != ROMAN)
      obuf << "}";

    obuf << "}}" << std::endl;

    // print mark; not offset
    if (mark != MK_NONE)
      obuf << "\\rput" << print(here) << "{"
	   << "\\color{" << print(text) << "}"
	   << print(mark) << "}" << std::endl;

    return obuf.str();
  } // end of PSTricks::print_text()


  /* * * print_plate * * */
  // fill color; border; clipped flag; edges
  std::string PSTricks::print_plate(const Color& fc, const Pen2d& line,
				    bool cropped,
				    const std::vector<Edge2d>& edges) const
  {
    std::stringstream value, obuf; // obuf gets lines broken

    // send straight to output (no line breaking)
    value << comment("Plate")
	  << update_fill_style(fc, line.color(), line.width());

    // break long line
    obuf << "\\pspolygon";

    // If not all our edges will be drawn, use fillcolor, then re-draw
    if (cropped)
      {
	if (fc.is_unset())
	  obuf << "[linewidth=0pt]";

	else
	  obuf << "[linecolor=" << print(fc) << ",linewidth=0pt]";
      }

    std::vector<Edge2d>::const_iterator p(edges.begin());

    const Pair init((*p).tail()); // store first vertex

    obuf << print(init);

    // on first pass, check for visible edges
    bool some_edge_drawn(false);

    for (p=edges.begin(); p!=edges.end(); ++p)
      {
	if ((*p).seen())
	  some_edge_drawn=true;

	obuf << print((*p).head());
      }

    // close path if necessary
    if ((*(--p)).head() != init)
      obuf << print(init);

    value << break_lines(obuf.str(), PST_path_break) << std::endl;

    // If not cropped, we're done. Otherwise, re-draw border if necessary.
    if (some_edge_drawn && cropped &&
	line.color() != m_fill && Length(0) < line.width())
      {
	value << comment("Plate boundary");

	bool in_path(false); // current path started?

	for (p=edges.begin(); p!=edges.end(); ++p)
	  {
	    if ((*p).seen())
	      {
		if (!in_path) // start path
		  {
		    value << "\\psline[fillstyle=none]"
			  << print((*p).tail());
		    in_path = true;
		  }

		value << print((*p).head());
	      }

	    else // not seen
	      {
		if (in_path) // end path segment
		  value << std::endl;

		in_path = false;
	      }
	  } // all edges re-examined

	if (in_path) // last edge visible
	  value << std::endl;

      } // finished re-drawing border

    // Cuts handled by Plate3d::print_to()
    return value.str();
  } // end of PSTricks::print_plate()


  /* * * print_outline * * */
  // line/base; edges
  std::string PSTricks::print_outline(const Pen2d& line, const Pen2d& base, 
				      const std::vector<Edge2d>& edges) const
  {
    const Length lw(line.width());
    const Color  lc(line.color());

    if ( (lw < Length(EPIX2_EPSILON)) || (lc.is_unset()))
      return "";

    // else
    // border width command, return value, path data buffer
    std::stringstream cmd, value, obuf;

    const Color  bc(base.color());
    const Length bw(base.width());

    value << comment("Outline")
	  << update_line_style(lc, lw, bc);

    cmd << "[fillstyle=none";

    // draw border if base is wider and of different color
    if ((bc != lc) && (lw < bw))
      {
	Length tmp_b(bw), tmp_l(lw);
	tmp_b += (tmp_l *= -1); // subtract lw from bw
	tmp_b *= 0.5;

	cmd << ",border=" << print(tmp_b);
      }

    cmd << "]";

    bool in_path=false; // current path started?

    std::vector<Edge2d>::const_iterator p=edges.begin();

    for (p=edges.begin(); p!=edges.end(); ++p)
      {
	if ((*p).seen())
	  {
	    if (!in_path) // start path
	      {
		obuf << "\\psline" << cmd.str() << print((*p).tail());
		in_path = true;
	      }

	    obuf << print((*p).head());
	  }

	else // not seen
	  {
	    if (in_path) // end path segment
	      obuf << std::endl;

	    in_path = false;
	  }
      }

    if (in_path) // last edge visible
      obuf << std::endl;

    value << break_lines(obuf.str(), PST_path_break);
    return value.str();
  } // end of print_outline(Pen2d, Pen2d, std::vector<Edge2d>)


  std::string PSTricks::print_outline(const Pen2d& line, const Pen2d& base, 
				      const Edge2d& e) const
  {
    const Length lw = line.width();
    const Color lc = line.color();

    if ( (lw == Length(0)) || (lc.is_unset()))
      return "";

    // else
    std::stringstream cmd, obuf; // line style command, and our return value

    const Color bc = base.color();
    const Length bw = base.width();


    obuf << comment("Cut")
	 << update_line_style(lc, lw, bc);

    cmd << "[fillstyle=none";

    // draw border if base is wider and of different color
    if ((bc != lc) && (lw < bw))
      {
	Length tmp_b(bw), tmp_l(lw);
	tmp_b += (tmp_l *= -1); // subtract lw from bw
	tmp_b *= 0.5;

	cmd << ",border=" << print(tmp_b);
      }

    cmd << "]";

    obuf << "\\psline" << cmd.str();

    obuf << print(e.tail()) << print(e.head()) << std::endl;

    return obuf.str();
  } // end of print_outline(Pen2d, Pen2d, Edge)


  std::string PSTricks::file_header(void) const
  {
    std::stringstream obuf, comment_buf;
    std::string comment;
    using std::endl;

    comment_buf << "Created by " << version() << " on " << date_and_time();
    comment = comment_buf.str();
    std::string separator(comment.size()+8, '%');
    std::string blank(comment.size(), ' ');

    obuf << separator << endl
	 << "%%  " << blank   << "  %%" << endl
	 << "%%  " << comment << "  %%" << endl
	 << "%%  " << blank   << "  %%" << endl
	 << separator << endl;

    return obuf.str();
  }

  std::string PSTricks::pic_header(const Pair& sw, const Pair& ne, 
				   const Length& len,
				   const std::set<Color>& pallet) const
  {
    using std::endl;
    std::stringstream obuf;

    /*
      << m_face_fill.print_to(*this)
      << m_line.color().print_to(*this)
    */

    obuf << file_header();

    std::set<Color>::const_iterator cp;
    for (cp=pallet.begin(); cp!=pallet.end(); ++cp)
      obuf << (*cp)->print_to(*this); 

    // TO DO: offsets
    obuf << "\\psset{fillstyle=solid"
	 << ",fillcolor=" << print(m_fill) << "}" << endl
	 << "\\psset{linecolor=" << print(m_edge)
	 << ",linewidth=" << print(m_edge_width) << "}" << endl
	 << "\\psset{unit=1" << len.units() << "}" << endl
	 << "\\begin{pspicture}" << print(sw) << print(ne) << endl;

    return obuf.str();
  }

  std::string PSTricks::pic_footer(void) const
  {
    std::stringstream obuf;
    obuf << "\\end{pspicture}" << std::endl;
    return obuf.str();
  }

  std::string PSTricks::print_color(const std::string& model,
				    const std::string& name,
				    const double d1, const double d2,
				    const double d3) const
  {
    std::stringstream obuf;
    obuf << "\\new" << model;

    if (model == "cmy")
      {
	obuf << "k";
	double bk(min(min(d1,d2),d3));

	obuf << "color{" << name << "}{"
	     << d1-bk << " " << d2-bk << " " << d3-bk << " " << bk
	     << "}" << std::endl;
      }

    else
      obuf << "color{" << name << "}{"
	   << d1 << " " << d2 << " " << d3
	   << "}" << std::endl;

    return obuf.str();
  }

  std::string PSTricks::print_color(const std::string& model,
				    const std::string& name,
				    const double d1, const double d2,
				    const double d3, const double d4) const
  {
    std::stringstream obuf;
    obuf << "\\new" << model << "color{" << name << "}{"
	 << d1 << " " << d2 << " " << d3 << " " << d4 << "}" << std::endl;
    return obuf.str();
  }

  /*
  // *complete* figure header/footer, including pic_header/footer
  std::string PSTricks::fig_header(const Pair& sw, const Pair& ne,
				   const Length& len,
				   const std::set<Color>& pallet) const
  {
    std::stringstream obuf;
    obuf << "\\begin{figure}[hbt]" << std::endl
	 << "\\begin{center}" << std::endl
	 << pic_header(sw, ne, len, pallet);

    return obuf.str();
  }

  std::string PSTricks::fig_footer(void) const
  {
    std::stringstream obuf;
    obuf << pic_footer()
	 << "\\end{center}" << std::endl
	 << "\\end{figure}" << std::endl;

    return obuf.str();
  }

  std::string PSTricks::fig_caption(const std::string& caption) const
  {
    std::stringstream obuf;
    obuf << "\\caption{" << caption << "}" << std::endl;
    return obuf.str();
  }

  std::string PSTricks::fig_xref(const std::string& xref) const
  {
    std::stringstream obuf;
    obuf << "\\label{" << xref << "}" << std::endl;
    return obuf.str();
  }
  */

  std::string PSTricks::comment(const std::string& msg) const
  {
    std::stringstream obuf;
    obuf << "%%  " << msg << std::endl;

    return obuf.str();
  }

  // Utility functions
  /*
  std::string epix2_rule(const Length& x, const Length& y,
			 const Length& dx=Length(0), 
			 const Length& dy=Length(0))
  {
    std::stringstream buf;
    buf << "\\kern " << dx << " \\rule[" << dy << "]"
	<< "{" << x << "}{" << y << "}";

    return buf.str();
  }
  */

  static std::string epix2_rule(const double x, const double y)
  {
    std::stringstream buf;
    buf << "\\rule{" << x << "pt}{" << y << "pt}";

    return buf.str();
  }

  std::string PSTricks::print(const align_type arg) const
  {
    std::string val;
    switch(arg) 
      {
      case B:
	val="B";
	break;
      case Br:
	val="Br";
	break;
      case Bl:
	val="Bl";
	break;
      case c:
	val="c";
	break;
      case r:
	val="l"; // opposite of LaTeX
	break;
      case tr:
      case rt:
	val="bl";
	break;
      case t:
	val="b";
	break;
      case tl:
      case lt:
	val="br";
	break;
      case l:
	val="r";
	break;
      case bl:
      case lb:
	val="tr";
	break;
      case b:
	val="t";
	break;
      case br:
      case rb:
	val="tl";
	break;
      }

    return val;
  } // end of PSTricks::print(align_type)

  std::string PSTricks::print(const mark_type mark) const
  {
    std::string val="$";

    switch(mark)
      {
      case MK_NONE:
	val = ""; // N.B. reassign, not append
	break;
      case MK_CIRC:
	val += "\\circle";
	break;
      case MK_SPOT:
	val += "\\circle*";
	break;
      case MK_RING:
	val += "\\circle";
	break;
      case MK_DOT:
	val += "\\circle*";
	break;
      case MK_DDOT:
	val += "\\circle*";
	break;
      case MK_PLUS:
	val += "+";
	break;
      case MK_OPLUS:
	val += "\\oplus";
	break;
      case MK_TIMES:
	val += "\\times";
	break;
      case MK_OTIMES:
	val += "\\otimes";
	break;
      case MK_DIAMOND:
	val += "\\diamond";
	break;
      case MK_UP:
	val += "\\bigtriangleup";
	break;
      case MK_DOWN:
	val += "\\bigtriangledown";
	break;
      case MK_BOX:
	val += epix2_rule(2, 2); // Magic numbers
	break;
      case MK_BBOX:
	val += epix2_rule(1.5, 1.5);
	break;
      case MK_HTICK:
	val += epix2_rule(4, 0.5);
	break;
      case MK_VTICK:
	val += epix2_rule(0.5, 4);
	break;
      }

    if (mark != MK_NONE)
      val +="$";
    return val;
  } // end of PSTricks::print(mark_type)


  std::string PSTricks::print(const font_type f) const
  {
    std::string val = "\\";

    switch(f)
      {
      case ROMAN:
	val = ""; // N.B. reassign
	break;
      case ITALIC:
	val += "textit";
	break;
      case BFACE:
	val += "textbf";
	break;
      case TTYPE:
	val += "texttt";
	break;
      }

    return val;
  } // end of PSTricks::print(font_type)


  std::string PSTricks::print(const Pair& arg) const
  {
    std::stringstream o;
    o << "(" << ePiX2::trunc(arg.x1()) << "," << ePiX2::trunc(arg.x2()) << ")";
    return o.str();
  }

  std::string PSTricks::print(const Color& col) const
  {
    return col.name();
  }

  std::string PSTricks::print(const Length& len) const
  {
    std::stringstream o;
    o << len.magnitude() << len.units();
    return o.str();
  }

} /* end of namespace */
