/* 
 * Pair_Map.cc -- ePiX2::Pair_Map(_Base) class functions
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 21, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include "Constants.h"
#include "Functions.h"
#include "Pair.h"
#include "Mask.h"
#include "Pair_Map.h"

namespace ePiX2 {

  Pair Ident(Pair arg) { return arg; }

  Pair_Map::Func::Func(Pair (*f)(Pair))
      : mapping(f)
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::Func::Func()" << std::endl;
#endif
  }

  Pair_Map::Func* Pair_Map::Func::clone() const
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::Func::clone()" << std::endl;
#endif
    return new Func(*this);
  }

  Pair Pair_Map::Func::operator() (const Pair arg) const
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::Func::operator(Pair)" << std::endl;
#endif
    return mapping(arg);
  }


  // Affine
  Pair_Map::Affine::Affine(const Mask::Mask& P, const Mask::Mask& C)
    : p_bl(P.bl()), c_bl(C.bl()),
      scale(Pair(P.h_size()/C.h_size(), P.v_size()/C.v_size()))
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::Affine::Affine()" << std::endl;
#endif
  }


  Pair_Map::Affine* Pair_Map::Affine::clone() const
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::Affine::clone()" << std::endl;
#endif
    return new Affine(*this);
  }

  Pair Pair_Map::Affine::operator() (const Pair arg) const
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::Affine::operator(Pair)" << std::endl;
#endif
    return p_bl + (scale&(arg - c_bl));
  }

  // Rotate
  Pair_Map::Rotate::Rotate(const double th, const Pair ctr)
    : angle(th), center(ctr) { }

  Pair_Map::Rotate* Pair_Map::Rotate::clone() const
  {
    return new Rotate(*this);
  }

  Pair Pair_Map::Rotate::operator() (const Pair arg) const
  {
    Pair dir(arg-center);
    double C(Cos(angle)), S(Sin(angle));
    return center + Pair(C*dir.x1() - S*dir.x2(), C*dir.x2() + S*dir.x1());
  }

  // Reflect
  Pair_Map::Reflect::Reflect(const Pair p1, const Pair p2)
    : fix1(p1), fix2(p2) { }

  Pair_Map::Reflect::Reflect(const double th, const Pair ctr)
    : fix1(ctr), fix2(ctr + Pair(Cos(th), Sin(th))) { }


  Pair_Map::Reflect* Pair_Map::Reflect::clone() const
  {
    return new Reflect(*this);
  }

  Pair Pair_Map::Reflect::operator() (const Pair arg) const
  {
    if (norm(fix2-fix1) > EPIX2_EPSILON)
      {
	Pair dir(J(fix2-fix1));
	dir *= recip(norm(dir));
	return arg - (2*(dir|(arg-fix1)))*dir;
      }
    else // "reflect" about fix1=fix2
      return 2*fix1 - arg;
  }


  // Pair_Map class functions
  Pair_Map::Pair_Map()
    : outer(new Func(Ident)), inner(0)
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::Pair_Map()" << std::endl;
#endif
  }

  Pair_Map::Pair_Map(const Map_Base& the_map)
    : outer(the_map.clone()), inner(0)
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::Pair_Map(const Map_Base&)" << std::endl;
#endif
  }

  Pair_Map::Pair_Map(const Pair_Map& pm)
    : outer(pm.outer->clone()), inner(pm.inner->clone())
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::Pair_Map(const Pair_Map&)" << std::endl;
#endif
  }

  Pair_Map& Pair_Map::operator=(const Pair_Map& pm)
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::operator=" << std::endl;
#endif
    if (this != &pm)
      {
	Map_Base* tmp_outer(pm.outer->clone());
	Pair_Map* tmp_inner(pm.inner->clone());

	delete outer;
	outer = tmp_outer;

	delete inner;
	inner = tmp_inner;
      }

    return *this;
  }

  Pair_Map::~Pair_Map()
  {
#ifdef DEBUG
    std::cerr << "~Pair_Map()" << std::endl;
#endif
    if (inner != 0)
      delete inner;

    delete outer;
  }

  Pair_Map* Pair_Map::clone() const
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::clone()" << std::endl;
#endif
    return new Pair_Map(*this);
  }

  Pair Pair_Map::operator() (const Pair arg) const
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::operator(Pair)" << std::endl;
#endif
    Pair value(arg);

    if (inner != 0)
      value = (*inner)(arg);

    return (*outer)(value);
  }

  Pair_Map Pair_Map::operator() (const Pair_Map& precomp) const
  {
#ifdef DEBUG
    std::cerr << "Pair_Map::operator(Pair_Map)" << std::endl;
#endif
    Pair_Map comp(*outer); // initialize with our mapping

    if (inner != 0) 
      comp.inner = ((*inner)(precomp)).clone();

    else
      comp.inner = precomp.clone();

    return comp;
  }


  // Named constructors
  Pair_Map Function(Pair f(Pair))
  {
#ifdef DEBUG
    std::cerr << "Pair_Map Function" << std::endl;
#endif
    return *(new Pair_Map::Func(f));
  }

  Pair_Map Affine(const Mask::Mask& P, const Mask::Mask& C)
  {
#ifdef DEBUG
    std::cerr << "Pair_Map Affine" << std::endl;
#endif
    return *(new Pair_Map::Affine(P, C));
  }

  Pair_Map Rotate(const double th, const Pair ctr)
  {
    return *(new Pair_Map::Rotate(th, ctr));
  }

  // fixed points
  Pair_Map Reflect(const Pair p1, const Pair p2)
  {
    return *(new Pair_Map::Reflect(p1, p2));
  }

  // fixed point and angle of axis
  Pair_Map Reflect(const double th, const Pair ctr)
  {
    return *(new Pair_Map::Reflect(th, ctr));
  }
} /* end of namespace */
