/* 
 *  Picture.cc -- epix2::Picture class
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 23, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include <fstream>
#include <sstream>

#include "Pair.h"
#include "Mask.h"
#include "Pair_Map.h"
#include "Errors.h"

#include "Styles.h"

#include "Tile.h"
#include "Tile_1d.h" // for m_page.m_border
#include "Tile_2d.h" // for m_page.m_backing

#include "Output.h"

#include "Screen.h"
#include "Picture.h"

namespace ePiX2 {

  Picture::Picture(const Pair arg1, const Pair arg2, std::string sz)
    : m_data(Screen(arg1, arg2)), m_page(Screen(Pair(), Pair())),
      m_hoff(Length(0)), m_voff(Length(0))

  {
    get_lengths(sz, m_width, m_height);

    // reconcile height units
    m_height.to(m_width.units());

    // page has actual size
    m_page=Screen(Pair(0,0), Pair(m_width.magnitude(), m_height.magnitude()));
  }

  // same syntax as size
  Picture& Picture::offset(const std::string& sz)
  {
    get_lengths(sz, m_hoff, m_voff);

    // reconcile units
    m_voff.to(m_hoff.units());
    return *this;
  }


  // superimpose a layer by affine scaling
  Picture& Picture::paste(const Screen& child)
  {
    m_data.paste(child);
    return *this;
  }

  Picture& Picture::paste(const Screen& child, const Mask::Rect& inset)
  {
    m_data.paste(child, inset);
    return *this;
  }

  /*
  Picture& Picture::paste(const Screen& child, const Mask::Rect& cutout, 
		      const Mask::Rect& inset)
  {
    m_data.paste(child, cutout, inset);
    return *this;
  }
  */

  // or by a specified Pair_Map
  Picture& Picture::paste(const Screen& child, const Pair_Map& f)
  {
    m_data.paste(child, f);
    return *this;
  }

  /*
  Picture& Picture::paste(const Screen& child, const Mask::Rect& cut, 
			  const Pair_Map& f)
  {
    m_data.paste(child, cut, f);
    return *this;
  }
  */

  Picture& Picture::clear(void)
  {
    m_data.clear();
    m_page.clear();
    return *this;
  }

  Picture& Picture::crop(void)
  {
    m_data.crop();
    return *this;
  }

  Picture& Picture::crop_to(const Mask::Rect& mask)
  {
    m_data.crop_to(mask);
    return *this;
  }

  Picture& Picture::border(const Color& color, std::string len)
  {
    m_page.border(color, len);
    return *this;
  }

  Picture& Picture::border(const Pen& pen)
  {
    m_page.border(pen);
    return *this;
  }

  Picture& Picture::backing(const Color& color)
  {
    m_page.backing(color);
    return *this;
  }

  Picture& Picture::operator<< (const Color& col)
  {
    m_data << col;
    return *this;
  }

  Picture& Picture::operator<< (const Tile& T)
  {
    m_data << T;
    return *this;
  }

  Screen& Picture::screen(void)
  {
    return m_data;
    //    Screen& val = m_data; return val;
  }

  void Picture::print_to(const char* filename, const Format& fmt)
  {
#ifdef DEBUG
    std::cerr << "Picture::print_to" << std::endl;
#endif
    std::ofstream outfile(filename);

    if (!outfile.is_open())
      {
	// TO DO: More sophisticated handling...
	std::stringstream err_msg;
	err_msg << "Cannot open file \"" << filename << "\"";
	epix_warning(err_msg.str());
	return;
      }

    m_page.paste(m_data);

    outfile << fmt.pic_header(m_page.m_mask->bl(), m_page.m_mask->tr(),
			      m_width, m_page.m_pallet);

    // underlayer
    if (!m_page.m_backing->is_empty())
      outfile << m_page.m_backing->print_to(fmt);

    std::vector<Tile*>::const_iterator curr;
    for (curr = m_page.m_tiles.begin(); curr != m_page.m_tiles.end(); ++curr)
      {
	(*curr)->realign(m_width.units());
	outfile << (*curr)->print_to(fmt);
      }

    if (!m_page.m_border->is_empty())
      outfile << m_page.m_border->print_to(fmt);

    outfile << fmt.pic_footer();
    outfile.close();
  } // end of print_to()
} // end of namespace
