/* 
 *  Polygon.cc -- epix2::Polygon and implementation classes
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 10, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef DEBUG
#include "Debug.h"
#endif

#include <cmath>

#include "Functions.h"
#include "Point.h"
#include "Vector.h"

#include "Shard_2d.h"

#include "Polygon.h"

namespace ePiX2 {

  // Triangle with specified vertices
  Polygon::Polygon(const Point& v1, const Point& v2, const Point& v3)
  {
    Point ctr = v1 + (1.0/3)*((v2-v1) + (v3-v1));
    Vector N=(v1-ctr)*(v2-ctr);

    // TO DO: Check for collinearity, use segment if appropriate
    if (norm(N) > EPIX2_EPSILON)
      N.normalize();

    Plate3d data(N);

    data.add(Edge3d(v1, v2, true));
    data.add(Edge3d(v2, v3, true));
    data.add(Edge3d(v3, v1, true));

    data.update();
    add_shard(data);
  }

  // regular n-gon in the plane spanned by v1, v2, with a vertex at v1
  Polygon::Polygon(const Vector& v1, const Vector& v2, const unsigned int n)
  {
    Point ctr = v1.tail();
    Vector e2 = v2;
    e2.move_to(ctr);
    e2 %= v1; // orthogonalize
    e2.normalize(); // TO DO: may throw...
    e2 *= norm(v1);

    Vector N=v1*e2;

    // TO DO: Check for collinearity, return segment if appropriate
    if (norm(N) > EPIX2_EPSILON)
      N.normalize();

    Plate3d data(N);

    double th = 2*M_PI/n;
    double d_th = th;
    Point curr = ctr+v1, next = ctr + cos(th)*v1 + sin(th)*e2;

    for (unsigned int i=0; i<n; ++i)
      {
	data.add(Edge3d(curr, next, true));
	curr = next;
	th += d_th;
	next = ctr + cos(th)*v1 + sin(th)*e2;
      }

    data.update();
    add_shard(data);
  }

  Polygon* Polygon::clone(void) const
  {
#ifdef DEBUG
    std::cerr << "Polygon::clone()" << std::endl;
#endif
    return new Polygon(*this);
  }
} // end of namespace

