/*
 * SVG.cc -- ePiX2 SVG output format
 *
 * This file is part of ePiX, a program for creating high-quality
 * figures in LaTeX
 *
 * Version 2.0pre
 *
 * Last Change: September 5, 2006
 */

/*
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 */
 
/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include <string>
#include <sstream>
#include <set>
#include <vector>

#include "Constants.h"
#include "Functions.h"

#include "Edge_Data.h"

#include "Color.h"

#include "Styles.h"
#include "Utils.h"

#include "SVG.h"

namespace ePiX2 {

  // formatting string to break lines in path data lists
  std::string SVG_path_break_init()
  {
    std::stringstream buf;
    buf << std::endl << "  "; // newline, two spaces

    return buf.str();
  }

  const std::string SVG_path_break(SVG_path_break_init());

  unsigned int d_to_uint(const double arg)
  {
    return (unsigned int) floor(arg*0XFF);
  }

  SVG::SVG() { }

  // private functions
  std::string SVG::update_fill_style(const Color& face,
				     const Color& edge,
				     const Length& edge_width) const
  {
    return ""; // we contain no state...
  }

  std::string SVG::update_line_style(const Color& line,
				     const Length& line_width,
				     const Color& base) const
  {
    return "";
  }


  // draw visible portion(s) of <edges> in <pen>
  std::string SVG::draw_outline_data(const std::vector<Edge2d>& edges,
				     const Pen2d& pen) const
  {
    std::stringstream obuf;

    std::vector<Edge2d>::const_iterator p;

    bool in_path(false); // current path started?

    for (p=edges.begin(); p!=edges.end(); ++p)
      {
	if ((*p).seen())
	  {
	    if (!in_path) // start path
	      {
		obuf << "<path d=\"M" << print((*p).tail()) << " ";

		in_path = true;
	      }

	    obuf << "L" << print((*p).head());
	  }

	else // not seen
	  {
	    if (in_path) // end path segment
	      obuf << "\"" << std::endl
		   << "fill=\"none\" "
		   << "stroke=\"" << print(pen.color()) << "\" "
		   << "stroke-width=\"" << print(pen.width()) << "\"/>"
		   << std::endl;

	    in_path = false;
	  }
      } // all edges re-examined

    if (in_path) // last edge visible
      obuf << "\"" << std::endl
	   << "fill=\"none\" "
	   << "stroke=\"" << print(pen.color()) << "\" "
	   << "stroke-width=\"" << print(pen.width()) << "\"/>"
	   << std::endl;

    //    return break_lines(obuf.str(), SVG_path_break);
    return obuf.str();
  } // end of draw_outline_data()


  // Mid-level functions
  // TO DO
  std::string  SVG::print_text(const Pair& here, const Pair& here_offset,
			       const std::string& label,
			       const mark_type mark,
			       const Color& text, const Color& back,
			       const Length& pad, const Pen2d& edge,
			       const align_type align,
			       const font_type font,
			       const double angle) const
  {
    return "";
    /*
    std::stringstream obuf, cmd;

    if (!back.is_unset())
      {
	// collect style command, update label_pad (framesep) if nec.
	cmd << "[" << "fillcolor=" << print(back);

	if (pad != m_text_pad)
	  {
	    m_text_pad = pad;
	    // print straight to output stream
	    obuf << "\\psset{framesep=" << print(pad) << "}" << std::endl;
	  }

	if (edge.color() != m_edge) // set linecolor
	  {
	    if (edge.color().is_unset())
	      cmd << ",linestyle=none";

	    else
	      cmd << ",linecolor=" << print(edge.color());
	  }

	if (edge.width() != m_edge_width) // set linewidth
	  cmd << ",linewidth=" << print(edge.width());

	cmd << "]";
      }
    // N.B. No label decorations without background color
    else
      cmd << "[fillstyle=none,linestyle=none]";

    // cmd complete

    obuf << "\\rput";

    if (align != c) // not centered
      obuf << "[" << print(align) << "]";

    if (fabs(angle) > EPIX2_EPSILON) // not vertical
      obuf << "{" << angle << "}";

    // text label gets offset
    obuf << print(here_offset) << "{";

    if (!back.is_unset()) // we have a background color
      obuf << "\\psframebox" << cmd.str();

    obuf << "{";

    // label text proper
    obuf << "\\color{" << print(text) << "}";

    if (font != ROMAN)
      obuf << print(font) << "{";

    obuf << label;

    if (font != ROMAN)
      obuf << "}";

    obuf << "}}" << std::endl;

    // print mark; not offset
    if (mark != MK_NONE)
      obuf << "\\rput" << print(here) << "{"
	   << "\\color{" << print(text) << "}"
	   << print(mark) << "}" << std::endl;

    return obuf.str();
    */
  } // end of SVG::print_text()


  /* * * print_plate * * */
  // fill color; border; clipped flag; edges
  std::string SVG::print_plate(const Color& fc, const Pen2d& line,
			       bool cropped,
			       const std::vector<Edge2d>& edges) const
  {
    std::stringstream value, obuf; // obuf gets lines broken

    // send straight to output (no line breaking)
    value << comment("Plate");

    // break long line
    obuf << "<path d=\"";

    std::vector<Edge2d>::const_iterator p(edges.begin());

    const Pair init((*p).tail()); // store first vertex

    obuf << "M" << print(init);

    // on first pass, check for visible edges
    bool some_edge_drawn(false);

    for (p=edges.begin(); p!=edges.end(); ++p)
      {
	if ((*p).seen())
	  some_edge_drawn=true;

	obuf << "L" << print((*p).head());
      }

    obuf << "z\"";  // close path

    value << break_lines(obuf.str(), SVG_path_break) << std::endl;


    // set fill color
    if (fc.is_unset())
      value << "fill=\"none\" ";

    else
      value << "fill=\"" << print(fc) << "\" ";

    // set line color
    if (cropped)
      value << "stroke=\"none\""; // re-draw boundary below

    else
      value << "stroke=\"" << print(line.color()) << "\" "
	    << "stroke-width=\"" << print(line.width()) << "\"";

    // close stanza
    value << "/>" << std::endl;

    // If not cropped, we're done. Otherwise, re-draw border if necessary.
    if (some_edge_drawn && cropped && Length(0) < line.width())
      value << comment("Plate boundary")
	    << draw_outline_data(edges, line);

    return value.str();
  } // end of SVG::print_plate()


  /* * * print_outline * * */
  // line/base; edges
  std::string SVG::print_outline(const Pen2d& line, const Pen2d& base, 
				 const std::vector<Edge2d>& edges) const
  {
    const Length lw(line.width());
    const Color  lc(line.color());

    if ( (lw < Length(EPIX2_EPSILON)) || (lc.is_unset()))
      return "";

    // else
    std::stringstream value;

    const Color  bc(base.color());
    const Length bw(base.width());

    value << comment("Outline");

    // base is wider
    if (lw < bw)
      value << draw_outline_data(edges, base);

    // line has different color?
    if ((lc != bc) || !(lw < bw))
      value << draw_outline_data(edges, line);

    return value.str();

  } // end of print_outline(Pen2d, Pen2d, std::vector<Edge2d>)


  std::string SVG::print_outline(const Pen2d& line, const Pen2d& base, 
				 const Edge2d& e) const
  {
    std::vector<Edge2d> edges;
    edges.push_back(e);

    print_outline(line, base, edges);
  } // end of print_outline(Pen2d, Pen2d, Edge)


  std::string SVG::file_header(void) const
  {
    std::stringstream obuf;
    using std::endl;

    obuf << "<?xml version=\"1.0\" standalone=\"yes\"?>" << endl
	 << "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\"" << endl
	 << "\"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">" << endl;

    return obuf.str();
  }

  std::string SVG::pic_header(const Pair& sw, const Pair& ne, 
			      const Length& len,
			      const std::set<Color>& pallet) const
  {
    using std::endl;
    std::stringstream obuf;

    // TO DO: Offsets
    obuf << file_header()
	 << "<svg width=\"" << print(len)
	 << "\" height=\""
	 << (ne.x2() - sw.x2()) << len.units()
	 << "\"  viewBox=\"" << sw.x1() << " " << sw.x2() << " "
	 << ne.x1() << " " << ne.x2() << "\"" << endl
	 << "xmlns=\"http://www.w3.org/2000/svg\" version=\"1.1\">"
	 << endl;

    return obuf.str();
  }

  std::string SVG::pic_footer(void) const
  {
    std::stringstream obuf;
    obuf << "</svg>" << std::endl;
    return obuf.str();
  }

  // RGB or CMY
  std::string SVG::print_color(const std::string& model,
			       const std::string& name,
			       const double d1, const double d2,
			       const double d3) const
  {
    std::stringstream obuf;
    obuf << "rgb(";

    if (model == "cmy")
      obuf << d_to_uint(1-d1) << ", "
	   << d_to_uint(1-d2) << ", "
	   << d_to_uint(1-d3);

    else
      obuf << d_to_uint(d1) << ", "
	   << d_to_uint(d2) << ", "
	   << d_to_uint(d3);

    obuf << ")";

    return obuf.str();
  }

  // print CMYK color
  std::string SVG::print_color(const std::string& model,
			       const std::string& name,
			       const double d1, const double d2,
			       const double d3, const double d4) const
  {
    std::stringstream obuf;
    obuf << "rgb("
	 << d_to_uint(1-d1-d4) << ", "
	 << d_to_uint(1-d2-d4) << ", "
	 << d_to_uint(1-d3-d4) << ")";

    return obuf.str();
  }


  std::string SVG::comment(const std::string& msg) const
  {
    std::stringstream obuf;
    obuf << "<!-- " << msg << " -->" << std::endl;

    return obuf.str();
  }


  // TO DO
  std::string SVG::print(const align_type arg) const
  {
    return "";
    /*
    std::string val;
    switch(arg) 
      {
      case B:
	val="B";
	break;
      case Br:
	val="Br";
	break;
      case Bl:
	val="Bl";
	break;
      case c:
	val="c";
	break;
      case r:
	val="l"; // opposite of LaTeX
	break;
      case tr:
      case rt:
	val="bl";
	break;
      case t:
	val="b";
	break;
      case tl:
      case lt:
	val="br";
	break;
      case l:
	val="r";
	break;
      case bl:
      case lb:
	val="tr";
	break;
      case b:
	val="t";
	break;
      case br:
      case rb:
	val="tl";
	break;
      }

    return val;
    */
  } // end of SVG::print(align_type)

  // TO DO
  std::string SVG::print(const mark_type mark) const
  {
    return "";
    /*
    std::string val="$";

    switch(mark)
      {
      case MK_NONE:
	val = ""; // N.B. reassign, not append
	break;
      case MK_CIRC:
	val += "\\circle";
	break;
      case MK_SPOT:
	val += "\\circle*";
	break;
      case MK_RING:
	val += "\\circle";
	break;
      case MK_DOT:
	val += "\\circle*";
	break;
      case MK_DDOT:
	val += "\\circle*";
	break;
      case MK_PLUS:
	val += "+";
	break;
      case MK_OPLUS:
	val += "\\oplus";
	break;
      case MK_TIMES:
	val += "\\times";
	break;
      case MK_OTIMES:
	val += "\\otimes";
	break;
      case MK_DIAMOND:
	val += "\\diamond";
	break;
      case MK_UP:
	val += "\\bigtriangleup";
	break;
      case MK_DOWN:
	val += "\\bigtriangledown";
	break;
      case MK_BOX:
	val += epix2_rule(2, 2); // Magic numbers
	break;
      case MK_BBOX:
	val += epix2_rule(1.5, 1.5);
	break;
      case MK_HTICK:
	val += epix2_rule(4, 0.5);
	break;
      case MK_VTICK:
	val += epix2_rule(0.5, 4);
	break;
      }

    if (mark != MK_NONE)
      val +="$";
    return val;
    */
  } // end of SVG::print(mark_type)


  // TO DO
  std::string SVG::print(const font_type f) const
  {
    return "";
    /*
    std::string val = "\\";

    switch(f)
      {
      case ROMAN:
	val = ""; // N.B. reassign
	break;
      case ITALIC:
	val += "textit";
	break;
      case BFACE:
	val += "textbf";
	break;
      case TTYPE:
	val += "texttt";
	break;
      }

    return val;
    */
  } // end of SVG::print(font_type)


  std::string SVG::print(const Pair& arg) const
  {
    std::stringstream o;
    o << ePiX2::trunc(arg.x1()) << " " << ePiX2::trunc(arg.x2()) << " ";
    return o.str();
  }

  std::string SVG::print(const Color& col) const
  {
    return col.print_to(*this);
  }

  std::string SVG::print(const Length& len) const
  {
    std::stringstream o;
    o << len.magnitude() << len.units();
    return o.str();
  }

} /* end of namespace */
