/* 
 * Screen.cc -- ePiX2::Screen class and classes for importing
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 21, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include <vector>
#include <algorithm>

#include "Errors.h"

#include "Pair.h"
#include "Pair_Map.h"
#include "Mask.h"

#include "Color.h"
#include "Styles.h"
#include "Tile.h"

#include "Tile_1d.h"
#include "Tile_2d.h"

#include "Screen.h"

namespace ePiX2 {

  typedef std::vector<Tile*>::const_iterator Tp;

  Screen::Screen(const Pair arg1, const Pair arg2)
    : m_mask(new Mask::Rect(arg1, arg2)),
      m_border(new Outline2d(Pen2d::Xfine, Pen2d::Xfine, false)),
      m_backing(new Plate2d(Neutral(), Pen2d::Xfine, false))
  {
#ifdef DEBUG
    std::cerr << "Screen(Pair, Pair)" << std::endl;
#endif
  }

  Screen::Screen(const Mask::Mask& bbox)
    : m_mask(bbox.clone()),
      m_border(new Outline2d(Pen2d::Xfine, Pen2d::Xfine, false)),
      m_backing(new Plate2d(Neutral(), Pen2d::Xfine, false))
  {
#ifdef DEBUG
    std::cerr << "Screen(Mask::Mask)" << std::endl;
#endif
  }

  Screen::Screen(const Screen& scr)
    : m_mask(scr.m_mask->clone()),
      m_border(new Outline2d(Pen2d::Xfine, Pen2d::Xfine, false)),
      m_backing(new Plate2d(Neutral(), Pen2d::Xfine, false)),
      m_pallet(scr.m_pallet)
  {
#ifdef DEBUG
    std::cerr << "Screen(const Screen&) " << std::endl;
#endif
    if (!scr.m_backing->is_empty())
      {
	delete m_backing;
	m_backing = scr.m_backing->clone();
      }

    for(Tp p=scr.m_tiles.begin(); p!= scr.m_tiles.end(); ++p)
      (*this) << *((*p)->clone());
  }

  Screen& Screen::operator= (const Screen& scr)
  {
#ifdef DEBUG
    std::cerr << "Screen::operator=" << std::endl;
#endif
    if (this != &scr)
      {
	// copy scr's tiles before freeing ours
	std::vector<Tile*> tmp;
	for(Tp p=scr.m_tiles.begin(); p!= scr.m_tiles.end(); ++p)
	  (*this) << *((*p)->clone());

	Mask::Mask* tmp_mask(scr.m_mask->clone());

	Outline2d* tmp_border(scr.m_border->clone());
	Plate2d* tmp_backing(scr.m_backing->clone());

	// okay, safe to muck with our data
	delete m_mask;
	m_mask = tmp_mask;

	delete m_border;
	m_border  = tmp_border;

	delete m_backing;
	m_backing = tmp_backing;

	m_pallet  = scr.m_pallet;

	for (Tp p=m_tiles.begin(); p!=m_tiles.end(); ++p)
	  delete *p;

	swap(m_tiles, tmp);
      }
    return *this;
  }



  Screen::~Screen()
  {
#ifdef DEBUG
    std::cerr << "~Screen()" << std::endl;
#endif

    delete m_mask;
    delete m_border;
    delete m_backing;

    for(Tp p=m_tiles.begin(); p!= m_tiles.end(); ++p)
      delete *p;

#ifdef DEBUG
    std::cerr << "  end of ~Screen()" << std::endl;
#endif
  }

  /*
  // Should be unnecessary; each "add_tile" call imports colors
  void Screen::import_pallet(const Screen& child)
  {
    std::set<Color>::const_iterator s;
    for(s=child.m_pallet.begin(); s!=child.m_pallet.end(); ++s)
      m_pallet.insert(*s);
  }
  */

  // superimpose a layer by affine scaling
  Screen& Screen::paste(const Screen& child)
  {
#ifdef DEBUG
    std::cerr << "Screen::paste(child)" << std::endl;
#endif
    if (this == &child)
      epix_warning("Pasting a Screen to itself has no effect");

    else
      {
	Pair_Map affine(Affine(*m_mask, *(child.m_mask)));

	if (!child.m_border->is_empty())
	  {
	    // child is const, so must copy tile
	    Outline2d* temp(child.m_border->clone());

	    // operator<< clones mapped Tile
	    (*this) << temp->map_by(affine);

	    delete temp;
	  }

	if (!child.m_backing->is_empty())
	  {
	    // child is const, so must copy tile
	    Plate2d* temp(child.m_backing->clone());

	    // operator<< clones mapped Tile
	    (*this) << temp->map_by(affine);

	    delete temp;
	  }

	// import tiles
	for(Tp p=child.m_tiles.begin(); p!= child.m_tiles.end(); ++p)
	  {
	    // same comments as above
	    Tile* temp((*p)->clone());
	    (*this) << temp->map_by(affine);
	    delete temp;
	  }
      }
    return *this;
  }

  // map child to inset
  Screen& Screen::paste(const Screen& child, const Mask::Rect& inset)
  {
#ifdef DEBUG
    std::cerr << "Screen::paste(child, inset)" << std::endl;
#endif
    if (this == &child)
      epix_warning("Pasting a Screen to itself has no effect");

    else
      {
	Pair_Map affine(Affine(inset, *(child.m_mask)));

	if (!child.m_border->is_empty())
	  {
	    // child is const, so must copy tile
	    Outline2d* temp(child.m_border->clone());

	    // operator<< clones mapped Tile
	    (*this) << temp->map_by(affine);

	    delete temp;
	  }

	if (!child.m_backing->is_empty())
	  {
	    Plate2d* temp(child.m_backing->clone());
	    (*this) << temp->map_by(affine);
	    delete temp;
	  }

	for(Tp p=child.m_tiles.begin(); p!= child.m_tiles.end(); ++p)
	  {
	    Tile* temp((*p)->clone());
	    (*this) << temp->map_by(affine);
	    delete temp;
	  }
      }
    return *this;
  }

  /*
  // map child.cut to inset
  ////////  N.B. This function has not been kept up-to-date!!
  Screen& Screen::paste(const Screen& child,
			const Mask::Rect& cutout, const Mask::Rect& inset)
  {
    if (this == &child)
      epix_warning("Pasting a Screen to itself has no effect");

    else
      {
	Pair_Map affine=Affine(inset, cutout);

	for(Tp p=child.m_tiles.begin(); p!= child.m_tiles.end(); ++p)
	  {
	    Tile* temp((*p)->clone());

	    temp->crop_to(cutout);
	    if (!temp->is_empty())
	      (*this) << temp->map_by(affine);

	    delete temp;
	  }
      }
    return *this;
  }
  */

  // or by a specified Pair_Map
  Screen& Screen::paste(const Screen& child, const Pair_Map& f)
  {
#ifdef DEBUG
    std::cerr << "Screen::paste(child, f)" << std::endl;
#endif
    if (this == &child)
      epix_warning("Pasting a Screen to itself has no effect");

    else
      {
	if (!child.m_border->is_empty())
	  {
	    Tile* temp(child.m_border->clone());
	    (*this) << temp->map_by(f);
	    delete temp;
	  }

	if (!child.m_backing->is_empty())
	  {
	    Tile* temp(child.m_backing->clone());
	    (*this) << temp->map_by(f);
	    delete temp;
	  }

	for(Tp p=child.m_tiles.begin(); p!= child.m_tiles.end(); ++p)
	  {
	    Tile* temp((*p)->clone());
	    (*this) << temp->map_by(f);
	    delete temp;
	  }
      }
    return *this;
  }

  /*
  //////// N.B. This function has not been kept up-to-date
  Screen& Screen::paste(const Screen& child,
			const Mask::Rect& cutout, const Pair_Map& f)
  {
    if (this == &child)
      epix_warning("Pasting a Screen to itself has no effect");

    else
      {
	for(Tp p=child.m_tiles.begin(); p!= child.m_tiles.end(); ++p)
	  {
	    Tile* temp((*p)->clone());
	    temp->crop_to(cutout);

	    if (!temp->is_empty())
	      (*this) << temp->map_by(f);

	    delete temp;
	  }
      }
    return *this;
  }
  */

  Screen& Screen::clear(void)
  {
#ifdef DEBUG
    std::cerr << "Screen::clear()...";
#endif
    delete m_border;
    m_border = new Outline2d(Pen2d::Xfine, Pen2d::Xfine, false);

    delete m_backing;
    m_backing = new Plate2d(Neutral(), Pen2d::Xfine, false);

    m_tiles.clear();

#ifdef DEBUG
    std::cerr << "done" << std::endl;
#endif
    return *this;
  }

  Screen& Screen::crop(void)
  {
    std::vector<Tile*> visibles;

    // no need to crop border or backing, since they came from m_mask
    for (Tp curr=m_tiles.begin(); curr!=m_tiles.end(); ++curr)
      {
	(*curr)->crop_to(*m_mask);

	if (!(*curr)->is_empty())
	  visibles.push_back(*curr);
      }

    swap(m_tiles, visibles);
    return *this;
  }

  Screen& Screen::crop_to(const Mask::Mask& mask)
  {
    std::vector<Tile*> visibles;

    // TO DO: re-create border/backing form mask; need to get attributes...
    if (!m_border->is_empty())
      m_border->crop_to(mask);

    if (!m_backing->is_empty())
      m_backing->crop_to(mask);

    for (Tp curr=m_tiles.begin(); curr!=m_tiles.end(); ++curr)
      {
	(*curr)->crop_to(mask);

	if (!(*curr)->is_empty())
	  visibles.push_back(*curr);
      }

    swap(m_tiles, visibles);
    return *this;
  }

  Screen& Screen::border(const Color& color, std::string len)
  {
#ifdef DEBUG
    std::cerr << "Screen::border(" << color.name() << ")" << std::endl;
#endif
    delete m_border;
    m_border = new Outline2d(m_mask->border(color, Length(len)));

    (*this) << color; // add color to pallet

    return *this;
  }


  Screen& Screen::border(const Pen& pen)
  {
#ifdef DEBUG
    std::cerr << "Screen::border(Pen)" << std::endl;
#endif
    delete m_border;
    m_border = new Outline2d(m_mask->border(pen));

    (*this) << pen.color(); // add color to pallet

    return *this;
  }



  Screen& Screen::backing(const Color& color)
  {
#ifdef DEBUG
    std::cerr << "Screen::backing(" << color.name() << ")" << std::endl;
#endif
    delete m_backing;

    m_backing = new Plate2d(m_mask->backing(color));

    (*this) << color; // add color to pallet

    return *this;
  }


  Screen& Screen::operator<< (const Color& col)
  {
    m_pallet.insert(col);
    return *this;
  }

  Screen& Screen::operator<< (const Tile& T)
  {
    m_tiles.push_back(T.clone());
    T.add_colors_to(*this);
    return *this;
  }
} /* end of namespace */
