/* 
 *  Shard.h -- ePiX2 class for Object representation
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*
 * This file provides the Shard interface, the internal representation
 * data for Objects. Shards can be 0-, 1-, or 2-dimensional (Glyph,
 * Outline, or Plate), and possess geometric data (location, list of
 * vertices, normal vector, etc.) as well as attributes (color, gloss,
 * transparency, line styles, font, etc.)
 */

#ifndef EPIX2_SHARD
#define EPIX2_SHARD

#include <vector>

#include "Constants.h"

namespace ePiX2 {
  class Point;
  class Vector;
  class Gyro;

  class Color;
  class Fog;

  class Camera;
  class Knife;
  class Light;

  class Color;
  class Length;
  class Ink;
  class Pen;
  class Stencil;

  class Screen;
  class Layer;

  class Shard {
  public:
    virtual ~Shard(void) { }

    virtual Shard* clone(void) const = 0;
    // virtual bool operator== (const Shard&) const = 0;

    virtual bool is_empty() const = 0;

    // half-space and planar cuts
    virtual Shard*  chop(const Knife&, std::vector<Shard*>&) = 0;
    virtual Shard* slice(const Knife&, std::vector<Shard*>&) = 0;

    // set attributes (no-ops by default; children re-define as needed)
    virtual Shard& front_pen(const Pen&);
    virtual Shard& front_ink(const Ink&);
    virtual Shard& back_pen(const Pen&);
    virtual Shard& back_ink(const Ink&);

    virtual Shard& line_pen(const Pen&);
    virtual Shard& base_pen(const Pen&);
    virtual Shard& stencil(const Stencil&);

    virtual Shard& text(const Color&);
    virtual Shard& back(const Color&);
    virtual Shard&  pad(const Length&);
    virtual Shard& edge(const Pen&);

    virtual Shard& align(const align_type);
    virtual Shard&  font(const font_type);
    virtual Shard& angle(const double);
    virtual Shard& offset(std::string); // need to pass by value
    virtual Shard& fix() = 0; // make ink/pen/stencil colors permanent

    // photograph to specified Screen
    virtual void shoot(Screen&, const Camera&, const Color& ambient,
		       const std::vector<Light>& spots, const Fog&,
		       const photo_type) const = 0;

    // For sorting before adding to BSP.
    // Returns area (2-dim), (# segments)*EPSILON (1-dim), or 0 (0-dim)
    virtual double priority(void) const = 0;

    virtual void add_to(Layer*) const = 0;

    // Express in specified coordinate basis and scale
    virtual Shard* in_coords(const Gyro&) = 0;

    // We occlude X from vpt?
    virtual bool hides(const Point& vpt, const Point& X) const = 0;

    // Our location, for reflectivity calculation
    virtual Point here(void) const = 0;
    // Unit normal vector
    virtual Vector perp(void) const = 0;

  }; // end of class Shard

  // Sort by priority for addition to BSP tree
  class Shard_Area {
  public:
    bool operator() (const Shard* S1, const Shard* S2);
  }; // end of class Shard_Area

} /* end of namespace */

#endif /* EPIX2_OBJECT_REP */
