/* 
 * Shard_0d.cc -- ePiX2::Label and Marker classes
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include <sstream>

#include "Functions.h"

#include "Point.h"
#include "Vector.h"
#include "Gyro.h"

#include "Length.h"

#include "Color.h"
#include "Fog.h"

#include "Knife.h"
#include "Styles.h"

#include "Camera.h"
#include "Screen.h"

#include "Layer.h"

#include "Tile_0d.h"
#include "Shard_0d.h"

namespace ePiX2 {

  Glyph3d::Glyph3d(const Vector& perp, const Length& hoff, const Length& voff,
		   const std::string& label, const mark_type mark,
		   const Stencil& stencil)
    : m_perp(perp), m_hoff(hoff), m_voff(voff),
      m_label(label), m_mark(mark), m_stencil(stencil), m_seen(true)
  {
#ifdef DEBUG
    Point loc(here());
    std::cerr << "Glyph3d() " << label << " at ("
	      << loc.x1() << "," << loc.x2() << "," << loc.x3()
	      << ")" << std::endl;
#endif
  }

  Glyph3d* Glyph3d::clone(void) const
  {
#ifdef DEBUG
    Point loc3(m_perp.tail());
    std::cerr << "Glyph3d::clone() creating "
	      << "Glyph3d at ("
	      << loc3.x1() << "," << loc3.x2() << "," << loc3.x3() 
	      << ")" << std::endl;
#endif
    return new Glyph3d(*this);
  }

  // half-space cut
  Glyph3d*  Glyph3d::chop(const Knife& k, std::vector<Shard*>&)
  {
    if (k.m_blade.height(here()) < 0)
      m_seen=false;

    return this;
  }

  Glyph3d* Glyph3d::slice(const Knife& k, std::vector<Shard*>&)
  {
    if (fabs(k.m_blade.height(here())) > EPIX2_EPSILON)
      m_seen=false;

    return this;
  }

  Glyph3d& Glyph3d::stencil(const Stencil& b)
  {
    m_stencil.apply(b);
    return *this;
  }

  Glyph3d& Glyph3d::text(const Color& col)
  {
    m_stencil.text(col);
    return *this;
  }

  Glyph3d& Glyph3d::back(const Color& col)
  {
    m_stencil.back(col);
    return *this;
  }

  Glyph3d&  Glyph3d::pad(const Length& len)
  {
    m_stencil.pad(len);
    return *this;
  }

  Glyph3d& Glyph3d::edge(const Pen& pen)
  {
    m_stencil.edge(pen);
    return *this;
  }


  Glyph3d& Glyph3d::align(const align_type a)
  {
    m_stencil.align(a);
    return *this;
  }

  Glyph3d&  Glyph3d::font(const font_type F)
  {
    m_stencil.font(F);
    return *this;
  }

  Glyph3d& Glyph3d::angle(const double th)
  {
    m_stencil.angle(th);
    return *this;
  }


  Glyph3d& Glyph3d::offset(std::string arg)
  {
    get_lengths(arg, m_hoff, m_voff); // parse arg into lengths
    return *this;
  }

  Glyph3d& Glyph3d::fix()
  {
    m_stencil.fix();

    return *this;
  }

  // photograph to specified Screen
  void Glyph3d::shoot(Screen& S, const Camera& cam, const Color& ambient,
		      const std::vector<Light>& no_op, const Fog& fog,
		      photo_type T) const
  {
#ifdef DEBUG
    std::cerr << "Glyph3d::shoot()" << std::endl;
#endif

    if (!m_seen)
      return;

    // else
    Vector to_cam = cam.viewpt() - here();
    double dist_cam = norm(to_cam);
    to_cam *= 1.0/dist_cam;

    // const double cos_th = (to_cam|m_perp); // cosine of face normal

    // convert to Color and Pen2d
    Color my_text(cam(fog.mutes(m_stencil.text(), dist_cam)));
    Color my_back(cam(fog.mutes(m_stencil.back(), dist_cam)));
    Color my_edge(cam(fog.mutes(m_stencil.edge().color(), dist_cam)));

    // TO DO: Appropriate shading...?
    /*
    my_text *= cos_th; // face normal shading
    my_back *= cos_th; // face normal shading
    my_edge *= cos_th; // face normal shading
    */

    // apply fog only
    Pen2d edge(my_edge, m_stencil.edge().width());

    // add to screen
    S << Glyph2d(cam(here()), m_hoff, m_voff, m_label, m_mark, 
		 my_text, my_back, m_stencil.m_pad, edge,
		 m_stencil.m_align, m_stencil.m_font,
		 m_stencil.m_angle, m_seen);
  } // end of Glyph3d::shoot()

  bool Glyph3d::is_empty(void) const
  {
    return !m_seen;
  }

  Glyph3d* Glyph3d::in_coords(const Gyro& gyro)
  {
    m_perp = gyro.in_coords(m_perp);
    return this;
  }

  // place in BSP
  void Glyph3d::add_to(Layer* T) const
  {
#ifdef DEBUG
    std::cerr << "Glyph3d::add_to(Layer*)" << std::endl;
#endif
    if (is_empty())
      return;

    //else
    double ht((T->perp()).height(here()));

    if (fabs(ht) < EPIX2_EPSILON)  // morally coplanar; adjust m_perp
      {
	Point here(m_perp.tail()); // save location
	m_perp = T->perp();        // get new perp
	m_perp.move_to(here);      // move to old location
	T->add(this);
      }

    else if (ht <= -EPIX2_EPSILON) // add to back
      T->add_b(this);

    else //  if (min_ht >= EPIX2_EPSILON)
      T->add_f(this);
  }

  Point Glyph3d::here(void) const
  {
    return m_perp.tail();
  }

  // Unit normal vector
  Vector Glyph3d::perp(void) const
  {
    return m_perp;
  }
} // end of namespace

