/* 
 * Shard_1d.cc -- epix2::Outline3d (planar path) class functions
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef DEBUG
#include "Debug.h"
#endif

#include <cmath>
#include <sstream>


#include "Errors.h"
#include "Functions.h"

#include "Edge_Data.h"

#include "Point.h"
#include "Vector.h"

#include "Color.h"
#include "Fog.h"

#include "Gyro.h"
#include "Knife.h"

#include "Styles.h"
#include "Lighting.h"

#include "Camera.h"
#include "Screen.h"

#include "Shard.h"
#include "Layer.h"

#include "Tile_1d.h"

#include "Shard_0d.h"
#include "Shard_1d.h"

namespace ePiX2 {

  typedef std::vector<Edge3d>::const_iterator Ep3;

  // build: perp vector, fill, edge, all edges drawn
  Outline3d::Outline3d(const Pen& line, const Pen& base, bool cl,
		       const Vector& perp)
    : m_line(line), m_base(base), m_clipped(cl), m_perp(perp),
      m_area(0), m_updated(false)
  {
#ifdef DEBUG
    std::cerr << "Outline3d()" << std::endl;
#endif
  }

  Outline3d::Outline3d(const Vector& perp)
    : m_line(Pen()), m_base(Pen::Xfine), m_clipped(false), m_perp(perp),
      m_area(0), m_updated(false)
  {
#ifdef DEBUG
    std::cerr << "Outline3d()" << std::endl;
#endif
  }

  Outline3d* Outline3d::clone() const
  {
#ifdef DEBUG
    std::cerr << "Outline3d::clone()" << std::endl;
#endif
    return new Outline3d(*this);
  }

  Outline3d::~Outline3d()
  {
#ifdef DEBUG
    std::cerr << "  ~Outline3d" << std::endl;
#endif
  }

  Outline3d& Outline3d::line_pen(const Pen& b)
  {
#ifdef DEBUG
    std::cerr << "Outline3d::line_pen()" << std::endl;
#endif
    m_line.apply(b);
    return *this;
  }

  Outline3d& Outline3d::base_pen(const Pen& b)
  {
#ifdef DEBUG
    std::cerr << "Outline3d::base_pen()" << std::endl;
#endif
    m_base.apply(b);
    return *this;
  }

  Outline3d& Outline3d::fix()
  {
#ifdef DEBUG
    std::cerr << "Outline3d::fix()" << std::endl;
#endif
    m_line.fix();
    m_base.fix();
    return *this;
  }

  // half-space cut; collect new Shards in cuts
  Outline3d* Outline3d::chop(const Knife& knife, std::vector<Shard*>& cuts)
  {
#ifdef DEBUG
    std::cerr << "Outline3d::chop()" << std::endl;
#endif
    if (is_empty())
      {
	update();
	return this;
      }

    // else chop edges
    const double eps(EPIX2_EPSILON); // half-thickness of cutting plane

    const Vector blade(knife.m_blade); // points toward retained part
    const Pen pen(knife.m_pen);

    std::vector<Edge3d> bord;

    std::vector<Point> cut_ends; // store cut ends

    // Examine heights of edges serially, check for crossings.
    // For clipping, a "crossing" occurs when max height > eps and 
    // min height < eps.
    Point T_c, H_c; // current head, tail
    double T_ht, H_ht, max_ht, min_ht;

    for (Ep3 curr=m_border.begin(); curr!=m_border.end(); ++curr)
      {
	T_c = (*curr).tail();
	H_c = (*curr).head();

	T_ht = blade.height(T_c);
	H_ht = blade.height(H_c);

	max_ht = max(T_ht, H_ht);
	min_ht = min(T_ht, H_ht);

        if (min_ht > eps) // no crossing
          bord.push_back(*curr);

        else if ( -eps <= min_ht && max_ht <= eps) // edge lies in cut plane
	  {
	    bord.push_back(*curr);
    // Outline3d* new_cut(new Outline3d(pen, Pen::Xfine, false, m_perp));
	    // new_cut->add(Edge3d(T_c, H_c, true));
	    // cuts.push_back(new_cut);
	  }

        else // part of edge is outside
          {
            if (max_ht > eps) // part of edge is inside
              {
		Point cut = T_c + (T_ht/(T_ht - H_ht))*(H_c - T_c);
                cut_ends.push_back(cut);

		// TO DO: push appropriate marker onto cuts

                if (T_ht > eps) // crossing from inside
                  {
                    // unchopped portion
                    bord.push_back(Edge3d(T_c, cut, (*curr).seen()));

                    if (cut_ends.size() == 2) // found second crossing
		      {
			m_clipped=true;
			m_updated=false;

			bord.push_back(Edge3d(cut_ends.at(1),
					      cut_ends.at(0), false));

			cut_ends.clear();
		      }
                  }

                else // crossing from outside
                  {
                    if (cut_ends.size() == 2) // found second crossing
		      {
			m_clipped=true;
			m_updated=false;

			bord.push_back(Edge3d(cut_ends.at(0),
					      cut_ends.at(1), false));

			cut_ends.clear();
		      }

                    bord.push_back(Edge3d(cut, H_c, (*curr).seen()));
                  }
              } // end of crossing; else edge is outside, no action
          } // end of edge partially outside
      } // all edges examined

    m_border.clear();
    swap(m_border,bord);

    if (!m_updated)
      update();

    return this;
  } // end of Outline3d::chop()


  // planar cut
  Outline3d* Outline3d::slice(const Knife& knife, std::vector<Shard*>& cuts)
  {
#ifdef DEBUG
    std::cerr << "Outline3d::slice()" << std::endl;
#endif
    if (is_empty())
      {
	update();
	return this;
      }

    // else chop edges
    const double eps(EPIX2_EPSILON); // half-thickness of cutting plane

    const Vector blade(knife.m_blade); // points toward retained part
    const Pen pen(knife.m_pen);

    std::vector<Edge3d> bord;

    std::vector<Point> cut_ends; // store cut ends

    // Examine heights of edges serially, check for crossings.
    // For clipping, a "crossing" occurs when max height > eps and 
    // min height < eps.
    Point T_c, H_c; // current head, tail
    double T_ht, H_ht, max_ht, min_ht;

    for (Ep3 curr=m_border.begin(); curr!=m_border.end(); ++curr)
      {
	T_c = (*curr).tail();
	H_c = (*curr).head();

	T_ht = blade.height(T_c);
	H_ht = blade.height(H_c);

	max_ht = max(T_ht, H_ht);
	min_ht = min(T_ht, H_ht);

	if ( -eps <= min_ht && max_ht <= eps) // edge lies in cut plane
	  {
	    Outline3d* new_cut(new Outline3d(pen, Pen::Xfine, false, m_perp));
	    new_cut->add(Edge3d(T_c, H_c, true));
	    cuts.push_back(new_cut);
	  }

        else if (min_ht > eps || max_ht < -eps) // no crossing
          continue;

        else // edge crosses
          {
	    // TO DO: push appropriate marker onto cuts
	    // (Allocate Glyph3d, push onto cuts)
	    // cut_ends.push_back(T_c + (T_ht/(T_ht - H_ht))*(H_c - T_c));
          }
      } // all edges examined

    m_border.clear(); // cuts will be appended by caller

    if (!m_updated)
      update();

    return this;
  } // end of Outline3d::slice()


  // photograph to specified Screen; may assume we're "updated"
  void Outline3d::shoot(Screen& S, const Camera& cam, const Color& ambient,
			const std::vector<Light>& spots, const Fog& fog,
			const photo_type SHADE) const
  {
#ifdef DEBUG
    std::cerr << "Outline3d::shoot()" << std::endl;
#endif
    if (is_empty())
      return;

    //else
    Vector to_camera = cam.viewpt() - here();
    double distance_to_camera = norm(to_camera);
    to_camera *= 1.0/distance_to_camera;

    // Inefficient, but no default Pen2d ctor...iron out later
    Pen2d line(cam(m_line.color()), m_line.width());
    Pen2d base(cam(m_base.color()), m_base.width());

    // TO DO: Is this the "right" behavior?
    if (SHADE != X_RAY)
      {
	Color the_line(ambient.filter(m_line.color()));
	Color the_base(ambient.filter(m_base.color()));

	if (SHADE == PHOTO) // not FLASH or TRACE
	  {
	    // weighted face normal shading; Magic Number
	    the_line *= 0.25*(3+fabs(to_camera|m_perp));
	    the_base *= 0.25*(3+fabs(to_camera|m_perp));
	  }

	line = Pen2d(cam(fog.mutes(the_line, distance_to_camera)),
		     m_line.width());
	base = Pen2d(cam(fog.mutes(the_base, distance_to_camera)),
		     m_base.width());
      }

    Outline2d value(line, base, m_clipped);

    // project edges
    for (Ep3 p=m_border.begin(); p!=m_border.end(); ++p)
      value.add(Edge2d(cam((*p).tail()), cam((*p).head()), (*p).seen()));

    // add to screen
    S << value;
  } // end of Outline3d::shoot()


  // Update (mutable) location and area
  void Outline3d::update() const
  {
#ifdef DEBUG
    std::cerr << "Outline3d::update()" << std::endl;
    print(m_perp, "  normal vector");
#endif
    if (!m_updated)
      {
	m_area = EPIX2_EPSILON*m_border.size();
	m_updated=true;
      }
  }


  double Outline3d::priority() const
  {
#ifdef DEBUG
    std::cerr << "Outline3d::priority()" << std::endl;
#endif
    if (!m_updated)
      update();

    return m_area;
  }


  Outline3d* Outline3d::in_coords(const Gyro& gyro)
  {
#ifdef DEBUG
    std::cerr << "Outline3d::in_coords()" << std::endl;
#endif
    m_perp = gyro.in_coords(m_perp);

    // convert edge data to B coordinates
    std::vector<Edge3d> new_border;
    for (Ep3 p=m_border.begin(); p!=m_border.end(); ++p)
      new_border.push_back(gyro.in_coords(*p));

    swap(m_border, new_border);

    return this;
  } // end of Outline3d::in_coords


  // place in BSP
  void Outline3d::add_to(Layer* T) const
  {
#ifdef DEBUG
    std::cerr << "Outline3d::add_to()" << std::endl;
#endif
    if (is_empty())
      return;

    // else
    Vector N(T->perp());

    Ep3 p(m_border.begin());

    double tmp_ht(N.height((*p).tail()));
    double max_ht(tmp_ht);
    double min_ht(tmp_ht);

    // calculate heights of vertices wrt T
    for(p=m_border.begin(); p!=m_border.end(); ++p)
      {
	tmp_ht = N.height((*p).head());

	min_ht = min(min_ht, tmp_ht);
	max_ht = max(max_ht, tmp_ht);
      }

    if (max_ht <= 0) // add to back
      {
	T->add_b(this);
      }

    else if (0 <= min_ht)
      {
	T->add_f(this);
      }

    else if (-EPIX2_EPSILON <= min_ht && max_ht <= EPIX2_EPSILON)
      {
	// merging non-coplanar outlines can get us here
	Outline3d* tmp(clone());
	tmp->m_perp = T->perp(); // reorient
	T->add(tmp); // coplanar
	delete tmp;
      }
    else // split
      {
	std::vector<Shard*> scratch; // store cuts, and discard

        Outline3d* S_top(clone());
	T->add_f(S_top->chop(Knife(N, Pen::Xfine), scratch));
	delete S_top; // previous line clones S_top

        Outline3d* S_bot(clone());
	T->add_b(S_bot->chop(Knife(-N, Pen::Xfine), scratch));
	delete S_bot;
      }
  } // end of Outline3d::add_to(Layer*)


  // Our location, for reflectivity calculation
  Point Outline3d::here() const
  {
    return m_perp.tail();
  }

  // Unit normal vector at barycenter
  Vector Outline3d::perp() const
  {
#ifdef DEBUG
    std::cerr << "Outline3d::perp()" << std::endl;
#endif
    return m_perp;
  }


  // specal to Outline3d
  Outline3d& Outline3d::add(const Edge3d& E)
  {
    m_border.push_back(E);
    m_updated = false;
    return *this;
  }

  bool Outline3d::is_empty() const
  {
#ifdef DEBUG
    std::cerr << "Outline3d::is_empty()" << std::endl
	      << "  border.size = " << m_border.size() << std::endl;
#endif
    return (m_border.size() == 0 ||
	    m_line.color().is_unset() ||
	    m_line.width() == Length(0));
  }
} // end of namespace

