/* 
 * Shard_2d.cc -- epix2::Plate3d class functions
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef DEBUG
#include "Debug.h"
#endif

#include <cmath>
#include <sstream>

#include "Errors.h"
#include "Functions.h"

#include "Edge_Data.h"
#include "Visibility.h"

#include "Point.h"
#include "Vector.h"

#include "Color.h"
#include "Fog.h"

#include "Gyro.h"

#include "Knife.h"

#include "Styles.h"
#include "Lighting.h"

#include "Camera.h"
#include "Screen.h"
// #include "Picture.h"

#include "Shard.h"
#include "Layer.h"

#include "Tile_2d.h"

#include "Shard_1d.h"
#include "Shard_2d.h"

namespace ePiX2 {

  // static function
  double area(const Point& x0, const Point& x1, const Point& x2)
  {
    Vector V1(x1-x0), V2(x2-x0);
    double IP=(V1|V2);

    return 0.5*sqrt((V1|V1)*(V2|V2) - IP*IP); // half norm of cross product
  }

  // where edge (T,H) crosses N perp
  Point crossing(const Vector& N, const Point& T, const Point& H)
  {
#ifdef DEBUG
    std::cerr << "crossing()" << std::endl;
#endif
    double T_ht = N.height(T);
    double H_ht = N.height(H);

    return T + (T_ht/(T_ht - H_ht))*(H-T);
  }


  // aliases
  typedef std::vector<Edge3d>::const_iterator Ep3;
  typedef std::vector<Wire3d>::const_iterator Wp3;

  // build: perp vector, fill, edge, all edges drawn
  Plate3d::Plate3d(const Ink& ink, const Pen& pen,
		   const Ink& ink_b, const Pen& pen_b,
		   bool cl, const Vector& perp)
    : m_front_ink(ink), m_front_pen(pen),
      m_back_ink(ink_b), m_back_pen(pen_b),
      m_clipped(cl), m_perp(perp), m_area(-1), m_updated(true) { }

  Plate3d::Plate3d(const Vector& N)
    : m_front_ink(White()), m_front_pen(Pen()),
      m_back_ink(White()), m_back_pen(Pen()),
      m_clipped(false), m_perp(N), m_area(-1), m_updated(true) { }


  /*
  Plate3d Plate3d::Empty = Plate3d(Ink(), Pen::Xfine,
				   Ink(), Pen::Xfine, false, Vector());
  */

  Plate3d* Plate3d::clone(void) const
  {
    return new Plate3d(*this);
  }


  // half-space portion; add new edge data to cuts
  Plate3d* Plate3d::chop(const Knife& knife, std::vector<Shard*>& cuts)
  {
#ifdef DEBUG
    std::cerr << "Plate3d::chop()" << std::endl;
#endif

    if (m_border.size() == 0)
      return this;

    // else chop edges
    const double eps(EPIX2_EPSILON); // half-thickness of cutting plane

    const Vector blade(knife.m_blade); // points toward retained part
    const Pen pen(knife.m_pen);

    std::vector<Edge3d> bord; // new border

    std::vector<Point> cut_ends; // store cut ends

    // Examine heights of edges serially, check for crossings.
    // For clipping, a "crossing" occurs when max height > eps and 
    // min height < eps.
    Point T_c, H_c; // current head, tail
    double T_ht, H_ht, max_ht, min_ht;

    for (Ep3 curr=m_border.begin(); curr!=m_border.end(); ++curr)
      {
	T_c = (*curr).tail();
	H_c = (*curr).head();

	// drop null edges
	if (T_c == H_c)
	  continue;

	T_ht = blade.height(T_c);
	H_ht = blade.height(H_c);

	max_ht = max(T_ht, H_ht);
	min_ht = min(T_ht, H_ht);

        if (min_ht > eps) // no crossing
          bord.push_back(*curr);

        else if ( -eps <= min_ht && max_ht <= eps) // edge lies in cut plane
          {
            bord.push_back(*curr);
	    // TO DO: add edge to cuts
          }

        else // part of edge is outside
          {
            if (max_ht > eps) // part of edge is inside
              {
		Point cut = T_c + (T_ht/(T_ht - H_ht))*(H_c - T_c);
                cut_ends.push_back(cut);

                if (T_ht > eps) // crossing from inside
                  {
                    // unchopped portion
                    bord.push_back(Edge3d(T_c, cut, (*curr).seen()));

                    if (cut_ends.size() == 2) // found second crossing
		      {
			if (cut_ends.at(0) != cut_ends.at(1))
			  {
			    m_clipped=true;
			    m_updated=false;

			    bord.push_back(Edge3d(cut_ends.at(1),
						  cut_ends.at(0), false));

			    if (pen.width() != Length(0))
			      {
				Outline3d* new_cut(new Outline3d(pen, 
								 Pen::Xfine,
								 false,
								 m_perp));
				new_cut->add(Edge3d(cut_ends.at(1),
						    cut_ends.at(0), true));
				cuts.push_back(new_cut);
			      }
			  }
			cut_ends.clear();
		      }
                  }

                else // crossing from outside
                  {
                    if (cut_ends.size() == 2) // found second crossing
		      {
			if (cut_ends.at(0) != cut_ends.at(1))
			  {
			    m_clipped=true;
			    m_updated=false;

			    bord.push_back(Edge3d(cut_ends.at(0),
						  cut_ends.at(1), false));

			    if (pen.width() != Length(0))
			      {
				Outline3d* new_cut(new Outline3d(pen, 
								 Pen::Xfine,
								 false,
								 m_perp));
				new_cut->add(Edge3d(cut_ends.at(0), 
						    cut_ends.at(1), true));
				cuts.push_back(new_cut);
			      }
			  }

			cut_ends.clear();
		      }

                    bord.push_back(Edge3d(cut, H_c, (*curr).seen()));
                  }
              } // end of crossing; else edge is outside, no action
          } // end of edge partially outside
      } // all edges examined

    swap(m_border,bord);

    if (!m_updated)
      update();

    return this;
  } // end of Plate3d::chop()


  // planar cut; typically stores all useful data in cuts
  Plate3d* Plate3d::slice(const Knife& knife, std::vector<Shard*>& cuts)
  {
#ifdef DEBUG
    std::cerr << "Plate3d::slice()" << std::endl;
#endif
    if (m_border.empty() || m_perp.same_perp(knife.m_blade))
      return this;

    // else
    const double eps(EPIX2_EPSILON); // half-thickness of cutting plane

    const Vector blade(knife.m_blade);
    const Pen pen(knife.m_pen);

    std::vector<Point> cut_ends; // store cut ends

    // Examine heights of edges serially, check for crossings.
    // For clipping, a "crossing" occurs when max height > eps and 
    // min height < eps.
    double T_ht, H_ht, max_ht, min_ht;
    Point T_c, H_c; // current head, tail

    for (Ep3 curr=m_border.begin(); curr!=m_border.end(); ++curr)
      {
	T_c = (*curr).tail();
	H_c = (*curr).head();

	// drop null edges
	if (T_c == H_c)
	  continue;

	T_ht = blade.height(T_c);
	H_ht = blade.height(H_c);

	max_ht = max(T_ht, H_ht);
	min_ht = min(T_ht, H_ht);

	if (min_ht > eps) // no crossing
	  continue;

	else if ( -eps <= min_ht && max_ht <= eps) // edge lies in cut plane
	  {
	    Outline3d* new_cut(new Outline3d(pen, Pen::Xfine, false, m_perp));
	    new_cut->add(Edge3d(T_c, H_c, true));
	    cuts.push_back(new_cut);
	  }

	// blade cuts edge
	else if (max_ht > eps)
	  {
	    cut_ends.push_back(T_c + (T_ht/(T_ht - H_ht))*(H_c - T_c));

	    // found second crossing
	    if (cut_ends.size() == 2 &&
		cut_ends.at(0) != cut_ends.at(1))
	      {
		Outline3d* C(new Outline3d(pen, Pen::Xfine, false, m_perp));
		C->add(Edge3d(cut_ends.at(0), cut_ends.at(1), true));
		cuts.push_back(C);
	      }
	  }
      } // all edges examined 

    m_border.clear(); // discard edges
    return this;
  }

  Plate3d& Plate3d::front_pen(const Pen& b)
  {
    m_front_pen.apply(b);
    return *this;
  }

  Plate3d& Plate3d::front_ink(const Ink& b)
  {
    m_front_ink.apply(b);
    return *this;
  }

  Plate3d& Plate3d::back_pen(const Pen& b)
  {
    m_back_pen.apply(b);
    return *this;
  }

  Plate3d& Plate3d::back_ink(const Ink& b)
  {
    m_back_ink.apply(b);
    return *this;
  }

  Plate3d& Plate3d::fix()
  {
    m_front_pen.fix();
    m_front_ink.fix();

    m_back_pen.fix();
    m_back_ink.fix();

    return *this;
  }


  // photograph to specified Screen; may assume we're "updated"
  void Plate3d::shoot(Screen& S, const Camera& cam, const Color& ambient,
		      const std::vector<Light>& spots, const Fog& fog,
		      const photo_type SHADE) const
  {
#ifdef DEBUG
    std::cerr << "Plate3d::shoot()" << std::endl;
#endif
    Point here(m_perp.tail());
    Vector to_cam(cam.viewpt() - here);
    double dist_cam = norm(to_cam);
    to_cam *= 1.0/dist_cam;

    const double cos_th(to_cam|m_perp); // cosine of face normal

    // get Ink and Pen
    Ink ink(m_front_ink);
    Pen pen(m_front_pen);

    if (cos_th < 0) // face points away
      {
	ink = m_back_ink;
	pen = m_back_pen;
      }

    // convert to Color and Pen2d
    Color face = ink.color();

    Color my_edge = pen.color();
    my_edge *= fabs(cos_th); // face normal shading
    Pen2d edge(cam(fog.mutes(my_edge, dist_cam)), pen.width());

    if (SHADE == PHOTO)
      {
	// color reflected by face to camera by ambient light
	face = face_ambient(ink, m_perp, to_cam, ambient);

	std::vector<Light>::const_iterator p;
	for (p=spots.begin(); p!=spots.end(); ++p)
	  {
	    Vector tmp_dir((*p).here() - here);

	    if (norm(tmp_dir) < EPIX2_EPSILON)
	      continue; // ignore spot light on face

	    else
	      {
		tmp_dir.normalize();

		// color reflected from spot light
		Color tmp_col(face_spot(ink, m_perp, tmp_dir, to_cam,
					(*p).incident(fog, here)));
		face = face.superpose(tmp_col); // add contribution from spot
	      }
	  }
	// face = light reflected from us; apply fog
	face = fog.mutes(face, dist_cam);
      }

    else if (SHADE == FLASH) // use "true" colors with face normal shading
      face *= 0.5*(1+fabs(cos_th));

    else if (SHADE == TRACE)
      face = White();

    else if (SHADE == X_RAY)
      face = Neutral(); // unfilled

    Plate2d value(cam(face), edge, m_clipped);

    // project edges
    for (Ep3 p=m_border.begin(); p!=m_border.end(); ++p)
      value.add(Edge2d(cam((*p).tail()), cam((*p).head()), (*p).seen()));

    S << value; // add to screen
  } // end of Plate3d::shoot()


  /*
  void Plate3d::shoot(Picture& S, const Camera& cam, const double d,
		    photo_type SHADE) const
  {
    shoot(S.screen(), cam, d, SHADE);
  }
  */


  // Returns area, for sorting before adding to BSP.
  // Update (mutable) location while we're at it...
  void Plate3d::update(void) const
  {
#ifdef DEBUG
    std::cerr << "Plate3d::update()" << std::endl;
#endif
    if (!m_updated)
      {
	double val=0;
	// Vector barycenter; // zero vector at Origin

	Ep3 p = m_border.begin();
	Point O_ref = (*p).tail();

	// Shard_Bases are convex, so needn't worry about local orientation
	for (p=m_border.begin(); p!=m_border.end(); ++p)
	  {
	    val += ePiX2::area(O_ref, (*p).tail(), (*p).head()); // area
	    // barycenter += ((*p).tail() - Origin);                // location
	  }

	// barycenter *= 1.0/m_border.size(); // arithmetic mean

	// Point here = barycenter.head();
	// m_perp.move_to(m_here);

	m_area = val;
	m_updated=true;
      }
  }


  double Plate3d::priority(void) const
  {
#ifdef DEBUG
    std::cerr << "Plate3d::priority()" << std::endl;
#endif
    if (!m_updated)
      {
	epix_warning("Plate3d::priority(): Shard not up to date");
	update();
      }
    return m_area;
  }


  Plate3d* Plate3d::in_coords(const Gyro& gyro)
  {
#ifdef DEBUG
    std::cerr << "Plate3d::in_coords()" << std::endl;
#endif
    if (!m_updated)
      {
	epix_warning("Plate3d::in_coords(): Shard not up to date");
	update();
      }

    // compute new border and cuts
    std::vector<Edge3d> new_border;
    for (Ep3 p=m_border.begin(); p!=m_border.end(); ++p)
      new_border.push_back(gyro.in_coords(*p));

    // Now it's safe to update our data
    m_perp = gyro.in_coords(m_perp);

    swap(m_border, new_border);

    m_area *= (c*c); // area scales like c^2
    m_updated=true;

    return this;
  } // end of Plate3d::in_coords

  bool Plate3d::hides(const Point& vpt, const Point& X) const
  {
    if (is_empty())
      return false;

    // else
    Ep3 p=m_border.begin();
    Point vtx1 = (*p).tail();
    Point vtx2, vtx3;

    // check triangle-by-triangle
    for (Ep3 p=m_border.begin(); p!=m_border.end(); ++p)
      {
	vtx2 = (*p).tail();
	vtx3 = (*p).head();

	if ( (!on_same_side(vtx1, vtx2, vtx3, vpt, X)) &&
	     on_same_side(vtx1, vtx2, vpt, X,  vtx3) &&
	     on_same_side(vtx2, vtx3, vpt, X,  vtx1) &&
	     on_same_side(vtx3, vtx1, vpt, X,  vtx2) )
	  return true;
      }

    return false;
  }


  // place in BSP
  void Plate3d::add_to(Layer* T) const
  {
    const double eps = EPIX2_EPSILON;

#ifdef DEBUG
    std::cerr << "Plate3d::add_to()" << std::endl;
#endif
    if (m_border.size() < 2)
      return;

    // else
    Vector N = T->perp();

    Ep3 p = m_border.begin();

    double tmp_ht = N.height((*p).tail());
    double max_ht = tmp_ht;
    double min_ht = tmp_ht;

    // calculate heights of vertices wrt this
    for(p=m_border.begin(); p!=m_border.end(); ++p)
      {
	tmp_ht = N.height((*p).head());

	min_ht = min(min_ht, tmp_ht);
	max_ht = max(max_ht, tmp_ht);
      }

    if (-eps <= min_ht && max_ht <= eps)
      T->add(this); // coplanar

    else if (max_ht <= eps)
      T->add_b(this); // add to back

    else if (min_ht >= -eps)
      T->add_f(this);

    else // split
      {
	std::vector<Shard*> scratch; // collect and discard cuts
        Plate3d* S_top(clone());
	T->add_f(S_top->chop(Knife(N, Pen::Xfine), scratch));
	delete S_top; // previous line clones

        Plate3d* S_bot(clone());
	T->add_b(S_bot->chop(Knife(-N, Pen::Xfine), scratch));
        delete S_bot;
      }
  } // end of Plate3d::add_to(Layer*)


  // Our location, for reflectivity calculation
  Point Plate3d::here(void) const
  {
    if (!m_updated)
      {
	epix_warning("Plate3d::here(): Shard not up to date");
	update();
      }

    return m_perp.tail();
  }

  // Unit normal vector at barycenter
  Vector Plate3d::perp(void) const
  {
    return m_perp;
  }

  // specal to Plate3d
  Plate3d* Plate3d::add(const Edge3d& E)
  {
    m_border.push_back(E);
    m_updated = false;
    return this;
  }

  bool Plate3d::is_empty(void) const
  {
    return (m_border.size() == 0);
  }
} // end of namespace

