/* 
 * Sphere.cc -- ePiX2::Sphere class
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 14, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef DEBUG
#include "Debug.h"
#endif

#include <cmath>

#include "Functions.h"
#include "Point.h"
#include "Vector.h"

#include "Edge_Data.h"

// Uncomment as necessary
// #include "Shard_0d.h"
// #include "Shard_1d.h"
#include "Shard_2d.h"

#include "Sphere.h"

namespace ePiX2 {

  // static helper
  Point sph(const double rad,
	    const unsigned int i, const unsigned int j,
	    const unsigned int n1, const unsigned int n2)
  {
    return Point(rad*cos(2*i*M_PI/n1)*sin(j*M_PI/n2),
		 rad*sin(2*i*M_PI/n1)*sin(j*M_PI/n2),
		 -rad*cos(j*M_PI/n2));
  }

  Sphere::Sphere(const Point& ctr, const double rad,
		 const unsigned int n_long, const unsigned int n_lat)
    : m_ctr(ctr), m_rad(rad)
  {
    Point oo, po, pp, op; // vertices of quadrilaterals

    for (unsigned int i=0; i<n_long; ++i)
      for (unsigned int j=0; j<n_lat; ++j)
	{
	  // will move our frame to ctr 
	  oo = sph(rad, i, j, n_long, n_lat);
	  po = sph(rad, i+1, j, n_long, n_lat);
	  pp = sph(rad, i+1, j+1, n_long, n_lat);
	  op = sph(rad, i, j+1, n_long, n_lat);

	  if (j == n_lat-1) // facet touches north pole, pp == op
	    {
	      Plate3d facet(((po-oo)*(op-oo)).normalize());

	      facet.add(Edge3d(oo,po,true));
	      facet.add(Edge3d(po,pp,true));
	      facet.add(Edge3d(op,oo,true));

	      facet.update();
	      add_shard(facet);
	    }

	  else if (j == 0) // facet touches south pole, oo == po
	    {
	      Plate3d facet(((pp-oo)*(op-oo)).normalize());

	      facet.add(Edge3d(po,pp,true));
	      facet.add(Edge3d(pp,op,true));
	      facet.add(Edge3d(op,oo,true));

	      facet.update();
	      add_shard(facet);
	    }

	  else
	    {
	      Plate3d facet(((po-oo)*(op-oo)).normalize());

	      facet.add(Edge3d(oo,po,true));
	      facet.add(Edge3d(po,pp,true));
	      facet.add(Edge3d(pp,op,true));
	      facet.add(Edge3d(op,oo,true));

	      facet.update();
	      add_shard(facet);
	    }
	}
    Object::move_to(ctr);
  }

  // If Sphere contains data, must define copy_ctor, operator=
  Sphere::Sphere(const Sphere& obj)
    : Object(obj), m_ctr(obj.m_ctr), m_rad(obj.m_rad) { }


  Sphere& Sphere::operator=(const Sphere& obj)
  {
    if (this != &obj) // May want more lenient condition
      {
	// allocate any heap data, free resources, do whatever else...
	Object::operator=(obj);

	m_ctr = obj.m_ctr;
	m_rad = obj.m_rad;
      }

    return *this;
  }

  Sphere* Sphere::clone(void) const
  {
    return new Sphere(*this);
  }
} // end of namespace

