/* 
 * Styles.cc -- ePiX2 style Ink, Pen, and Stencil
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "Functions.h"
#include "Color.h"
#include "Styles.h"

namespace ePiX2 {

  /* * * Ink * * */
  Ink::Ink(const Color& col, double r, bool p)
    : m_color(col), m_gloss(r), m_perm(p) { }

  // replace if we're not permanent
  Ink& Ink::apply(const Ink& col)
  {
    if (!m_perm)
      *this = col;

    return *this;
  }

  // mix colors if we're not permanent
  Ink& Ink::blend(const Ink& col, const double d)
  {
    if (!m_perm)
      {
	m_color.blend(col.m_color, clip_0_1(d));
	// TO DO: handle gloss
	m_perm = col.m_perm;
      }

    return *this;
  }

  // Color negative
  Ink& Ink::invert(void)
  {
    if (!m_perm)
      m_color.invert();

    return *this;
  }


  Color  Ink::color() const     { return m_color; }
  Ink&   Ink::gloss(const double r) { m_gloss=r; return *this; }
  double Ink::gloss() const     { return m_gloss; }
  double Ink::alpha() const     { return m_alpha; }
  void   Ink::fix()             { m_perm=true; }
  bool   Ink::indelible() const { return m_perm; }


  /* * * Pen * * */
  Pen::Pen() : m_ink(Black()), m_width(PLAIN_WIDTH) { }

  Pen::Pen(const Ink& p, const Length& wid)
    : m_ink(p), m_width(wid) { }

  Pen::Pen(const Ink& p, std::string wid)
    : m_ink(p), m_width(Length(wid)) { }

  Pen Pen::Xfine = Pen(Black(), Length(0));


  // set color and width if color not permanent
  Pen& Pen::apply(const Pen& p)
  {
    if (!m_ink.indelible())
      {
	m_ink.apply(p.m_ink);
	m_width = p.m_width;
      }
    return *this;
  }

  /*
  Pen& Pen::apply(const Ink& p)
  {
    m_ink.apply(p);
    return *this;
  }
  */


  // named constructors
  Pen plain(const Color& col)
  {
    return Pen(col, PLAIN_WIDTH);
  }

  Pen bold(const Color& col)
  {
    return Pen(col, BOLD_WIDTH);
  }

  Pen bbold(const Color& col)
  {
    return Pen(col, BBOLD_WIDTH);
  }

  Color  Pen::color() const { return m_ink.color(); }
  Length Pen::width() const { return m_width; }
  double Pen::gloss() const { return m_ink.gloss(); }
  double Pen::alpha() const { return m_ink.alpha(); }
  Pen& Pen::fix() { m_ink.fix(); return *this; }

  Duo_Pen::Duo_Pen(const Pen& line, const Pen& base)
    : m_line(line), m_base(base) { }

  Duo_Pen::Duo_Pen(const Color& c_line, const std::string& w_line,
		       const Color& c_base, const std::string& w_base)
    : m_line(Pen(c_line, w_line)), m_base(Pen(c_base, w_base)) { }

  Pen Duo_Pen::line() const { return m_line; }
  Pen Duo_Pen::base() const { return m_base; }


  /* * * Pen2d * * */
  Pen2d::Pen2d(const Color& col, const Length& wid)
    : m_color(col), m_width(wid) { }

  Pen2d Pen2d::Xfine(Black(), Length(0));

  bool Pen2d::operator== (const Pen2d& pen) const
  {
    return ((m_color == pen.m_color) && (m_width == pen.m_width));
  }

  bool Pen2d::operator!= (const Pen2d& pen) const
  {
    return !(*this == pen);
  }

  Color  Pen2d::color() const { return m_color; }
  Length Pen2d::width() const { return m_width; }


  /* * * Stencil * * */
  Stencil::Stencil(const Ink& text, const Ink& back,
		   const Length& pad, const Pen& edge,
		   align_type align, font_type font, const double th)
    : m_text(text), m_back(back), m_pad(pad), m_edge(edge), 
      m_align(align), m_font(font), m_angle(th) { }

  Stencil& Stencil::apply(const Stencil& arg)
  {
    m_text.apply(arg.m_text);
    m_back.apply(arg.m_back);
    m_pad = arg.m_pad;
    m_edge.apply(arg.m_edge);

    m_align = arg.m_align;
    m_font  = arg.m_font;
    m_angle = arg.m_angle;

    return *this;
  }

  Stencil& Stencil::text(const Color& col)
  {
    m_text.apply(col);
    return *this;
  }

  Stencil& Stencil::back(const Color& col)
  {
    m_back.apply(col);
    return *this;
  }

  Stencil&  Stencil::pad(const Length& pad)
  {
    m_pad = pad;
    return *this;
  }

  Stencil& Stencil::edge(const Pen& pen)
  {
    m_edge.apply(pen);
    return *this;
  }

  Stencil& Stencil::align(const align_type a)
  {
    m_align = a;
    return *this;
  }

  Stencil&  Stencil::font(const font_type f)
  {
    m_font = f;
    return *this;
  }

  Stencil& Stencil::angle(const double th)
  {
    m_angle = th;
    return *this;
  }

  Stencil& Stencil::fix()
  {
    m_text.fix();
    m_back.fix();
    m_edge.fix();

    return *this;
  }

  Stencil& text_style()
  {
    static Stencil* current_text_style(new Stencil());
    return *current_text_style;
  }

  Color  Stencil::text() const { return m_text.color(); }
  Color  Stencil::back() const { return m_back.color(); }
  Length Stencil::pad()  const { return m_pad; }
  Pen    Stencil::edge() const { return m_edge; }
} /* end of namespace */
