/* 
 * Tile_2d.cc -- epix2::Plate2d class functions
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 27, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef DEBUG
#include "Debug.h"
#endif

// #include <cmath>
#include <sstream>

#include "Functions.h"
#include "Pair.h"
#include "Pair_Map.h"
#include "Scissors.h"

#include "Edge_Data.h"

#include "Color.h"
#include "Fog.h"

#include "Styles.h"

#include "Screen.h"

#include "Output.h"

#include "Tile.h"
#include "Tile_1d.h"
#include "Tile_2d.h"

namespace ePiX2 {

  // aliases
  typedef std::vector<Edge2d>::const_iterator Ep2;
  typedef std::vector<Wire2d>::const_iterator Wp2;

  Plate2d::Plate2d(const Color& color, const Pen2d& pen, bool cl)
    : m_color(color), m_pen(pen), m_clipped(cl) { }

  //  Plate2d Plate2d::Empty = Plate2d(Black(), Pen2d(), false);

  Plate2d* Plate2d::clone(void) const
  {
    return new Plate2d(*this);
  }


  Plate2d& Plate2d::map_by(const Pair_Map& f)
  {
    if (m_border.size() == 0)
      return *this;

    std::vector<Edge2d> bord;

    // TO DO: Subdivide edges if f doesn't preserve lines
    for (Ep2 p=m_border.begin(); p!=m_border.end(); ++p)
      bord.push_back(Edge2d(f(p->tail()), f(p->head()), (*p).seen()));

    swap(m_border, bord);
    return *this;
  }


  // implemented analogously to Plate3d::chop(Knife)
  Plate2d& Plate2d::crop_by(const Scissors& S)
  {
    if (m_border.size() == 0)
      return *this;

    // else crop edges
    const double eps=EPIX2_EPSILON; // half-thickness of cutting plane

    std::vector<Edge2d> bord;
    std::vector<Pair> cut_ends; // store cut ends

    // Examine heights of edges serially, check for crossings.
    // For clipping, a "crossing" occurs when max height > eps and 
    // min height < eps.
    Pair T_c, H_c; // current head, tail
    double T_ht, H_ht, max_ht, min_ht;

    for (Ep2 curr=m_border.begin(); curr!=m_border.end(); ++curr)
      {
	T_c = (*curr).tail();
	H_c = (*curr).head();

	T_ht = S.height(T_c);
	H_ht = S.height(H_c);

	max_ht = max(T_ht, H_ht);
	min_ht = min(T_ht, H_ht);

        if (min_ht > eps) // no crossing
          bord.push_back(*curr);

        else if ( -eps <= min_ht && max_ht <= eps) // edge lies in cut plane
	  bord.push_back(*curr);

        else // part of edge is outside
          {
            if (max_ht > eps) // part of edge is inside
              {
		Pair cut(S.cuts(*curr));
                cut_ends.push_back(cut);

                if (T_ht > eps) // crossing from inside
                  {
                    // unchopped portion
                    bord.push_back(Edge2d(T_c, cut, (*curr).seen()));

                    if (cut_ends.size() == 2) // found second crossing
		      {
			if (cut_ends.at(0) != cut_ends.at(1))
			  {
			    m_clipped=true;
			    bord.push_back(Edge2d(cut_ends.at(1),
						  cut_ends.at(0), false));
			  }
			cut_ends.clear();
		      }
                  }

                else // crossing from outside
                  {
                    if (cut_ends.size() == 2) // found second crossing
		      {
			if (cut_ends.at(0) != cut_ends.at(1))
			  {
			    m_clipped=true;
			    bord.push_back(Edge2d(cut_ends.at(0),
						  cut_ends.at(1), false));
			  }
			cut_ends.clear();
		      }

                    bord.push_back(Edge2d(cut, H_c, (*curr).seen()));
                  }
              } // end of crossing; else edge is outside, no action
          } // end of edge partially outside
      } // all edges examined

    swap(m_border,bord);

    return *this;
  } // end of crop_by(Scissors)


  Plate2d& Plate2d::crop_to(const Mask::Mask& M)
  {
    M.crop(this);
    return *this;
  }

  void Plate2d::add(const Edge2d& E)
  {
    m_border.push_back(E);
  }

  bool Plate2d::is_empty(void) const
  {
    return (m_border.size() == 0);
  }


  std::string Plate2d::print_to(const Format& fmt) const
  {
    // cull null edges
    std::vector<Edge2d> bord;
    for (Ep2 p=m_border.begin(); p!=m_border.end(); ++p)
      if ((*p).tail() != (*p).head())
	bord.push_back(*p);

    if (bord.size() < 3) // must be at least a triangle
      return "";

    std::ostringstream obuf;
    obuf << fmt.print_plate(m_color, m_pen, m_clipped, bord);
    return obuf.str();
  }


  // rewrite according to Screen style storage scheme
  void Plate2d::add_colors_to(Screen& S) const
  {
    if (!m_pen.color().is_unset() && (Length(0) < m_pen.width()))
      S << m_pen.color();

    if (!m_color.is_unset())
      S << m_color;
  }
} // end of namespace

