/* 
 * Vector.cc -- ePiX2::Vector class and operations
 *
 * This file is part of ePiX, a preprocessor for creating high-quality 
 * line figures in LaTeX
 *
 * Version 2.0pre
 * Last Change: September 19, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <ahwang@mathcs.holycross.edu>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include "Debug.h"
#endif

#include "Constants.h"
#include "Errors.h"
#include "Functions.h"

#include "Vector.h"

namespace ePiX2 {

  Vector::Vector(const Point& tail, double dx1, double dx2, double dx3)
    : m_tail(tail),
      m_head(tail.x1()+dx1, tail.x2()+dx2, tail.x3()+dx3) { }

  Vector::Vector(const Point& tail, const Point& head)
    : m_tail(tail), m_head(head) { }


  // component functions
  double Vector::x1(void) const
  {
    return m_tail.x1();
  }
  double Vector::x2(void) const
  {
    return m_tail.x2();
  }
  double Vector::x3(void) const
  {
    return m_tail.x3();
  }

  double Vector::dx1(void) const
  {
    return m_head.x1() - m_tail.x1();
  }
  double Vector::dx2(void) const
  {
    return m_head.x2() - m_tail.x2();
  }
  double Vector::dx3(void) const
  {
    return m_head.x3() - m_tail.x3();
  }

  Point Vector::tail(void) const
  {
    return m_tail;
  }
  Point Vector::head(void) const
  {
    return m_head;
  }

  // affine translation
  Vector& Vector::move_to(const Point& arg)
  {
#ifdef DEBUG
    std::cerr << "Vector::move_to()" << std::endl;
#endif
    m_head = Point(arg.x1() + dx1(), arg.x2() + dx2(), arg.x3() + dx3());
    m_tail = arg;
    return *this;
  }

  Vector& Vector::operator*= (const double c)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator*=(double)" << std::endl;
#endif
    m_head = Point(x1() + c*dx1(), x2() + c*dx2(), x3() + c*dx3());

    return *this; 
  }

  // increment operators
  Vector& Vector::operator+= (const Vector& arg)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator+=(Vector)" << std::endl;
#endif
    if (m_tail != arg.m_tail)
      epix_warning("Implicit parallel translation");

    m_head = Point(m_head.x1() + arg.dx1(),
		   m_head.x2() + arg.dx2(),
		   m_head.x3() + arg.dx3());
    
    return *this;
  }

  Vector& Vector::operator-= (const Vector& arg)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator-=(Vector)" << std::endl;
#endif
    if (m_tail != arg.m_tail)
      epix_warning("Implicit parallel translation");
  
    m_head = Point(m_head.x1() - arg.dx1(),
		   m_head.x2() - arg.dx2(),
		   m_head.x3() - arg.dx3());

    return *this;
  }

  // cross product
  Vector& Vector::operator*= (const Vector& arg)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator*=(Vector)" << std::endl;
#endif
    if (m_tail != arg.m_tail)
      epix_warning("Implicit parallel translation");

    double a1(dx1()), a2(dx2()), a3(dx3());
    double b1(arg.dx1()), b2(arg.dx2()), b3(arg.dx3());

    m_head = Point(x1() + a2*b3 - a3*b2,
		   x2() + a3*b1 - a1*b3,
		   x3() + a1*b2 - a2*b1);
    return *this;
  }

  // warns if we're the zero vector
  Vector& Vector::normalize(void)
  {
#ifdef DEBUG
    std::cerr << "Vector::normalize()" << std::endl;
#endif
    double denom(norm(*this));
    if (denom < EPIX2_EPSILON)
      epix_warning("Normalizing the zero vector, no action");

    else
      {
	m_head = Point(x1() + dx1()/denom,
		       x2() + dx2()/denom,
		       x3() + dx3()/denom);
      }

    return *this;
  }

  // projection to arg(perp)
  Vector& Vector::operator%= (const Vector& arg)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator%=(Vector)" << std::endl;
#endif
    m_head = m_head % arg;
    m_tail = m_tail % arg;

    return *this;
  }


  // reflect us, treating perp as a mirror
  Vector& Vector::reflect_by(const Vector& perp)
  {
#ifdef DEBUG
    std::cerr << "Vector::reflect_by()" << std::endl;
#endif
    Point tail_tmp(m_tail % perp), head_tmp(m_head % perp);

    m_tail = m_tail + 2*(tail_tmp - m_tail);
    m_head = m_head + 2*(head_tmp - m_head);

    return *this;
  }


  // Create an orthonormal basis with axis as 3rd element, then rotate
  // in perp plane
  Vector& Vector::rotate_by(const double angle, const Vector& axis)
  {
#ifdef DEBUG
    std::cerr << "Vector::rotate_by()" << std::endl;
#endif
    double len(norm(axis));

    if (len < EPIX2_EPSILON) // axis is null, no action
      return *this;

    // else
    Vector tmp3(axis);
    tmp3.normalize();

    // Try displacement to tail as a second basis element
    Point ctr(tmp3.m_tail);
    Vector tmp1((m_tail - ctr) %= tmp3); // displacement to axis

    if (norm(tmp1) < EPIX2_EPSILON)   // tail is on axis
      {
	tmp1 = (m_head - ctr) %= tmp3; // try head

	if (norm(tmp1) < EPIX2_EPSILON) // we're on axis
	  return *this;
      }

    // else tmp1 is non-zero, orthog to tmp3
    tmp1.normalize();
    Vector tmp2(tmp3*tmp1);

    // get tail coords w.r.t. ctr
    double t1(tmp1.height(m_tail));
    double t2(tmp2.height(m_tail));
    double t3(tmp3.height(m_tail));

    double h1(tmp1.height(m_head));
    double h2(tmp2.height(m_head));
    double h3(tmp3.height(m_head));

    double c(Cos(angle));
    double s(Sin(angle));

    m_tail = ctr + (c*t1-s*t2)*tmp1 + (s*t1+c*t2)*tmp2 + t3*tmp3;
    m_head = ctr + (c*h1-s*h2)*tmp1 + (s*h1+c*h2)*tmp2 + h3*tmp3;

    return *this;
  }

  double Vector::height(const Point& arg) const
  {
#ifdef DEBUG
    std::cerr << "Vector::height()" << std::endl;
#endif
    double len(norm(*this));
    if (len < EPIX2_EPSILON)
      {
	epix_warning("Requested height w.r.t. zero vector");
	return 0;
      }

    else
      return (1.0/len)*((*this)|(arg-m_tail));
  }

  bool Vector::same_perp(Vector arg) const
  {
#ifdef DEBUG
    std::cerr << "Vector::same_perp()" << std::endl;
#endif
    // arg.m_tail in our perp plane, and arg parallel
    return (fabs((*this)|(arg.m_tail - m_tail)) < EPIX2_EPSILON &&
      norm(arg.move_to(m_tail) *= (*this)) < EPIX2_EPSILON);
  }


  // intersection of perp with E
  /*
  Point Vector::perp_hits(const Point& q1, const Point& q2) const
  {
    Vector dir=q2-q1;
    dir.move_to(m_tail);
    double t=-((q1-m_tail)|(*this))/(dir|(*this));

    return q1 + t*dir;
  }
  */

  /*
  Point Vector::perp_hits(const Edge& E) const
  {
    return perp_hits(E.first, E.second);
  }
  */

  // non-member functions
  Vector E_1(const Point& base)
  {
    return Vector(base, 1, 0, 0);
  }
  Vector E_2(const Point& base)
  {
    return Vector(base, 0, 1, 0);
  }
  Vector E_3(const Point& base)
  {
    return Vector(base, 0, 0, 1);
  }

  // Positions and components
  Point tail(const Vector& arg)
  {
    return arg.tail();
  }
  Point head(const Vector& arg)
  {
    return arg.head();
  }

  double dx1(const Vector& arg)
  {
    return arg.dx1();
  }
  double dx2(const Vector& arg)
  {
    return arg.dx2();
  }
  double dx3(const Vector& arg)
  {
    return arg.dx3();
  }

  // vector space operations; warning if vectors not based at same point
  Vector operator+ (Vector arg1, const Vector& arg2)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator+(Vector, Vector)" << std::endl;
#endif
    return arg1 += arg2;
  }
  Vector operator- (Vector arg1, const Vector& arg2)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator-(Vector, Vector)" << std::endl;
#endif
    return arg1 -= arg2;
  }
  Vector operator- (Vector arg)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator-(Vector)" << std::endl;
#endif
    return arg *= -1;
  }

  // cross product
  Vector operator* (Vector arg1, const Vector& arg2)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator*(Vector, Vector)" << std::endl;
#endif
    return arg1 *= arg2;
  }

  // scalar multiplication
  Vector operator* (const double c, Vector v)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator*(double, Vector)" << std::endl;
#endif
    return v *= c;
  }

  double norm(const Vector& arg)
  {
#ifdef DEBUG
    std::cerr << "Vector::norm()" << std::endl;
#endif
    double v1(arg.dx1()), v2(arg.dx2()), v3(arg.dx3());
    return sqrt(v1*v1 + v2*v2 + v3*v3); // can't use dot product
  }

  // (in)equality
  bool operator == (const Vector& arg1, const Vector& arg2)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator==" << std::endl;
#endif
    return (arg1.tail() == arg2.tail() && norm(arg2-arg1) < EPIX2_EPSILON);
  }

  bool operator != (const Vector& arg1, const Vector& arg2)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator!=" << std::endl;
#endif
    return !(arg1 == arg2);
  }
 
  double operator| (const Vector& arg1, const Vector& arg2)
  {
#ifdef DEBUG
    std::cerr << "Vector::operator|" << std::endl;
#endif
    if (arg1.m_tail != arg2.m_tail)
      epix_warning("Implicit translation in dot product");

    return
      (arg1.dx1())*(arg2.dx1()) +
      (arg1.dx2())*(arg2.dx2()) +
      (arg1.dx3())*(arg2.dx3());
  }

  // Linear dependence
  bool lin_dependent (const Vector& v1, const Vector& v2)
  {
    return (norm(v1*v2) <= EPIX2_EPSILON*norm(v1)*norm(v2));
  }

  bool lin_dependent (const Vector& v1, const Vector& v2, const Vector& v3)
  {
    return (fabs((v1*v2)|v3) <= EPIX2_EPSILON*norm(v1)*norm(v2)*norm(v3));
  }


  // Point - Point
  Vector operator- (const Point& head, const Point& tail)
  {
    return Vector(tail, head);
  }

  // non-members
  Point operator+ (const Point& arg, const Vector& T)
  {
    return Point(arg.x1() + T.dx1(), arg.x2() + T.dx2(), arg.x3() + T.dx3());
  }

  Point operator- (const Point& arg, const Vector& T)
  {
    return Point(arg.x1() - T.dx1(), arg.x2() - T.dx2(), arg.x3() - T.dx3());
  }

  // project to N perp; identity map if N == 0
  Point operator% (Point arg, const Vector& N)
  {
    const double len(norm(N));

    if (len > EPIX2_EPSILON)
      {
	double mult(((arg.x1() - N.tail().x1())*N.dx1() +
		     (arg.x2() - N.tail().x2())*N.dx2() +
		     (arg.x3() - N.tail().x3())*N.dx3())/(len*len));
	arg = arg-(mult*N);
      }
   
    return arg;
  }
}; /* end of namespace ePiX2 */
