/* 
 *  Vector.h -- epix2::Vector class and operators
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: September 19, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/* 
 * Vector.h
 *
 * This file provides the Vector class, conceptually a displacement in 
 * 3-dimensional affine space, implemented as a m_tail and components.
 *
 *  Constructors
 *   - m_tail and components: Vector(tail, Vx1, Vx2, Vx3)
 *   - tail and head; Vector(tail, head)
 *   - Standard basis (at origin)
 *
 *  Position and momenta:
 *   - x1(), dx1(), etc.
 *   - translate tail to arg: move_to(Point arg)
 *   - projected length of (arg-tail): height(Point arg)
 *
 *  Point/Vector manipulators
 *   - head(), tail();
 *   - Point - Point = Vector; Point + Vector = Point;
 *   - convex linear combination: midpoint(Point, Point, double=0.5);
 *
 *  Algebraic operators for vectors with same m_tail
 *   - Vector sum, difference, scalar multiplication
 *   - Euclidean operations 
 *     dot product (v|w)
 *     norm
 *     cross product (v*w)
 *   - (In)equality, linear dependence
 *
 *  Affine operations with respect to line/plane through w.tail()
 *     orthogonalization (v%=w: v projected to w perp)
 *     reflection        (v^=w: v reflected in w perp)
 *     projection        (v&=w: v projected to w)
 */

#ifndef EPIX2_VECTOR
#define EPIX2_VECTOR

#include "Point.h"

namespace ePiX2 {

  class Vector;
  double operator| (const Vector& arg1, const Vector& arg2);

  class Vector {
    friend class Point;

  public:  
    // Cartesian constructor
    explicit Vector(const Point& tail = Origin, 
		    double dx1=0, double dx2=0, double dx3=0);

    Vector(const Point& tail, const Point& head);

    // component functions
    Point tail() const;
    Point head() const;

    double x1() const;
    double x2() const;
    double x3() const;

    double dx1() const;
    double dx2() const;
    double dx3() const;

    // affine translation
    Vector& move_to(const Point&);

    // scalar multiplication
    Vector& operator*= (const double);

    // if norm != 0, make unit
    Vector& normalize();

    // vector addition/subtraction
    Vector& operator+= (const Vector& arg);
    Vector& operator-= (const Vector& arg);

    // cross product
    Vector& operator*= (const Vector& arg);

    // dot product
    friend double operator| (const Vector& arg1, const Vector& arg2);

    // affine projection to plane perp to arg
    Vector& operator%= (const Vector& arg);

    // named operations
    Vector& reflect_by(const Vector&);
    Vector& rotate_by(const double angle, const Vector& axis);

    double height(const Point&) const;

    bool same_perp(Vector) const;

  private:
    Point m_tail;
    Point m_head;

  }; // end of class Vector


  // Standard basis
  Vector E_1(const Point& base=Origin);
  Vector E_2(const Point& base=Origin);
  Vector E_3(const Point& base=Origin);

  // Positions and components
  Point tail(const Vector& arg);
  Point head(const Vector& arg);

  double dx1(const Vector& arg);
  double dx2(const Vector& arg);
  double dx3(const Vector& arg);

  // vector space operations; warning if vectors not based at same point
  Vector operator+ (Vector arg1, const Vector& arg2);
  Vector operator- (Vector arg1, const Vector& arg2);
  Vector operator- (Vector arg);

  // cross product
  Vector operator* (Vector arg1, const Vector& arg2);

  // scalar multiplication
  Vector operator* (const double c, Vector v);
  double norm(const Vector& arg);

  // (in)equality
  bool operator == (const Vector& arg1, const Vector& arg2);
  bool operator != (const Vector& arg1, const Vector& arg2);

  // Linear dependence
  bool lin_dependent (const Vector&, const Vector&);
  bool lin_dependent (const Vector&, const Vector&, const Vector&);

  // Affine operations
  // Point - Point
  Vector operator- (const Point& head, const Point& tail);

  // quasi-increment operators; Vector+Point deliberately not provided
  Point operator+ (const Point&, const Vector&);
  Point operator- (const Point&, const Vector&);

  // "modulo": project Point to N perp
  Point operator% (Point, const Vector& N);

} // end of namespace

#endif /* EPIX2_VECTOR */
