/* 
 *  World.cc -- epix2::World class
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 */

/* 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifdef DEBUG
#include <iostream>
#endif

#include <vector>

#include "Constants.h"

#include "Color.h"
#include "Styles.h"
#include "Knife.h"
#include "Lighting.h"

#include "Screen.h"
#include "Picture.h"

#include "Layer.h"
#include "Shard.h"
#include "Object.h"

#include "World.h"

namespace ePiX2 {

  typedef std::vector<Object*>::iterator Op;
  typedef std::vector<Object*>::const_iterator Opc;

  class Camera;
  World::World(void)
    : m_fog(White(), 0), m_ambient(White()),
      m_layers(new Layer(Vector())), m_updated(false) { }

  World::World(const World& W)
    : m_fog(W.m_fog), m_ambient(W.m_ambient), m_spots(W.m_spots)
  {
    // copy scenery
    for (Opc curr=W.m_scenery.begin(); curr!=W.m_scenery.end(); ++curr)
      m_scenery.push_back((**curr).clone());

    // if W is current, build our tree (easier than copying W's tree)
    if (W.m_updated)
      build_tree();
  }


  World& World::operator= (const World& W)
  {
    if (this == &W)
      return *this;

    // else
    // copy W's Objects
    std::vector<Object*> tmp;
    for (Opc curr=W.m_scenery.begin(); curr!=W.m_scenery.end(); ++curr)
      tmp.push_back((**curr).clone());

    // delete our Objects
    for (Opc curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      delete *curr;

    // assign
    m_fog = W.m_fog;
    m_ambient = W.m_ambient;
    m_spots = W.m_spots;

    swap(m_scenery, tmp);

    // simpler than copying W's tree
    if (W.m_updated)
      build_tree();

    return *this;
  }


  World::~World(void)
  {
    for (Opc curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      delete *curr;
  }


  // modify scenery
  World& World::operator<< (const Object& obj)
  {
#ifdef DEBUG
    std::cerr << "World::operator<< (Object) " << std::endl;
#endif
    m_scenery.push_back(obj.clone());
    m_updated=false;

    return *this;
  }

  World& World::operator<<(const World& W)
  {
    for (Opc curr=W.m_scenery.begin(); curr!=W.m_scenery.end(); ++curr)
      m_scenery.push_back((*curr)->clone());

    m_updated=false;
    return *this;
  }


  World& World::chop(const Knife& N)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->chop(N);

    m_updated=false;
    return *this;
  }

  World& World::slice(const Knife& N)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->slice(N);

    m_updated=false;
    return *this;
  }

  // private
  void World::build_tree(void)
  {
    if (!m_updated)
      {
	// collect shards
	std::vector<Shard*> all_frags;

	for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
	  (*curr)->shatter(all_frags);

#ifdef DEBUG
    std::cerr << "World::build_tree() got "
	      << all_frags.size() << " Shards" << std::endl;
#endif
	// decreasing order of area
	stable_sort(all_frags.begin(), all_frags.end(), Shard_Area());

	m_layers->clear();

	// and build
	m_layers->add(all_frags);
	m_updated=true;
      }
  }


  // apply attributes to scenery
  World& World::front_pen(const Pen& S)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->front_pen(S);

    return *this;
  }

  World& World::front_ink(const Ink& S)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->front_ink(S);

    return *this;
  }

  World& World::back_pen(const Pen& S)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->back_pen(S);

    return *this;
  }

  World& World::back_ink(const Ink& S)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->back_ink(S);

    return *this;
  }

  World& World::line_pen(const Pen& S)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->line_pen(S);

    return *this;
  }

  World& World::base_pen(const Pen& S)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->base_pen(S);

    return *this;
  }

  World& World::stencil(const Stencil& S)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->stencil(S);

    return *this;
  }

  // fine-grained stencil control
  World& World::text(const Color& color)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->text(color);

    return *this;
  }

  World& World::back(const Color& color)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->back(color);

    return *this;
  }

  World&  World::pad(const Length& len)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->pad(len);

    return *this;
  }

  World& World::edge(const Pen& pen)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->edge(pen);

    return *this;
  }


  World& World::align(const align_type A)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->align(A);

    return *this;
  }

  World&  World::font(const font_type F)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->font(F);

    return *this;
  }

  World& World::angle(const double theta)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->angle(theta);

    return *this;
  }

  // "h_offset x v_offset", e.g., "2x2" (pt), "4x-2pt", "1mm x -4pt"
  World& World::offset(std::string arg)
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->offset(arg);

    return *this;
  }

  World& World::fix()
  {
    for (Op curr=m_scenery.begin(); curr!=m_scenery.end(); ++curr)
      (*curr)->fix();

    return *this;
  }


  // set attributes on both sides
  World& World::paint(const Ink& ink)
  {
    front_ink(ink);
    back_ink(ink);
    return *this;
  }

  // face edges
  World& World::limn(const Pen& pen)
  {
    front_pen(pen);
    back_pen(pen);
    return *this;
  }

  // lines
  World& World::stroke(const Pen& pen)
  {
    line_pen(pen);
    base_pen(pen);
    return *this;
  }

  World& World::stroke(const Duo_Pen& pen)
  {
    line_pen(pen.line());
    base_pen(pen.base());
    return *this;
  }


  World& World::fog(const Color& col, const double k, const double d)
  {
    m_fog = Fog(col, k, d);
    return *this;
  }

  World& World::fog(const double k, const double d)
  {
    m_fog = Fog(White(), k, d);
    return *this;
  }

  World& World::clear_to(const double d)
  {
    m_fog.clear_to(d);
    return *this;
  }

  World& World::light(const Point& loc, const Color& color)
  {
    m_spots.push_back(Light(loc, color));
    return *this;
  }

  World& World::light(const Color& color)
  {
    m_ambient = color;
    return *this;
  }

  // map scenery (private)
  void World::shoot(Screen& S, const Camera& cam, const photo_type T)
  {
#ifdef DEBUG
    std::cerr << "World::shoot()" << std::endl;
#endif
    if (!m_updated)
      build_tree();

    m_layers->shoot(S, cam, m_ambient, m_spots, m_fog, T);
  }

  // public wrappers for shoot()
  void World::photo(Screen& S, const Camera& cam)
  {
    shoot(S, cam, PHOTO);
  }
  void World::flash(Screen& S, const Camera& cam)
  {
    shoot(S, cam, FLASH);
  }
  void World::trace(Screen& S, const Camera& cam)
  {
    shoot(S, cam, TRACE);
  }
  void World::x_ray(Screen& S, const Camera& cam)
  {
    shoot(S, cam, X_RAY);
  }

  void World::photo(Picture& S, const Camera& cam)
  {
    shoot(S.screen(), cam, PHOTO);
  }
  void World::flash(Picture& S, const Camera& cam)
  {
    shoot(S.screen(), cam, FLASH);
  }
  void World::trace(Picture& S, const Camera& cam)
  {
    shoot(S.screen(), cam, TRACE);
  }
  void World::x_ray(Picture& S, const Camera& cam)
  {
    shoot(S.screen(), cam, X_RAY);
  }
} // end of namespace
