/* 
 * Camera.h -- epix2::Camera interface
 *
 * This file is part of ePiX2, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 12, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * ePiX2 provides a Camera to render a World in a Picture. A Camera  *
 * has a location, orientation, lens, and color filter. The viewer   *
 * sits at a specified distance from a target, the point the Camera  *
 * looks at. The viewing angle is controlled by changing the viewer, *
 * target, and/or the distance between them.                         *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef EPIX2_CAMERA
#define EPIX2_CAMERA

#include "Constants.h"
#include "Point.h"
#include "Basis.h"
#include "Color.h"

namespace ePiX2 {

  class Pair;
  class Vector;

  class Camera {
  public:
    Camera(const Point vpt = Point(0,0,EPIX2_INFTY));
    ~Camera(void);

    // rotate
    Camera& tilt(const double angle); // up/down (pitch)
    Camera& pan(const double angle);  // left/right (yaw)
    Camera& roll(const double angle); // about viewing axis

    // fix target, move viewpt radially along eye()
    Camera& range(const double);
    // fix viewpt, move target radially along eye()
    Camera& focus(const double);

    Camera& at(const Point& arg);      // fix target, set viewpt
    Camera& look_at(const Point& arg); // fix viewpt, set target

    Camera& filter(const Color&);

    // set lens
    Camera& perspective(void); // default
    Camera& orthog(void); // faster
    Camera& fisheye(void);
    Camera& bubble(void);

    // Functions for internal use
    // Camera maps
    Color operator() (const Color&) const; // filter
    Pair  operator() (const Point&) const; // lens map
    bool  is_linear(void); // lens preserves lines?

    // situation
    Vector eye(void) const;
    Point viewpt(void) const;

  private:
    class Lens {
    public:
      // lens map(arg, orient, viewpt, distance)
      virtual Pair operator() (const Point&, 
	 const Basis&, const Point&, const double) const = 0;
      virtual bool is_linear(void) const = 0; // lens preserves lines?
      virtual ~Lens(void) { }
    }; // end of class Lens

    /* The implementation data comprise:
     *   - a Basis ({sea, sky, eye}, with the intuitive meanings:)
     *   - a viewpoint
     *   - a target (at the center of the field of view)
     *   - the distance from the viewpoint to the target
     *   - a "filter" color, White by default, for color separation, etc.
     *   - A Lens (below)
     *
     * The viewer sits at given <distance> from the <target>, in the
     * direction of the <eye>. The vectors <sea> and <sky> point
     * horizontally to the right and vertically upward.
     *
     * Conceptually, a Lens provides a mapping Point -> Pair. In the
     * implementation, the Camera ("body") passes the orientation, target,
     * and distance to the Lens operator. The Camera contains a pointer to
     * Lens, a base class. Adding a new Lens entails:
     *   - Declaring a new child of Camera::Lens
     *   - Adding a pointer-managing function to set the lens
     */
    Point the_viewpt; // center of projection
    Point the_target; // mapped to Screen origin
    Basis the_orient; // our {sea, sky, eye}
    double the_distance;

    Color the_filter;
    void adjust(); // re-orient according to target, viewpt
    Lens* the_lens;


    // actual lenses
    class Perspective : public Lens {
    public:
      Pair operator()
	(const Point&, const Basis&, const Point&, const double) const;
      bool is_linear(void) const { return true; }
    }; // end of class Perspective

    class Orthog : public Lens {
    public:
      Pair operator()
	(const Point&, const Basis&, const Point&, const double) const;
      bool is_linear(void) const { return true; }
    }; // end of class Orthog

    class Fisheye : public Lens {
    public:
      Pair operator()
	(const Point&, const Basis&, const Point&, const double) const;
      bool is_linear(void) const { return false; }
    }; // end of class Fisheye

    class Bubble : public Lens {
    public:
      Pair operator()
	(const Point&, const Basis&, const Point&, const double) const;
      bool is_linear(void) const { return false; }
    }; // end of class Bubble

  }; // end of class Camera

} /* end of namespace */

#endif /* EPIX2_CAMERA */
