/*
 *  Color.h -- epix2::Color class and helpers
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 25, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*
 * This file defines the ePiX2 color interface:
 *   - Color_Base: The abstract interface
 *   - Color: The user interface
 *       operator*=           Modify intensity
 *       blend(Color, double) Color mixing
 *       filter(Color)        Apply us as a filter
 *       superpose(Color)     Add colors, clip intensities
 *       invert()             Color negative
 *
 *   - Named primary constructors
 *
 * Scaling and filtering are implemented by converting to the filter's
 * color model. Color model conversions come from the LaTeX xcolor
 * package documentation, V 1.11 (2044/05/09) by Uwe Kern.
 */

#ifndef EPIX2_COLOR
#define EPIX2_COLOR

#include <string>

namespace ePiX2 {

  class Format;

  class Color_Base {
    friend class Color;

  protected:
    struct RGB_Densities
    {
      // ensures densities are in [0,1]
      RGB_Densities (const double r, const double g, const double b);

      /*
      double red()   const { return m_dens_red; }
      double green() const { return m_dens_green; }
      double blue()  const { return m_dens_blue; }
      */
      double m_dens_red, m_dens_green, m_dens_blue;
    };

  public:
    virtual ~Color_Base() { }
    virtual Color_Base* clone() const = 0;

    // apply us as filter
    virtual Color filter(const Color_Base&) const = 0;

    // blending (all non-const)
    // scale intensities
    virtual Color_Base& operator*= (const double) = 0;

    // clip d to [0,1], then return (1-d)*this + d*col
    virtual Color_Base& blend(const Color_Base&, const double) = 0;

    // add channels
    virtual Color_Base& superpose(const Color_Base&) = 0;

    // return negative color
    virtual Color_Base& invert() = 0;

    // internal colorname and densities
    virtual std::string name() const = 0;
    virtual std::string print_to(const Format&) const = 0;

    bool operator== (const Color_Base& col) const;
    virtual RGB_Densities to_rgb() const = 0;

  }; /* end of class Color_Base */


  // Color_Base handle
  class Color {
  public:
    Color();
    Color(const Color_Base&);
    Color(const Color&);

    Color& operator= (const Color&);

    ~Color();

    const Color_Base* operator->() const;
    double alpha() const;

    bool operator== (const Color&) const;
    bool operator< (const Color&) const; // needed for Screen::pallet
    bool operator!= (const Color& col) const { return !((*this) == col); }
    bool is_unset() const;

    Color filter(const Color&) const;

    Color& operator*= (const double c);     // scale the intensity
    Color& blend(const Color&, const double);
    Color& superpose(const Color&);
    Color& invert();
    Color& alpha(const double);

    std::string name() const;

    std::string print_to(const Format&) const;

  private:
    Color_Base* m_color;
    double m_alpha;

    Color (const Color_Base*);
  }; // end of class Color

  Color operator* (const double, const Color&);


  /*
   * * *  Primary color functions in the global namespace * * *
   *
   * A primary color constructor accepts a numerical argument, which is
   * reduced mod 4, then interpreted as a color density as follows:
   * d = -2 or 2 : white
   * d = -1: full anti-saturation
   * d =  0: black
   * d =  1: full saturation (default)
   *
   * The primary changes continuously, and linearly on each interval
   * [-2,-1], [-1,0], [0,1], and [1,2].
   *
   * Example:
   * Red(0.3) = rgb(0.3, 0, 0) = Red(-3.7)
   * Red(1.3) = rgb(1,0.3,0.3) = Red(-2.7)
   * Red(2.3) = rgb(0.7, 1, 1) = Red(-1.7)
   * Red(3.3) = rgb(0,0.7,0.7) = Red(-0.7)
   */
  Color RGB(const double r=0, const double g=0, const double b=0);
  Color CMY(const double c=1, const double m=1, const double y=1);
  Color CMYK(const double c=1, const double m=1, const double y=1,
	     const double k=0); // void same as CMYK(0,0,0,1);
  Color Gray(const double d=0);

  Color Red(const double d=1);
  Color Green(const double d=1);
  Color Blue(const double d=1);
  Color White(const double d=1);
  Color Black(const double d=1);

  Color Cyan(const double d=1);
  Color Magenta(const double d=1);
  Color Yellow(const double d=1);
  Color CMY_White(const double d=1);
  Color CMY_Black(const double d=1);

  Color CyanK(const double d=1);
  Color MagentaK(const double d=1);
  Color YellowK(const double d=1);
  Color CMYK_White(const double d=1);
  Color CMYK_Black(const double d=1);

  // color separation
  Color C_Process(const double d=1);
  Color M_Process(const double d=1);
  Color Y_Process(const double d=1);
  Color K_Process(const double d=1);

  // "conversion filters"
  Color Neutral();
  Color RGB_Neutral();
  Color CMY_Neutral();
  Color CMYK_Neutral();
  Color Gray_Neutral();

} /* end of namespace */

#endif /* EPIX2_COLOR */
