/* 
 * Mask.h -- ePiX2::Rect class
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 23, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 
 * This file provides the Mask class:
 *   - Alignment points, horizontal/vertical sizes, bounds
 *   - Tile cropping
 *
 * Rect: Rectangular Mask
 */

#ifndef EPIX2_MASK
#define EPIX2_MASK

#include <vector>
#include "Pair.h"

namespace ePiX2 {

  class Color;
  class Length;
  class Pen;
  class Edge2d;

  class Tile;
  class Plate2d;
  class Outline2d;

  namespace Mask {
    // Base class for masking and extracting from a Screen
    class Mask {
    public:
      virtual ~Mask(void) { }

      virtual Mask* clone() const = 0;

      // alignment points
      Pair tr(void) const;
      Pair tl(void) const;
      Pair bl(void) const;
      Pair br(void) const;

      Pair  t(void) const;
      Pair  b(void) const;
      Pair  l(void) const;
      Pair  r(void) const;
      Pair  c(void) const;

      // cropping facilities
      virtual void crop(Tile*) const = 0;
      virtual bool encloses(const Pair&) const = 0;

      // extreme and average coordinates, dimensions
      virtual double h_min(void)  const = 0;
      virtual double h_max(void)  const = 0;
      virtual double h_size(void) const = 0;
      virtual double h_avg(void)  const = 0;

      virtual double v_min(void)  const = 0;
      virtual double v_max(void)  const = 0;
      virtual double v_size(void) const = 0;
      virtual double v_avg(void)  const = 0;

      // convert us to Screen border or backing
      virtual Outline2d  border(const Pen&) const = 0;
      virtual Plate2d   backing(const Color&) const = 0;

      Outline2d  border(const Color&, const Length&) const;
    }; // end of class Mask

    class Rect : public Mask {
    public:
      Rect(void) { }
      Rect(const Pair& arg1, const Pair& arg2);

      Rect* clone() const;
      double h_min(void)  const;
      double h_max(void)  const;
      double h_size(void) const;
      double h_avg(void)  const;

      double v_min(void)  const;
      double v_max(void)  const;
      double v_size(void) const;
      double v_avg(void)  const;

      void crop(Tile*) const;
      bool encloses(const Pair&) const;

      Outline2d  border(const Pen&) const;
      Plate2d   backing(const Color&) const;

    private:
      double m_hmin, m_hmax;
      double m_vmin, m_vmax;
    }; // end of class Rect
  } // end of Mask namespace
} // end of ePiX2 namespace

#endif /* EPIX2_MASK */
