/* 
 * Pair.h -- ePiX2::Pair_Map(_Base) classes
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 21, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 
 * This file provides the classes
 * [] Pair_Map_Base
 *   - Pair-valued mapping wrapper
 *
 * [] Pair_Map_{Fcn, Affine, Rotate, Reflect}
 *   - Child classes that implement function maps, affine scaling, 
 *     rotation, and reflection
 *
 * [] Pair_Map
 *   - A composable wrapper class for Pair_Map_Base, which provides
 *     evaluation and composition ()-operators
 */

#ifndef EPIX2_PAIR_MAP
#define EPIX2_PAIR_MAP

#include <cmath>
#include "Pair.h"
#include "Mask.h"

namespace ePiX2 {

  class Pair_Map;

  // named constructors
  Pair_Map Function(Pair f(Pair));
  Pair_Map Affine(const Mask::Mask& prnt, const Mask::Mask& chld);
  Pair_Map Rotate(const double th=0, const Pair ctr=Pair(0,0));
  // fixed points
  Pair_Map Reflect(const Pair p1, const Pair p2);
  // fixed point and angle of axis
  Pair_Map Reflect(const double th=0, const Pair ctr=Pair(0,0));

  // composable wrapper class for Base
  class Pair_Map {
    class Map_Base;

    friend Pair_Map Function(Pair f(Pair));
    friend Pair_Map Affine(const Mask::Mask&, const Mask::Mask&);
    friend Pair_Map Rotate(const double, const Pair);
    friend Pair_Map Reflect(const Pair, const Pair);
    friend Pair_Map Reflect(const double, const Pair);

  public:
    Pair_Map(); // identity
    Pair_Map(const Map_Base&);

    Pair_Map(const Pair_Map&);
    Pair_Map& operator=(const Pair_Map&);
    ~Pair_Map();

    Pair_Map* clone() const;

    // evaluation
    Pair operator() (const Pair) const;
    // composition
    Pair_Map operator() (const Pair_Map&) const;

  private:
    Map_Base* outer; // the mapping we represent
    Pair_Map* inner; // precompose with this

    // Implementation interface
    class Map_Base {
    public:
      virtual ~Map_Base(void) { }
      virtual Map_Base* clone() const = 0;
      virtual Pair operator() (const Pair) const = 0; // evaluation
    }; // end of class Map_Base


    // Map_Base implementations
    class Func : public Map_Base {
    public:
      Func(Pair (*f)(Pair));

      Func* clone() const;
      Pair operator() (const Pair arg) const;

    private:
      Pair (*mapping)(Pair); // the mapping we represent
    }; // end of class Func

    class Affine : public Map_Base {
    public:
      // affine map from child onto parent
      Affine(const Mask::Mask& parent, const Mask::Mask& child);
      Affine* clone() const;
      Pair operator() (const Pair arg) const;

    private:
      Pair p_bl;  // parent lower left corner
      Pair c_bl;  // child
      Pair scale; // affine scale factor from child to parent
    }; // end of class Affine

    class Rotate : public Map_Base {
    public:
      Rotate(const double th=0, const Pair ctr=Pair(0,0));
      Rotate* clone() const;
      Pair operator() (const Pair arg) const;

    private:
      double angle;
      Pair center;
    }; // end of class Rotate

    class Reflect : public Map_Base {
    public:
      // fixed points
      Reflect(const Pair p1, const Pair p2);
      // fixed point and angle of axis
      Reflect(const double th=0, const Pair ctr=Pair(0,0));

      Reflect* clone() const;
      Pair operator() (const Pair arg) const;

    private:
      Pair fix1, fix2; // fixed points
    }; // end of class Reflect

  }; // end of Pair_Map class
} // end of ePiX2 namespace

#endif /* EPIX2_PAIR_MAP */
