/* 
 *  Shard_2d.h -- epix2::Plate3d class
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * This file provides:
 *
 *   Plate3d - A convex planar polygon or polyline and draw functions
 *          .        
 *    Front .  Back
 *          .        
 *   *------*- - - -o
 *   |xxxxxx:       
 *   |xxxxxx:       |
 *   |xxxxxx:       
 *   |xxxxxx:       |
 *   |xxxxxx:       
 *   *------*- - - -o
 *          .
 *          .        
 *          .
 *
 *   . . . Cutting plane
 *   ----- Visible edge
 *   - - - Edge of back portion
 *     :   Cut edge (undrawn)
 *     *   Visible vertex
 *     o   Vertex of "back" portion
 *     x   Interior of front portion
 * 
 *  A Plate3d can be "photographed" to a Plate2d in the following styles:
 *
 *   - photo:  Interiors shaded according to face normal
 *   - flash:  Interiors colored solid
 *   - trace:  Edges only, with hiding
 *   - x_ray:  Edges only, no hiding
 */

#ifndef EPIX2_SHARD_2D
#define EPIX2_SHARD_2D

#include <vector>
#include <string>

#include "Vector.h"

#include "Edge_Data.h"

#include "Styles.h"
#include "Shard.h"

namespace ePiX2 {
  class Point;

  class Layer;

  class Gyro;

  class Screen;
  class Camera;

  class Plate3d : public Shard {
  public:
    Plate3d(const Ink&, const Pen&, // front style (face and edge))
	    const Ink&, const Pen&, // back style
	    bool  clipped,              // all edges drawn?
	    const Vector&);             // perp vector

    Plate3d(const Vector&);

    Plate3d* clone(void) const;

    // half-space cut; Xfine leaves cuts undrawn
    Plate3d*  chop(const Knife&, std::vector<Shard*>&);

    // quasi-planar cut; no default Pen
    Plate3d* slice(const Knife&, std::vector<Shard*>&);

    Plate3d& front_pen(const Pen&);
    Plate3d& front_ink(const Ink&);
    Plate3d& back_pen(const Pen&);
    Plate3d& back_ink(const Ink&);

    Plate3d& fix();

    // photograph to specified Screen
    void shoot(Screen&, const Camera&, const Color& ambient,
	       const std::vector<Light>& spots, const Fog&,
	       photo_type) const;

    void update(void) const; // update area, location, and normal vector

    // Returns area, for sorting before adding to BSP, and updates location
    double priority(void) const;

    bool is_empty(void) const;
    Plate3d* in_coords(const Gyro&);

    // place in BSP
    void add_to(Layer*) const;

    // We occlude X from vpt?
    bool hides(const Point& vpt, const Point& X) const;

    // Our location, for reflectivity calculation
    Point here(void) const;
    // Unit normal vector
    Vector perp(void) const;

    // specal to Plate3d
    Plate3d* add(const Edge3d&);

  private:
    Ink m_front_ink;
    Pen m_front_pen;

    Ink m_back_ink;
    Pen m_back_pen;

    bool m_clipped; // some edge(s) undrawn

    // bookkeeping data
    Vector m_perp;          // unit normal; location = m_perp.tail()
    mutable double m_area;  // true area
    mutable bool m_updated; // area current?

    std::vector<Edge3d> m_border; // "default" appearance
  }; // end of class Plate3d

} /* end of namespace */

#endif /* EPIX2_FACE_REP */
