/* 
 *  Styles.h -- ePiX2 style data structures
 *
 * This file is part of ePiX, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*
 * This file provides classes for defining and manipulating the
 * appearances of path-like, surface-like, and textual objects.
 * Classes are built in stages from Colors, Lengths, and other
 * parameters that represent shininess, transparency, label padding, etc.
 *
 * - Ink
 *   Acts like a Color with shine, optional permanence
 *
 * The high-level classes below provide a relatively mnemonic get/set
 * interface:
 *
 * - Pen (3d or 2d)
 *   For drawing line-like objects; has Ink or Color, and width.
 *
 *   Line-like Objects use two Pen2ds:
 *   -------------------------------  <--
 *   ==== line (default: Black) ====     |--- base (default: White)
 *   -------------------------------  <--
 *
 *
 * - Surface-like objects
 *   Ink and a Pen2d, for filling regions and drawing their borders.
 * 
 *   o-----------------o
 *   |/////////////////|------ edge (Pen)
 *   |/////////////////|
 *   |//////////O------------- fill (Ink or Color)
 *   |/////////////////|
 *   |/////////////////|
 *   o-----------------o
 *
 * - Stencil
 *   Data for drawing text boxes: Text and background colors, a Nib (for
 *   drawing a box around a label), padding (a Length), alignment, font,
 *   and rotation angle.
 *
 *   o-----------------------o
 *   |                  O--------- back (Ink or Color)
 *   |                       |
 *   |                       |---- edge (Pen)
 *   |    Your quote here--------- text (Ink or Color)
 *   |                       | -
 *   |                       | |__ pad (Length)
 *   |                       | |
 *   o-----------------------o -
 *
 *       Stencil& text(const Ink&);
 *       Stencil& back(const Ink&);
 *       Stencil&  pad(const Length&);
 *       Stencil& edge(const Pen&);
 *
 *       Ink   text() const;
 *       Ink   back() const;
 *       Length pad() const;
 *       Pen   edge() const;
 *
 */
#ifndef EPIX2_STYLES
#define EPIX2_STYLES

#include "Constants.h"
#include "Length.h"
#include "Color.h"

namespace ePiX2 {

  class Ink {
  public:
    // A Color can be promoted to Ink
    // flat, non-permanent
    Ink(const Color& col, double r=0, bool p=false);

    // replace if we're not permanent
    Ink& apply(const Ink&);

    // mix colors if we're not permanent
    Ink& blend(const Ink&, const double);

    // Color negative if we're not permanent
    Ink& invert();

    // get
    Color  color() const;       // our Color
    Ink&   gloss(const double); // set/get reflectivity
    double gloss() const;
    double alpha() const;       // get transparency
    void   fix();               // make permanent

    bool indelible() const;

  private:
    Color  m_color;
    double m_gloss;
    double m_alpha;  // transparency
    bool m_perm;     // permanent?
  }; // end of class Ink


  // Ink and width for line-like scene elements
  class Pen {
  public:
    Pen();
    Pen(const Ink& p, const Length& wid);
    Pen(const Ink& p, std::string wid); // pass by value

    static Pen Xfine; // width 0, for Knife with invisible cuts

    // set color (if not permanent)
    Pen& apply(const Pen&);
    // Pen& apply(const Ink&);

    // get
    Color  color() const;
    Length width() const;
    double gloss() const;
    double alpha() const;

    Pen& fix();  // make ink permanent

  private:
    Ink  m_ink;
    Length m_width;  
  }; // end of class Pen


  // two pens, for convenience with 1-d Objects
  class Duo_Pen {
  public:
    Duo_Pen(const Pen& line, const Pen& base);

    Duo_Pen(const Color& c_line, const std::string& w_line,
	    const Color& c_base, const std::string& w_base);

    Pen line() const;
    Pen base() const;

  private:
    Pen m_line;
    Pen m_base;
  }; // end of Duo_Pen


  class Pen2d {
  public:
    Pen2d(const Color& col, const Length& wid);

    static Pen2d Xfine;

    // get
    Color  color() const;
    Length width() const;

    bool operator== (const Pen2d& pen) const;
    bool operator!= (const Pen2d& pen) const;

  private:
    Color  m_color;
    Length m_width;  
  }; // end of class Pen2d


  // for textual Elements
  class Stencil {
    friend class Glyph3d;
  public:
    Stencil(const Ink& text=Black(), const Ink& back=Neutral(),
	    const Length& pad=LABEL_PAD, const Pen& edge=Pen::Xfine,
	    align_type align=B, font_type font=ROMAN, const double th=0);

    // set
    Stencil& apply(const Stencil&);
    Stencil& text(const Color&);
    Stencil& back(const Color&);
    Stencil&  pad(const Length&);
    Stencil& edge(const Pen&);

    Stencil& align(const align_type);
    Stencil&  font(const font_type);
    Stencil& angle(const double);

    Stencil& fix();

    // get
    Color  text() const;
    Color  back() const;
    Length pad() const;
    Pen    edge() const;

  private:
    Ink    m_text;
    Ink    m_back;
    Length m_pad;
    Pen    m_edge;

    align_type m_align;
    font_type  m_font;
    double     m_angle; // stored in degrees
  }; // end of Stencil

  // current text style
  Stencil& text_style();

  // named Pen constructors
  Pen plain(const Color&);
  Pen  bold(const Color&);
  Pen bbold(const Color&);

} /* end of namespace */

#endif /* EPIX2_STYLES */
