/* 
 *  World.h -- epix2::World interface
 *
 * This file is part of ePiX2, a program for creating high-quality 
 * figures in LaTeX 
 *
 * Version 2.0pre
 * Last Change: August 16, 2006
 *
 * 
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 *
 *
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * ePiX2 provides a 3-dimensional scene description language, World. *
 * A World contains Objects (q.v.), which can be chopped, colored,   *
 * and lighted. The entire World can be "photographed" in various    *
 * combinations of face shading and edge hiding.                     *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef EPIX2_WORLD
#define EPIX2_WORLD

#include <vector>

#include "Constants.h"

#include "Color.h"
#include "Fog.h"
#include "Lighting.h"

namespace ePiX2 {

  class Camera;
  class Screen;
  class Picture;

  class Knife;

  class Ink;
  class Pen;
  class Stencil;

  class Object;

  class Layer;

  class World {
  public:

    World(void);

    World(const World&);
    World& operator= (const World&);
    ~World(void);

    // modify scenery
    World& operator<< (const Object&);
    World& operator<< (const World&);

    World& chop(const Knife&);  // chop by half-space
    World& slice(const Knife&); // planar slice

    // apply attribute(s):
    // 2-dimensional elements
    World& front_pen(const Pen&);
    World& front_ink(const Ink&);

    World& back_pen(const Pen&);
    World& back_ink(const Ink&);

    // 1-dimensional
    World& line_pen(const Pen&);
    World& base_pen(const Pen&);

    // and 0-dimensional
    World& stencil(const Stencil&);
    World& text(const Color&);
    World& back(const Color&);
    World&  pad(const Length&);
    World& edge(const Pen&);

    World& align(const align_type);
    World&  font(const font_type);
    World& angle(const double);
    // "h_offset x v_offset", e.g., "2x2" (pt), "4x-2pt", "1mm x -4pt"
    World& offset(std::string);

    World& fix();

    // for convenience
    World& paint(const Ink&); // color both faces
    World& limn(const Pen&);  // color both edges

    World& stroke(const Pen&); // set line style
    World& stroke(const Duo_Pen&);


    World& light(const Point&, const Color& col = White()); // point source
    World& light(const Color&);               // ambient

    // density and clear distance
    World& fog(const Color& c, const double, const double d=0);
    World& fog(const double, const double d=0); // White() fog
    World& clear_to(const double d);

    // map scenery
    void photo(Screen&,  const Camera&); // face-normal shading
    void photo(Picture&, const Camera&);

    void flash(Screen&,  const Camera&); // flat shading
    void flash(Picture&, const Camera&);

    void trace(Screen&,  const Camera&); // edges only, hiding
    void trace(Picture&, const Camera&);

    void x_ray(Screen&,  const Camera&); // edges only, no hiding
    void x_ray(Picture&, const Camera&);

  private:
    Fog m_fog;
    Color m_ambient;
    std::vector<Light> m_spots;

    std::vector<Object*> m_scenery;

    Layer* m_layers;
    bool m_updated; // m_layers current?

    void build_tree(void);
    void shoot(Screen&, const Camera&, const photo_type);

  }; // end of class World

} /* end of namespace */

#endif /* EPIX2_WORLD */
